/*
 *                 Copyright (c) Mar 28 10:38:21  MTCC
 *	Author: Michael Thomas
 *	Module: cgidbg.c
 *	Created: Fri Mar 28 10:38:21 1997
 *	Abstract:
 *	   main module of the CGI debugger. This thingy is
 *         basically not much more than a plumber setting
 *         up the proper pipes and such to get the job done.
 */

/* Edit History: 
 */

#include <html.h>
#include <sys/time.h>
#include <signal.h>
#include <wait.h>


#define JAVAPAGE "cgidbg.html"


char serverroot [256];



static void reaper ();
static void error(char * fmt, ...);

char * cgicmd, * cgiparm;
char * grabpost ();
int textmode = 0;

struct dbgstrm 
{
    int gdbin, gdbout;
    int cgiin, cgirfd;
    int pid;
    int dbgpid;
    int sts;
    int pgrp;
    int tmo;
} dbg;


/*
 * filter the input before we send it to the debugger.
 * traps ^C's and other things like that.
 * returns what type of input it is.
 */


#define NORMAL  0
#define META    1
#define SIGKEY  2

int GETS (char * bf) 
{
    char * p = bf;
    int rv = 0, n;

    /* yes, this unbuffered shit is wasteful */
    while ((n = read (0, p, 1)) > 0) {
        if (* p == '\n') {
            p++;
            break;
        }
        if (* p == '\003')
            rv = SIGKEY;
        p++;
    }
    * p = 0;    
    if (n <= 0)
        return -1;
    if (rv == SIGKEY)
        return SIGKEY;
    if (bf [0] == '~')
        return META;
    return NORMAL;
}

/* giveup after a certain amount of time on the accept calls */
void
giveup (int sigctx) 
{
    error ("giving up after %d seconds.\n", dbg.tmo);
    exit (1);
}


static void setcgi (char * str) 
{
    char * bf;
    char * cmd, * parm;

    if (str) {
        bf = alloca (strlen (str));
        unwebify (str, bf);
        if ((parm = strchr (bf, '?')))
            * parm++ = 0;
        else
            parm = "";
        cmd = bf;
        
        cgicmd = malloc (strlen (DEBUGGER)+strlen (cmd)+strlen (serverroot)+1);
        if (* cmd == 0)
            sprintf (cgicmd, "%s", DEBUGGER);
        else
            sprintf (cgicmd, "%s %s%s", DEBUGGER, serverroot, cmd);    
        cgiparm = strdup (parm);
    } else {
        cgicmd = malloc (strlen (DEBUGGER)+1);
        sprintf (cgicmd, "%s", DEBUGGER);
        cgiparm = strdup ("");
    }    
}


static int execcmd (char * cmd) 
{
    int gin [2], gout [2], cgiin [2];

    pipe (gin);
    pipe (gout);
    pipe (cgiin);
    

    /* gdbin/out refer to how gdb sees all this */
    dbg.gdbin = gin [1];
    dbg.gdbout = gout [0];
    dbg.cgiin = cgiin [1];
    dbg.cgirfd = cgiin [0];
    
    if ((dbg.pid = fork ())) {
        if (dbg.pid < 0) 
            return -1;
        /* parent */
        close (gin [0]);
        close (gout [1]);
        close (cgiin [0]);
    } else {
        if (dup2 (gin [0], 0) < 0)
            return -1;
        if (dup2 (gout [1], 1) < 0)
            return -1;
        if (dup2 (gout [1], 2) < 0)
            return -1;
        /* close dup'ed and unnecessary pipes */
        close (gin [0]);
        close (gin [1]);
        close (gout [0]);
        close (gout [1]);
        close (cgiin [1]);
        /* set shell lest they inherit csh or something
           bizarre for the IO redirection cruft. */
        setenv ("SHELL", SHELL, 1);
        /* need to do more goodies with gdb to setup IO & args */
        if (execl (SHELL, SHELL,  "-c", cmd, 0) < 0)
            return -1;
    }
    return 0;
}
    
    
static void reaper (int arg) 
{
    
    signal (SIGCHLD, reaper);
    wait (&dbg.sts);
    dbg.pid = -1;    
 }



static void syserror (char * str) 
{
    extern int errno;
    error ("%s: %s\n", str, strerror (errno));
}


static void 
error (char * fmt, ...) 
{
    va_list ap;
    char bf [512];

    VA_START (ap, fmt);
    vsprintf (bf, fmt, ap);
    va_end (ap);
    if (textmode) {
        printf (bf);
    } else {
        hterase ();
        preHTML ();
        startHTML ("cgidbg error", 0);
        par ();
        htprintf ("an error has occurred: %s\n", bf);
        endHTML ();
    }
    exit (0);
}

int safewrite (int fd, char * bf, int n) 
{
    while (n > 0) {
        int rv = write (fd, bf, n);
        if (rv <= 0)
            return rv;
        bf += rv;
        n -= rv;
    }
    return 1;
}



char *
grabpost (int tmo, int * plen) 
{
    struct timeval t;
    fd_set fds;
    char * bf, * p;
    int rv;
    
    FD_ZERO (&fds);
    FD_SET (0, &fds);
    t.tv_sec = 0;
    t.tv_usec = tmo*1000;
    * plen = 0;
    if (select (32, &fds, 0, 0, &t) <= 0) 
        return strdup ("");
    if (FD_ISSET (0, &fds)) {
        int tlen = 1024;
        p = bf = malloc (1024);
        while ((rv = read (0, p, 512)) > 0) {
            p += rv;
            /* check for overflow */
            if (p - bf + 512 > tlen) {
                char * t;
                tlen += 1024;
                t = realloc (bf, tlen);
                p = t + (p-bf);
                bf = t;
            }
        }
        * plen = p-bf;
        return bf;
    }
    return strdup ("");
}



void detach()
{
    chdir("/");
    if((dbg.dbgpid = fork()) > 0)
        exit(0);
    else if(dbg.dbgpid == -1) {
        exit(1);
    }
#ifndef NO_SETSID
    if((dbg.pgrp=setsid()) == -1) {
        syserror("setsid");
        exit(1);
    }
#else
    if((dbg.pgrp=setpgrp(getpid(),0)) == -1) {
        syserror ("setpgrp");
        exit(1);
    }
#endif    
}




int main (int ac, char ** av) 
{

    int lfd1, lfd2, dbgfd, cgifd;
    int dbgport = 0, cgiport = 0;

    char * post, * env;
    int plen;
    htelm * h;
    char hostname [256], jpage [256];
    char args [1024];
    fd_set fds, selmask;
    
    preHTML ();

    /* set the server root. we'll use the environment if
       possible, otherwise we use the compiled version */
    if ((env = getenv ("DOCUMENT_ROOT"))) {
        strcpy (serverroot, env);
        if ((env = strrchr (serverroot, '/'))) {
            /* make certain it isn't of the form path/path/path/ */
            if (env [1] == 0) {
                * env = 0;
                /* try again. */
                if ((env = strrchr (serverroot, '/')))
                    * env = 0;
            } else
                * env = 0;
        }
    } else
        strcpy (serverroot, SERVERROOT);    
    
    gethostname (hostname, 256);
    sprintf (jpage, "%s/%s", DEBUGHTML, JAVAPAGE);    
    h = htreadurl (jpage, 0, 0);

    if (h == 0)
        error ("%s is not installed.", jpage);
    if ((env = getenv ("QUERY_STRING")))
        setcgi (env);
    else
        setcgi (ac > 0 ? av [1] : 0);
    signal (SIGCHLD, reaper);    

    lfd1 = listenon (hostname, &dbgport);
    /* this should really be done on a per run basis...? */
    lfd2 = listenon (hostname, &cgiport);

    /* grab any post data at this point */
    post = grabpost (100, &plen);
    /* now, set up a listener socket for the applet
       to reconnect to us */
    htgotomark ("phead");
    htprintf ("Debugging (%s) with args=%s ", cgicmd, cgiparm);
    htgotomark ("params");
    htprintf ("<param name=host value=%s>\n", hostname);
    htprintf ("<param name=dbgport value=%d>\n", dbgport);
    /* insert code to make new window/frame */
    htgotomark ("script");
    htprintf ("window.open (\"%s/cgife?host=%s&port=%d\", \"debugoutput\", \"directories=no,location=no\");\n",
              DEBUGCGI, hostname, cgiport);
    /* emit the file */ 
    htdumpfile ();
    /* we detach from the mothership because after closing
       our channels down (thus signaling html is done), they
       kill us. How rude! We, however, can run *and* hide.
     */    
    detach ();    
    close (0);
    close (1);
    close (2);
    signal (SIGALRM, giveup);
    alarm (dbg.tmo = 60);
    textmode = 1;
    if ((dbgfd = accepton (lfd1, 0)) < 0)
        syserror ("cgidbg: can't accept");
    
    /*
     * now, dup the socket we just connected on the applet end
     * to (now) fully formed connection and field requests to
     * the debugger...
     */
    if (dup2 (dbgfd, 0) < 0)
        syserror ("can't dup2 net");    
    if (dup2 (dbgfd, 1) < 0)
        syserror ("can't dup2 net");
    if (dup2 (dbgfd, 2) < 0)
        syserror ("can't dup2 net");
    setlinebuf (stdout);
    setlinebuf (stderr);
    if ((cgifd = accepton (lfd2, 0)) < 0) 
        syserror ("can't accept");
    /* don't need the listeners anymore */
    close (lfd1);
    close (lfd2);
    if (execcmd (cgicmd) < 0)
        syserror ("can't fork debugger.");
    /* we no longer need cgifd */
    close (cgifd);
    /* prefeed a bit of cruft to make life easier
       Note: we leave the cgi program's stderr point at
       the *debugger's* stderr. This is reasonable since
       we don't want it to land in the log or (worse) in
       the output stream. all the sudden-like, we can
       see stderr. k00l beans.
     */

    /* now loop waiting for input to send to the debugger */
    FD_ZERO (&fds);
    FD_SET (dbg.gdbout, &fds);
    FD_SET (0, &fds);

    /* why, you ask do we use a pipe in between the cgi program
       and the browser? Two reasons: for some reason directly
       plumbing it seems to elicit a SIGTTIN. I'm not sure why
       as output isn't getting a SIGTTOU. Secondly, it just seemed
       like a not so bad idea that we might be able to intercept
       the browswer spew to perhaps send it back to the applet,
       hence I just go with the flow.
    */
    sprintf (args, "set args 0<&%d 1>&%d %s\n",
             dbg.cgirfd, cgifd, cgiparm);
    fprintf (stderr, "input from browser: %d bytes\n", plen);
    fprintf (stderr, "setting args to: %s\n", args);

    write (2, post, plen);
    
    write (dbg.gdbin, args, strlen (args));
    alarm (dbg.tmo = 900);
    if (plen)
        write (dbg.cgiin, post, plen);
    close (dbg.cgiin);
    while (1) {
        char bf [1024];
        
        int rv;
        selmask = fds;        
        rv = select (32, &selmask, 0, 0, 0);
        if (rv < 0) {
            if (errno == EINTR)
                continue;
            syserror ("select failed.");
        }
        /* check for gdb output: off to applet's debug window */
        if (FD_ISSET (dbg.gdbout, &selmask)) {
            alarm (dbg.tmo);
            if ((rv = read (dbg.gdbout, bf, 1024)) <= 0) {
                error ("debug stream broken.");
            }
            safewrite (1, bf, rv); 
        }

        /* check for network input */
        if (FD_ISSET (0, &selmask)) {
            alarm (dbg.tmo);
            switch (GETS (bf)) {
                case -1:
                    error ("net stream broken.");
                    break;
                case NORMAL:
                    safewrite (dbg.gdbin, bf, strlen (bf));
                    break;
                case META:
                    /* need to parse the incoming data for special hacks */
                    switch (bf [1]) {
                        case 'r':
                            write (dbg.gdbin, "run\n", 4);
#if 0
                            if (plen)
                                write (dbg.cgiin, post, plen);
                            close (dbg.cgiin);
#endif                            
                            break;
                        case 'i':
                            fprintf (stderr, "input from browser: %d bytes\n", plen);
                            write (2, post, plen);
                            break;
                        case 'I':
                            if (dbg.pid < 0)
                                break;
                            fprintf (stderr, "\003kill %d\n", dbg.pid);
                            
                            if (kill (dbg.pid, SIGHUP) < 0)
                                perror ("kill: ");
                            else
                                fprintf (stderr, "<int>\n");
                            break;
                        case 'K':
                            if (dbg.pid < 0)
                                break;
                            if (killpg (dbg.dbgpid, SIGINT) < 0)
                                perror ("kill: ");
                            else
                                fprintf (stderr, "<int>\n");
                            
                            break;                            

                        default:
                            fprintf (stderr, "bad meta command '%s'\n", bf);
                            break;
                    }
                    break;
                case SIGKEY:
                    printf ("<int>\n");
                    break;
                default:
                    printf ("unknown gets type.\n");
                    break;
            }
        }
        
    }
    exit (0);
}


