/*
 *                 Copyright (c) May  3 09:35:48  MTCC
 *	Author: Michael Thomas
 *	Module: spamlib.c
 *	Created: Sat May  3 09:35:48 1997
 *	Abstract:
 *	   a httpd based spam buster
 */

/* Edit History: 
 */

#include <html.h>
#include <time.h>
#include "spam.h"

#define DBHOST "fasolt"
#define SPAMDB "spam"
#define PEERTAB "peers"
#define LOGTAB "log"
#define GENERIC "/spam/generic.html"
#define ADMINHTML "/spam/admin.html"
#define PARSEHTML "/spam/parse.html"
#define LISTHTML "/spam/list.html"
#define LOGHTML "/spam/log.html"
#define LOGINHTML "/spam/login.html"
#define PEERLISTHTML "/spam/listpeer.html"

static char * buttons [] = { "nothing", "exact", "substr", "regex", "delete", 0 };


int sock = -1;
hrec * in, * login = 0,  * peers = 0;
int logLevel = 0;



hrec * getdefrec () 
{
    long tm;
    char * t, * env;
    
    hrec * h = newhrec ();
    hradd (h, "user", "");
    hradd (h, "spammer", "");
    hradd (h, "field", "");
    if (env = getenv ("SERVER_NAME"))
        hradd (h, "server", env);
    else
        hradd (h, "server", "");
    hraddint (h, "type", 0);
    tm = time (0); 
    t = ctime (&tm);
    chop (t);
    hradd (h, "cdate", t);
    return h;
}

hrec *
getpeerrec () 
{
    hrec * p = newhrec ();
    hradd (p, "name", "");
    hradd (p, "url", "");
    hraddint (p, "flags", 0);
    return p;    
}

hrec * 
getlogrec () 
{
    hrec * p = newhrec ();
    hraddint (p, "cdate", time (0));
    hradd (p, "reason", "");
    hradd (p, "hdr", "");
    return p;
}



listfull (char * var, char * val, hrec * h) 
{
    char url [1024], * s;
    if (! strcmp (var, "spammer")) {
        strcpy (url, "http://");
        if (s = getenv ("SERVER_NAME")) 
            strcat (url, s);
        if (s = getenv ("SCRIPT_NAME"))
            strcat (url, s);
        strcat (url, "?cmd=parse&user=");
        strcat (url, hrget (in, "user"));
        strcat (url, "&sessid=");
        strcat (url, hrget (in, "sessid"));
        strcat (url, "&hdr=");
        strcat (url, webify (hrget (h, "field")));
        strcat (url, webify (" "));
        strcat (url, webify (val));
        aHTML (url, val);
    } else if (! strcmp (var, "type")) {
        htprintf ("%s", buttons [hrgetint (h, "type")+1]);
    } else
        htprintf ("%s", val);
}

    


listspam (hrec * in) 
{
    char * full = hrisvis (in, "full");
    hrec * h = getdefrec ();
    void * ctx = 0;
    htelm * tbl;
    char * db = hrget (in, "table");

    if (* db == 0)
        db = "local";

    
    if (full == 0) {
        hterase ();
        preTEXT ();
    } else {
        htreadurl (LISTHTML, 0, 0);
        tbl = htgettabledef ("ptable");
        htgotomark ("tbody");
        httablestart (tbl);
    }
    while (hmfetchrows (sock,  h, db, &ctx) > 0) {
        if (full == 0) 
            htprintf ("%s\n", hrget (h, "spammer"));
        else {
            htbindtable (tbl, h, listfull);
        }
    }
    if (full)
        httableend ();
}




mirrorspam (hrec * in) 
{
    int full = hrisvis (in, "full") ? 1 : 0;
    void * ctx;
    hrec * h = getdefrec (), * p;
    int vers;

    hterase ();
    preTEXT ();
    if (hrisvis (in, "version"))
        vers = hrgetint (in, "version");
    else
        vers = MVERSION;
    /* check up on versions */
    if (vers > MVERSION) {
        htprintf ("502 incompatible version ours: %d yours: %d\n",
                  MVERSION, vers);
        
        return;
    }            
    htprintf ("200 Mirror Version %d on %s\n", MVERSION, getenv ("SERVER_NAME"));
    logf (LOG_NORMAL, "200 mirroring to host %s\n", getenv ("REMOTE_HOST"));
    for (p = peers; p; p = p->flink) {
        if ((hrgetint (p, "flags") & PEER_SHADOW) ||
            (full == 0 && strcmp (hrget (p, "name"), "local")))
            continue;
        /* dump the contents of the database */
        ctx = 0;
        while (hmfetchrowsf (sock, h, hrget (p, "name"), &ctx, "order by field,spammer")) {
            /* dump */
            htprintf ("user=%s&", webify (hrget (h, "user")));
            htprintf ("spammer=%s&", webify (hrget (h, "spammer")));
            htprintf ("field=%s&", webify (hrget (h, "field")));
            htprintf ("server=%s&", webify (hrget (h, "server")));
            htprintf ("type=%s&", webify (hrget (h, "type")));
            htprintf ("cdate=%s&", slashify (hrget (h, "cdate")));
            htprintf ("server=%s", slashify (hrget (h, "server")));
            htprintf ("\n");
        }
    }
}


            
            

            





char * exlist [] = { "subject:", "date:", "x-vm-v5-data:",
                     "mime-version:", "content-type:"
                     "content-transfer-encoding:", 
                     0
};

char * findqual = "where field='%s' and spammer='%s'";


hdrspam (hrec * in) 
{
    char * hdr, * next, * field;
    hrec * h, * hvals, * tb;
    int rv, asize = 1024, csize = 0, i;
    void * ctx;
    char * qual = malloc (1024);
    

    hdr = alloca (strlen (hrget (in, "hdr"))+1);    
    h = getdefrec ();
    hvals = newhrec ();
    /* make a private copy */
    strcpy (hdr, hrget (in, "hdr"));
    /* always append the entire header rule */
    strcpy (qual, "where field = '*' or ");
    csize += strlen (qual);
    
    for (; * hdr; hdr = next) {
        
        if (next = strchr (hdr, '\n')) 
            * next++ = 0;
        else
            next = hdr + strlen (hdr);
        /* make certain they didn't send us the entire
           mail message */
        if (* hdr == '\n' || * hdr == '\r' || * hdr == 0)
            break;
        /* strip the field before considering it. */
        if ((field = strchr (hdr, ' ')) ||
            (field = strchr (hdr, '\t'))) {
            int i;

            * field++ = 0;

            strlower (hdr);
            /* never consider these headers */
            for (i = 0; exlist [i]; i++)
                if (! strcmp (hdr, exlist [i]))
                    break;
            if (exlist [i])
                continue;
            /* trim any leading cruft */            
            while (isspace (* field))
                field++;
            /* now, go for any continuation lines */
            while (* next == ' ' || * next == '\t') {
                next [-1] = '\n';
                if (next = strchr (next, '\n')) 
                    * next++ = 0;
                else
                    next += strlen (next);
            }            
        } else {
            /* no field separator, punt */
            continue;
        }
        if (strlen (hdr) + 13 >= asize) {
            asize += 512;
            qual = realloc (qual, asize);
        }
        sprintf (qual+csize, "field='%s' or ", slashify (hdr));
        csize += strlen (qual+csize);
        hrappend (hvals, hdr, field, 0);
    }
    csize -= 3;
    qual [csize] = 0;
    ctx = 0;

    for (tb = peers; tb; tb = tb->flink) {
        if ((hrgetint (tb, "flags") & (PEER_SHADOW|PEER_ENABLED)) == 0)
            continue;
        
        while ((rv =
                hmfetchrowsf (sock, h, hrget (tb, "name"), &ctx, "%s", qual)) > 0) {
            int found = 0;
            /* field for this row */
            field = hrget (h, "field");
            /* this rule to all fields in the hvals */
            if (! strcmp (field, "*")) {
                int i;
                struct h_node * n;
                hrecelm * e;
            
                re_comp (hrget (h, "spammer"));
                for (i = 0; i < hvals->htbl->h_size; i++)
                    for (n = hvals->htbl->h_tab [i]; n; n = n->flink) {
                        e = n->h_obj;
                        if (re_exec (e->val)) {
                            htprintf ("501 spammer because %s@%s sez header contains '%s'\n",
                                      hrget (h, "user"), hrget (tb, "name"),
                                      hrget (h, "spammer"));
                            logf (LOG_NORMAL, "501 spammer because %s@%s sez header contains '%s'\n",
                                      hrget (h, "user"), hrget (tb, "name"),
                                      hrget (h, "spammer"));
                            free (qual);
                            return;
                        }
                    }
                continue;
            }        
            /* corresponding field in header missing? */
            if (hrget (hvals, field) == 0)
                continue;
            logf (LOG_DEBUG, ">>'%s' '%s' '%s'\n", field, hrget (h, "spammer"), hrget (hvals, field));
            switch (hrgetint (h, "type")) {
                case 0:             /* substr */
                    if (strstr (hrget (hvals, field), hrget (h, "spammer")))
                        found++;
                    break;
                case 1:             /* exact */
                    if (! strcmp (hrget (hvals, field), hrget (h, "spammer")))
                        found++;
                    break;
                case 2:             /* regex */
                    re_comp (hrget (h, "spammer"));
                    if (re_exec (hrget (hvals, field)))
                        found++;
                    break;
            }
            if (found) {
                htprintf ("500 spammer because %s@%s sez field '%s' contains '%s'\n",
                          hrget (h, "user"), hrget (tb, "name"),
                          hrget (h, "field"), hrget (h, "spammer"));
                logf (LOG_NORMAL, "500 spammer because %s@%s sez field '%s' contains '%s'\n",
                          hrget (h, "user"), hrget (tb, "name"),
                          hrget (h, "field"), hrget (h, "spammer"));
                free (qual);
                return;
            }               
        }
    }
    
    if (rv < 0) {
        logf (LOG_NORMAL, "201 msqlerr: %s\n", msqlErrMsg);
        htprintf ("201 msqlerr: %s\n", msqlErrMsg);
    }
    free (qual);
    logf (LOG_TRACK, "100 Ok.");
    htprintf ("100 Ok.\n");
}




loginspam (hrec * in) 
{
    char pwd [50];
    char * u = hrget (in, "user");

    
    
    if (hpwdfile ("/etc/passwd", u, pwd) < 0)
        error ("no such user %s", u);
    if ((login = hlogin (u, hrget (in, "pwd"), pwd)) == 0)
        error ("login failed.");
    setadmin (in);
}

cklogin (hrec * in) 
{
    if ((login = hvrfysession (in, 0)) == 0) 
        error ("login information invalid.");
}


startlogin (hrec * in) 
{
    htreadurl (LOGINHTML, 0, 0);
}


setadmin (hrec * in) 
{
    
    htelm * h = htreadurl (ADMINHTML, 0, 0);
    hrec * p;

    htrewind ();
    if (h == 0)
        error ("can't open template.");    
    /* bind the username and sessid to all the forms */
    if (login == 0) {
        login = newhrec ();
        hradd (login, "user", hrget (in, "user"));
        hradd (login, "sessid", hrget (in, "sessid"));
    }        
    while (htformnext ()) {
        htbindform (login);
    }
    /* add the peers to the list peers */
    for (p = peers; p; p = p->flink) {
        char * name = hrget (p, "name");
        htgotomark ("listrules");
        optionHTML (name, name, ! strcmp (name, "local"));
        if ((hrgetint (p, "flags") & PEER_SHADOW) == 0) {
            htgotomark ("peermisc");
            optionHTML (name, name, 0);
        }
    }
    /* if there is a name, look it up */
    if (p = lookpeer (hrget (in, "name"))) {
        int flg = hrgetint (p, "flags");
        htgotoform ("addpeer");
        if (flg & PEER_ENABLED)
            hradd (p, "enabled", "enabled");
        else
            hradd (p, "enabled", "disabled");
        if (flg & PEER_AUTOUPD)
            hradd (p, "auto", "auto");
        else
            hradd (p, "auto", "defer");
        if (flg & PEER_ALLRULES)
            hradd (p, "rules", "all");
        else
            hradd (p, "rules", "local");
        htbindform (p);
    }
    p = newhrec ();
    hraddint (p, "level", logLevel);
    htgotoform ("setlog");
    htbindform (p);
    hrfree (p);
}





listhdr (char * var, char * val, hrec * h) 
{
    int i = 0;
    char * p;
    
    if (! strcmp (var, "label")) {
        htprintf ("%s", val);
    } else if (! strcmp (var, "field")) {
        for (i = 0, p = hrget (h, "fval"); p; i++)
            if (p = strchr (p, '\n'))
                p++;
        if (i > 1)
            textHTML ("", val, hrget (h, "fval"), 40, i);
        else
            inputHTML ("", val, hrget (h, "fval"), input, 40, 512);
    } else if (! strcmp (var, "action")) {
        selectHTML ("", val, buttons, hrget (h, "button"));
    }
    return 1;            
}

        
    

parsespam (hrec * in) 
{
    char * hdr, * next, * field;
    int i;
    hrec * h, * db;
    htelm * tbl;
    void * ctx;

    /* this always uses the local table. */
    hputenv ("edittype", "normal");
    if (hrisvis (in, "check")) {
        htreadurl (GENERIC, 0, 0);
        htgotomark ("pageheader");
        hdrspam (in);
        return;
    }
    if (htreadurl (PARSEHTML, 0, 0) == 0)
        error ("can't read template.");
    htgotoform ("pform");
    h = newhrec ();
    db = getdefrec ();
    hradd (h, "user", hrget (in, "user"));
    hradd (h, "sessid", hrget (in, "sessid"));
    
    htbindform (h);
    hrfree (h);
    /* make a private copy */
    hdr = alloca (strlen (hrget (in, "hdr"))+1);    
    strcpy (hdr, hrget (in, "hdr"));
    tbl = htgettabledef ("ptable");
    h = newhrec ();
    htgotomark ("tbody");
    inputHTML ("", "table", "local", hidden, 0, 0);
    httablestart (tbl);    
    for (i = 0; * hdr; hdr = next, i++) {
        if (next = strchr (hdr, '\n')) 
            * next++ = 0;
        else
            next = hdr + strlen (hdr);
        /* make certain they didn't send us the entire
           mail message */
        if (* hdr == '\n' || * hdr == '\r' || * hdr == 0)
            break;
        /* strip the field before considering it. */
        if ((field = strchr (hdr, ' ')) ||
            (field = strchr (hdr, '\t'))) {
            int i;

            * field++ = 0;

            strlower (hdr);
            /* never consider these headers */
            for (i = 0; exlist [i]; i++)
                if (! strcmp (hdr, exlist [i]))
                    break;
            if (exlist [i])
                continue;
            /* trim any leading cruft */            
            while (isspace (* field))
                field++;
            /* now, go for any continuation lines */
            while (* next == ' ' || * next == '\t') {
                next [-1] = '\n';
                if (next = strchr (next, '\n')) 
                    * next++ = 0;
                else
                    next += strlen (next);
            }
            
        } else {
            /* no field separator, punt */
            continue;
        }
        hradd (h, "label", hdr);
        hraddf (h, "field", "f%d:%s", i, hdr);
        hraddf (h, "action", "a%d:%s", i, hdr);
        hradd (h, "fval", field);
        /* check to see if it's already there, and select
           the proper button if so. */
        ctx = 0;
        if (hmfetchrowsf (sock, db, "local", &ctx, findqual, hdr, field) > 0)
            hradd (h, "button", buttons [hrgetint (db, "type")+1]);
        else
            hradd (h, "button", buttons [0]);        
        htbindtable (tbl, h, listhdr);
        
    }
    httableend ();
    hrfree (h);    
}


editspam (hrec * in) 
{
    hrec * row = getdefrec (), * h;
    void * ctx = 0;
    char * db = hrget (in, "table"), * edittb;
    int i;
    htelm * tbl;

    if (* db == 0)
        db = "local";
    edittb = db;
    h = lookpeer (db);
    if (h == 0)
        error ("can't find peer %s's database", db);
    if (hrgetint (h, "flags") & PEER_SHADOW) {
        hputenv ("edittype", "shadow");
        /* the base table is kept in url */
        edittb = hrget (h, "url");
    }
    else
        hputenv ("edittype", "normal");
    
    htreadurl (PARSEHTML, 0, 0);
    htgotoform ("pform");
    h = newhrec ();
    hradd (h, "user", hrget (in, "user"));
    hradd (h, "sessid", hrget (in, "sessid"));
    
    htbindform (h);
    hrfree (h);
    h = newhrec ();
    tbl = htgettabledef ("ptable");
    htgotomark ("tbody");
    inputHTML ("", "table", edittb, hidden, 0, 0);    
    httablestart (tbl);
    i = -1;    
    while (hmfetchrowsf (sock, row, db, &ctx, "order by field,spammer") > 0) {
        i++;        
        hradd (h, "label", hrget (row, "field"));
        hraddf (h, "field", "f%d:%s", i, hrget (row, "field"));
        hraddf (h, "action", "a%d:%s", i, hrget (row, "field"));
        hradd (h, "fval", hrget (row, "spammer"));
        /* check to see if it's already there, and select
           the proper button if so. */
        hradd (h, "button", buttons [hrgetint (row, "type")+1]);
        htbindtable (tbl, h, listhdr);
    }
    httableend ();
    hrfree (h);    
}



addspam (hrec * in) 
{
    char hdr [256], bf [256], * p;
    hrec * u = getdefrec (); 
    int acnt = 0, dcnt = 0, i, rv;
    struct h_node * n;
    char * db = hrget (in, "table");
    int merge = 0;

    if (* db == 0)
        db = "local";
    htreadurl (GENERIC, 0, 0);
    htgotomark ("pagebody");

    if (! strcmp (hrget (in, "go"), "merge")) {
        if (! strcmp (db, "local"))
            error ("can't merge the local rules");
        /* otherwise, zot the current copy */
        hmdeletef (sock, db, "");
        merge = 1;
    }
    for (i = 0; i < in->htbl->h_size; i++) 
        for (n = in->htbl->h_tab [i]; n; n = n->flink) {
            if (sscanf (n->h_name, "f%*d:%s", hdr)) {
                hrecelm * e = n->h_obj;
                int type = 0;
                
                /* grab the associated action */
                strcpy (bf, n->h_name);
                bf [0] = 'a';
                p = hrget (in, bf);
                /* find which was selected */
                switch (* p) {
                    case 'n':
                        break;
                    case 'r':
                        type++;
                    case 's':
                        type++;
                    case 'e':                        
                    case 'd':
                        /* no difference for now... */
                        hradd (u, "user", hrget (in, "user"));
                        hradd (u, "field", hdr);
                        hradd (u, "spammer", e->val);
                        hraddint (u, "type", type);
                        if (* p == 'd') {
                            if (merge)
                                break;
                            rv = hmdeletef (sock, db, findqual,
                                       hdr, slashify (e->val));
                            if (rv > 0)
                                dcnt += rv;
                        }
                        else {
                            rv = hmupdatef (sock, u, db, findqual,
                                       hdr, slashify (e->val));
                            if (rv > 0)
                                acnt += rv;
                        }
                        if (rv < 0) {
                            htprintf ("sql error: %s", msqlErrMsg);
                            logf (LOG_NORMAL, "sql error: %s", msqlErrMsg);
                            br (1);
                        }
                        break;
                }
            }
        }
    htgotomark ("pageheader");
    htprintf ("Rules: %d updated, %d deleted.", acnt, dcnt);
    logf (LOG_NORMAL, "%s's rules updated by %s", db, hrget (in, "user"));
}


/* peering commands */





#define TDEF "create table %s (user char(50), spammer char(100), field char(30),\
cdate char(30), type int, server char(100))"
#define IDEF "create index i1 on %s (field)"

addpeer (hrec * in) 
{
    hrec * peer;
    int flags = 0, new = 0;
    char * name = hrget (in, "name");
    char upd [256];
    int rv;    

    if (hrisvis (in, "del"))
        delpeer (in);
    peer = getpeerrec ();
    hradd (peer, "name", name);
    sprintf (upd, "%s_update", name);
    /* create the table and index */
    if (lookpeer (hrget (in, "name")) == 0) {
        
        rv = msqlQueryf (sock, TDEF, name);
        if (rv < 0)
            error ("sql error: %s", msqlErrMsg);
        rv = msqlQueryf (sock, IDEF, name);
        if (rv < 0) {
            delpeer (in);
            error ("sql error: %s", msqlErrMsg);
        }
        rv = msqlQueryf (sock, TDEF, upd);
        if (rv < 0) {
            delpeer (in);
            error ("sql error: %s", msqlErrMsg);
        }
        new = 1;
    }
    /* create the peer record */
    if (! strcmp (hrget (in, "enabled"), "enabled")) 
        flags |= PEER_ENABLED;        
    if (! strcmp (hrget (in, "auto"), "auto")) 
        flags |= PEER_AUTOUPD;
    if (! strcmp (hrget (in, "rules"), "all"))
        flags |= PEER_ALLRULES;
    hraddint (peer, "flags", flags);
    hradd (peer, "url", hrget (in, "url"));    
    if (hmupdate (sock, peer, PEERTAB, "name", name) < 0) {
        if (new)
            delpeer (in);
        logf (LOG_NORMAL, "sql error: %s", msqlErrMsg);
        error ("sql error: %s", msqlErrMsg);
    }
    /* and now its update table */
    hradd (peer, "name", upd);
    hradd (peer, "url", name);
    hraddint (peer, "flags", PEER_SHADOW);

    if (hmupdate (sock, peer, PEERTAB, "name", upd) < 0) {
        if (new)
            delpeer (in);
        logf (LOG_NORMAL, "sql error: %s", msqlErrMsg);
        error ("sql error: %s", msqlErrMsg);
    }
    htreadurl (GENERIC, 0, 0);    
    htgotomark ("pageheader");
    htprintf ("%s has been installed as a spam peer", name);
    logf (LOG_NORMAL, "peer %s installed by %s", name, hrget (in, "user"));
}

           
    
    
    



delpeer (hrec * in) 
{
    char * name = hrget (in, "name");
    char upd [256];
    htreadurl (GENERIC, 0, 0);
    if (* name == 0)
        error ("you must supply a name");
    if (! strcmp (name, "local"))
        error ("can't delete the local database.");
    msqlQueryf (sock, "drop table %s", name);
    sprintf (upd, "%s_update", name);
    msqlQueryf (sock, "drop table %s", upd);
    hmdelete (sock, PEERTAB, "name", name);
    hmdelete (sock, PEERTAB, "name", upd);
    htgotomark ("pageheader");
    htprintf ("%s has been deleted as a spam peer", name);
    logf (LOG_NORMAL, "peer %s deleted by %s", name, hrget (in, "user"));    
}

    




listpeer (hrec * in) 
{
    hrec * p;
    htelm * tb;
    int flg;
    
    if (htreadurl (PEERLISTHTML, 0, 0) == 0) 
        error ("can't read template");

    tb = htgettabledef ("ptable");
    htgotomark ("tbody");
    httablestart (tb);
    
    for (p = peers; p; p = p->flink) {
        char str [100];
        
        flg = hrgetint (p, "flags");
        if (flg & PEER_SHADOW)
            continue;
        if (flg & PEER_ENABLED)
            strcpy (str, "enabled,");
        else
            strcpy (str, "disabled,");
        if (flg & PEER_AUTOUPD)
            strcat (str, "Updates: Auto,");
        else
            strcat (str, "Updates: Manual,");
        if (flg & PEER_ALLRULES)
            strcat (str, "Rules: All");
        else
            strcat (str, "Rules: Local Only");
        hradd (p, "strflags", str);
        
        htbindtable (tb, p, 0);
    }
    httableend ();
}

findpeer (hrec * in) 
{
    setadmin (in);
}

    


updpeer (hrec * in) 
{
    char bf [4096], tab [256];
    FILE * fp;
    char * name;
    int resp;
    hrec * h, * peer;
    int tot = 0, rv;

    name = hrget (in, "name");    
    if ((peer = lookpeer (name)) == 0)
        error ("invalid peer '%s'", name);
    sprintf (bf, "%s?cmd=mirror&version=%d%s",
             hrget (peer, "url"), MVERSION, 
             hrgetint (peer, "flags") & PEER_ALLRULES ? "&full" : "");
    if ((fp = fopenurl (bf, 0, 0)) == 0)
        error ("can't open spam server at %s\n", name);
    fgets (bf, 4096, fp);
    if (sscanf (bf, "%d", &resp) == 0)
        error ("invalid response: %s", bf);
    if (resp >= 300)
        error ("mirror error: %s", bf);
    /* we treat the entire thing as a post which we feed selectively
       to the post decombobulator */

    /* since we know that a real post can't have an
       unescaped newline, we'll take advantage of that
       so that it's our record delimiter. For the rest,
       we just treat each record as an individual post
       and suck it into an hrec. Clever, no?
     */
    sprintf (tab, "%s_update", name);
    h = getdefrec ();        
    while (fgets (bf, 4096, fp) != NULL) {
        chop (bf);
        hrgetorpost (h, bf, strlen (bf));
        
        rv = hmupdate (sock, h, tab, "field", hrget (h, "field"));
        if (rv < 0) {
            logf (LOG_NORMAL, "sql error: %s", msqlErrMsg);
            error ("sql error: %s", msqlErrMsg);
        }
        tot += rv;
    }
    if (hrisvis (in, "text") == 0) {
        htreadurl (GENERIC, 0, 0);
        htgotomark ("pageheader");
        htprintf ("%d records updated for %s", tot, name);
    } else
        printf ("%d records updated for %s\n", tot, name);
    
    logf (LOG_NORMAL, "peer %s updated by %s", name, hrget (in, "user"));    
    fclose (fp);
    hrfree (h);
}

    



miscpeer (hrec * in) 
{
    if (hrisvis (in, "del"))
        delpeer (in);
    else if (hrisvis (in, "list"))
        listpeer (in);
    else if (hrisvis (in, "add"))
        addpeer (in);
    else if (hrisvis (in, "update"))
        updpeer (in);
    else if (hrisvis (in, "find"))
        findpeer (in);
}


showlog (hrec * in) 
{
    hrec * h = getlogrec ();
    void * ctx = 0;
    htelm * tbl;

   
    htreadurl (LOGHTML, 0, 0);
    if (! strcmp (hrget (in, "full"), "full"))
        tbl = htgettabledef ("ptablefull");        
    else    
        tbl = htgettabledef ("ptable");
    htgotomark ("tbody");
    httablestart (tbl);
    while (hmfetchrowsf (sock,  h, LOGTAB, &ctx, "order by cdate desc") > 0) {
        long t = hrgetint (h, "cdate");
        hradd (h, "date", ctime (&t));
        hrclrflg (h, "date", HR_HIDDEN);            
        htbindtable (tbl, h, 0);
        hrsetflg (h, "date", HR_HIDDEN);                    
    }
    httableend ();
}


purgelog (hrec * in) 
{
    int since = hrgetint (in, "since");
    int tm = time (0), rv;
    
    if (since > 0) {
        tm -= since * 24 * 60 * 60;
        rv = hmdeletef (sock, LOGTAB, "where cdate < %d", tm);
    } else
        rv = hmdeletef (sock, LOGTAB, "");
    if (rv < 0) {
        logf (LOG_NORMAL, "sql error: %s", msqlErrMsg);
        error ("sql error:", msqlErrMsg);
    }
    if (hrisvis (in, "text") == 0) {
        htreadurl (GENERIC, 0, 0);
        htgotomark ("pageheader");
        htprintf ("%d records purged.", rv);        
    } else
        printf ("%d records purged.\n", rv);
}


setlog (hrec * in) 
{
    hrec * p = lookpeer ("local");
    unsigned long lvl = hrgetint (in, "level");
    unsigned long old = hrgetint (p, "flags");

    if (lvl < 0 || lvl > LOG_DEBUG)
        error ("invalid logging level");
    old &= ~ (LOG_DEBUG << LOG_SHIFT);
    old |= lvl << LOG_SHIFT;
    logLevel = lvl;
    hraddint (p, "flags", old);
    if (hmupdate (sock, p, PEERTAB, "name", "local") < 0) {
        logf (LOG_NORMAL, "msql error: %s\n", msqlErrMsg);
        error ("msql errorL %s", msqlErrMsg);
    }
    setadmin (in);
}



/*
 * read the peer database
 */


readpeers () 
{
    hrec * p;
    void * ctx;

    p = getpeerrec ();    
    ctx = 0;    
    while (hmfetchrowsf (sock, p, PEERTAB, &ctx, "") > 0) {
        /* if this is the local database, grab the logging level */
        if (! strcmp (hrget (p, "name"), "local"))
            logLevel = (unsigned long) (hrgetint (p, "flags")) >> LOG_SHIFT;
        hrlink (&peers, p);
        p = getpeerrec ();
    }
    hrfree (p);
}

/*
 * lookup a a peer's name
 */

hrec *
lookpeer (char * nm) 
{
    hrec * p;
    
    for (p = peers; p; p = p->flink)
        if (! strcmp (nm, hrget (p, "name")))
            return p;
    return 0;
}


void 
error (char * fmt, ...) 
{
    va_list ap;
    char bf [512];
    htelm * h;

    va_start (ap, fmt);
    vsprintf (bf, fmt, ap);
    va_end (ap);

    if (hrisvis (in, "text") == 0) {        
        hterase ();
        preHTML ();
        if ((h = htreadurl (GENERIC, 0, 0)) == 0)
            startHTML ("spam error", "/");
        else {
            htgotomark ("title");
            htprintf ("spam error");
        }
        htgotomark ("pageheader");
        par ();
        htprintf ("an error has occurred: %s\n", bf);
        if (h == 0)
            endHTML ();
        else
            htdumpfile ();
    } else
        fprintf (stderr, "spamerr: %s\n", bf);    
    exit (0);
}

void logf (int lvl, char * fmt, ...) 
{
    va_list ap;
    char bf [1024];
    hrec * h;

    if (lvl > logLevel)
        return;
    
    h = getlogrec ();
    va_start (ap, fmt);
    vsnprintf (bf, 127, fmt, ap);
    va_end (ap);
    hradd (h, "reason", bf);
    /* append headers if any */
    strncpy (bf, hrget (in, "hdr"), 512);
    bf [511] = 0;
    hradd (h, "hdr", bf);
    if (hminsert (sock, h, LOGTAB) < 0)
        error ("error writing log: %s", msqlErrMsg);
    
}

