#include <sys/file.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <html.h>

#ifdef DEBUG
#define PWDFILE "/user/mike/passwdhack"
#else
#define PWDFILE "/etc/passwd."
#endif
#define PWD     "/user/mike/public_html/pwd/pwd.html"
#define PWDFIND "/user/mike/public_html/pwd/pwdfind.html"
#define PWDLIST "/user/mike/public_html/pwd/pwdlist.html"
#define LOCKFILE "/user/mike/public_html/pwd/pwd.lock"
#define GENERIC "/user/mike/public_html/pwd/generic.html"
#define BASEURL "http://www.mtcc.com/~mike/pwd/"
#define PASSSEED "pw"
#define UPROG "/usr/sbin/userman"


hrec * hrecs = 0, * roadmap;
int nrecs;
int uidmax = 0;
void error(char * fmt, ...);

/*
 * the default record is sort of important. It basically
 * maps out types, and what will be fetched from various
 * database programs, etc.
 */

hrec * getdefrec () 
{
    hrec * p = newhrec ();

    hradd (p, "user", "");
    hradd (p, "pwd", "");
    hrtypeadd (p, "uid", "-1", HRF_INT, 0, 0);
    hradd (p, "gid", "-1");
    hradd (p, "fullname", "");
    hradd (p, "disabled", "no");
    hradd (p, "home", "/dev/null");
    hradd (p, "shell", "/bin/false");
    return p;
    
}



/*
 * Called pwd <db> <cmd>
 */

main (int ac, char ** av) 
{
    FILE * lock, * ap;
    hrec * params;
    char * cmd;    

    preHTML ();   
    lock = fopen (LOCKFILE, "r");

    if ((roadmap = hrreadfile ("/etc/roadmap.conf")) == 0)
        error ("can't read roadmap.conf");
    /* setup the default uid */
    sscanf (hrget (roadmap, "uid"), "%d", &uidmax);
    
    /* Be paranoid: grab an exclusive lock for all reasons. */
    flock (fileno (lock), LOCK_EX);
    nrecs = readdb ();
    params = hrgetparams (getdefrec (), ac, av);
    if (* (cmd = hrget (params, "cmd")) == 0) {
        if (ac < 2)
            error ("not enough arguments");
        cmd = av [1];
    }
    if (hinitlogin ("fasolt", "pwd", "zz", 60*60) < 0)
        error ("can't start up login database.");
    
    switch (* cmd) {
        case 'u':
            login (params);
            break;
        case 'f':              /* find */
            find (params);
            break;
        case 'l':              /* list */
            list (params);
            break;
        case 'a':              /* add */
            add (params);            
            break;
        case 'r':              /* remove */
            rem (params);
            break;
        default:
            error ("unknown command %s", cmd);
            break;
    }
    htdumpfile ();
    exit (0);
}

login (hrec * in) 
{
    char pwd [50];
    char * u = hrget (in, "user");
    hrec * login;
    htelm * h = htreadfile (PWD);
    
    if (h == 0)
        error ("can't open template.");
    
    if (hpwdfile ("/etc/passwd", u, pwd) < 0)
        error ("no such user %s", u);
    if ((login = hlogin (u, hrget (in, "pwd"), pwd)) == 0)
        error ("login failed.");
    hradd (in, "sessid", hrget (login, "sessid"));
    hloginbindall (in);
    
}

    





find (hrec * in)
{
    hrec * r;
    htelm * h = htreadfile (PWDFIND);
    char * pass;
    int gid;
    struct group *gr;

    if (h == 0)
        error ("can't read template file");

    
    if (hvrfysession (in, 0) == 0)
        error ("login error -- try logging in again.");
    
    if ((r = hrfind (hrecs, "user", hrget (in, "user"))) == 0) {
        r = getdefrec ();
        /* this could be done with an hrmerge for the most
           part, but it's have all kinds of other cruft eventually,
           so lets not burden it for now.
         */
        hradd (r, "user", hrget (in, "user"));
        hradd (r, "fullname", hrget (in, "user"));
        hraddf (r, "home", "%s/%s", hrget (roadmap, "home"),
                hrget (in, "user"));
        hradd (r, "shell", hrget (roadmap, "shell"));
        hraddf (r, "gid", "%d", grtol (hrget (roadmap, "gid")));
        /* this should probably be done a bit more intelligently */
        hraddf (r, "uid", "%d", uidmax+1);
        hradd (r, "disabled", "yes");
    }
    /* propogate the session id along the way */
    hradd (r, "sessid", hrget (in, "sessid"));
    hrsetflg (r, "pwd", HR_HIDDEN);
    htgotomark ("add");
    htprintf ("user: %s", hrget (r, "user"));
    
    htformnext ();
    htbindform (r);
    htgotomark ("group");
    setgrent ();
    if (* hrget (r, "gid"))
        gid = atol ( hrget (r, "gid"));
    else
        gid = 0;                /* s/b default gid  */
    while (gr = getgrent ()) 
        optionHTML (0, gr->gr_name, gid == gr->gr_gid ? 1 : 0);
    
    htgotomark ("rem");
    htformnext ();
    htbindform (r);
}

add (hrec * in)
{
    htelm * h = htreadfile (GENERIC);
    hrec * old;
    hrecerr err;
    

    if (h == 0)
        error ("couldn't read template.");
    if (hvrfysession (in, 0) == 0)
        error ("login error -- try logging in again.");
    if (hrvrfy (in, &err) < 0) {
        error ("error at field %s: %s", err.name, err.errmsg);
    }
    if (* hrget (in, "user") == 0)
        error ("You must supply an user name to add");
    
    if ((old = hrfind (hrecs, "user", hrget (in, "user"))) == 0) {
        /* use the incoming record as a base */
        old = in;
        if (strcmp (hrget (in, "pwd"), hrget (in, "pwd2")))
            error ("password's don't match");        
        if (* hrget (in, "pwd"))
            hradd (old, "pwd", crypt (hrget (in, "pwd"), PASSSEED));
        else
            hradd (old, "pwd", "*");
        hrlink (&hrecs, old);        
    } else {
        /* merge everything into old */
        hrsetflg (old, "pwd", HR_NOMERGE);
        hrsetflg (old, "pwd2", HR_NOMERGE);                
        hrmerge (old, in);
        if (strcmp (hrget (in, "pwd"), hrget (in, "pwd2")))
            error ("password's don't match");
        if (* hrget (in, "pwd"))
            hradd (old, "pwd", crypt (hrget (in, "pwd"), PASSSEED));
    }
    hraddf (old, "gid", "%d", grtol (hrget (old, "gid")));
    
    hrsetflg (old, "pwd2", HR_HIDDEN);
/*    writedb ();*/
/*    mkuser (old);*/
    nrecs++;
    htgotomark ("pageheader");    
    htprintf ("%s (%s) has been added.\n", hrget (old, "user"),
            hrget (old, "fullname"));
    htgotomark ("pagebody");
}




rem (hrec * in) 
{
    hrec * old;
    htelm * h = htreadfile (GENERIC);

    if (h == 0)
        error ("can't read template file.");
    if (hvrfysession (in, 0) == 0)
        error ("login error -- try logging in again.");
    
    if (* hrget (in, "user") == 0)
        error ("You must supply a user");
    if (old = hrfind (hrecs, "user", hrget (in, "user"))) {
        hrdelink (&hrecs, old);
        htgotomark ("pageheader");
        htprintf ("user %s has been deleted.\n", hrget (in, "user"));
        rmuser (old);
        return;            
    }
    htgotomark ("pageheader");    
    htprintf ("%s isn't a user\n");
}



int ucmp (const void * k1, const void * k2) 
{
    char * p1, * p2, * p;
    int rv;

    p1 = hrget (* (hrec **) k1, "user");    
    if (p = strrchr (p1, ' '))
        p1 = p+1;
    p2 = hrget (* (hrec **) k2, "user");    
    if (p = strrchr (p2, ' '))
        p2 = p+1;
    return strcasecmp (p1, p2);
}




listcell (char * var, char * val, hrec * r) 
{    
    char url [256], * s;
    int t;

    if (! strcmp (var, "user")) {
        strcpy (url, "http://");
        if (s = getenv ("SERVER_NAME")) 
            strcat (url, s);
        if (s = getenv ("SCRIPT_NAME"))
            strcat (url, s);
        strcat (url, "?cmd=find&user=");
        strcat (url, val);
        aHTML (url, val);
    } else if (! strcmp (var, "pwd") && ! strcmp (val, "*"))
        htprintf (" (login disabled account)");
    else 
        htprintf ("%s", val);
    return 1;
}




list (hrec * in) 
{
    hrec * p;
    hrec ** vec;
    htelm * h = htreadfile (PWDLIST);
    htelm * tbl;
    int cnt = 0, j;
    char host [100];

    if (h == 0)
        error ("can't open template.");
    if (hvrfysession (in, 0) == 0)
        error ("login error -- try logging in again.");
    vec = alloca (nrecs * sizeof (vec));
    for (p = hrecs; p; p = p->flink) {
        vec [cnt] = p;
        cnt++;
    }    

    htgotomark ("pageheader");
    gethostname (host, 99);
    
    htprintf ("Users on %s", host);
    
    htgotomark ("pagebody");
    qsort (vec, cnt, sizeof (vec), ucmp);
    center (1);
    tbl = htgettabledef ("list");
    httablestart (tbl);
    for (j = 0; j < cnt; j++) {
        p = vec [j];
        htbindtable (tbl, p, listcell);
    }
    httableend ();
    center (0);
}




writedb () 
{
    FILE * fp;
    hrec * h;
    

    fp = fopen (PWDFILE, "w");
    if (fp == 0)
        error ("can't write %s\n", PWDFILE);
    
    for (h = hrecs; h; h = h->flink)
        fprintf (fp, "%s:%s:%s:%s:%s:%s:%s\n",
                 hrget (h, "user"),
                 hrget (h, "pwd"),
                 hrget (h, "uid"),
                 hrget (h, "gid"),
                 hrget (h, "fullname"),
                 hrget (h, "home"),
                 hrget (h, "shell"));
    fclose (fp);
}

/*
 * make the user's directory's etc
 */

mkuser (hrec * h) 
{
    char bf [256];
    sprintf (bf, "%s create %s %s", UPROG, hrget (h, "user"),
             hrget (h, "home"));
    system (bf);
    /* need to also set the password using passwd -p */
}

/*
 * remove a user from the system
 */

rmuser (hrec * h) 
{
    char bf [256];
    sprintf (bf, "%s delete %s %s", UPROG, hrget (h, "user"),
             hrget (h, "home"));
    system (bf);
}



readdb () 
{
    hrec * p;
    struct passwd * pw;
    
    while (pw = getpwent ()) {
        p = getdefrec ();
        hradd (p, "user", pw->pw_name);
        hradd (p, "pwd", pw->pw_passwd);
        hraddf (p, "uid", "%d", pw->pw_uid);
        hraddf (p, "gid", "%d", pw->pw_gid);
        hradd (p, "fullname", pw->pw_gecos);
        if (! strcmp (pw->pw_passwd, "*"))
            hradd (p, "disabled", "yes");
        else
            hradd (p, "disabled", "no");
        hradd (p, "home", pw->pw_dir);
        hradd (p, "shell", pw->pw_shell);
        hrlink (&hrecs, p);
        if (pw->pw_uid > uidmax)
            uidmax = pw->pw_uid;
        nrecs++;
    }
    endpwent ();    
    return nrecs;
}

void error (char * fmt, ...);


void 
error (char * fmt, ...) 
{
    va_list ap;
    char bf [512];
    htelm * h;

    va_start (ap, fmt);
    vsprintf (bf, fmt, ap);
    va_end (ap);
    
    hterase ();
    preHTML ();
    if ((h = htreadfile (GENERIC)) == 0)
        startHTML ("pwd error", BASEURL);
    else {
        htgotomark ("title");
        htprintf ("pwd error");
    }
    htgotomark ("pageheader");
    par ();
    htprintf ("an error has occurred: %s\n", bf);
    if (h == 0)
        endHTML ();
    else
        htdumpfile ();    
    exit (0);
}

/* convert a group to an int */

grtol (char * grp) 
{
    struct group * g;
    setgrent ();
    while (g = getgrent ())
        if (! strcmp (g->gr_name, grp)) {
            setgrent ();
            return g->gr_gid;
        }    
    setgrent ();
    return 0;
}
