/*********************************************************************\
* 								      *
* 		    Copyright (c) 1986 Gordian Inc.		      *
* 								      *
* 	Author: Michael A Thomas				      *
* 	Module: hash.c						      *
* 	Created: Tue Sep 23 15:19:51 1986			      *
* 	Abstract:						      *
* 		Hash table primitives, h_get, h_put, h_del, h_init    *
* 								      *
\*********************************************************************/
/* Edit History: 
 */

#include "config.h"
#include "hash.h"

static inline int hash (), hashi ();


/***************************************************************************\
* 									    *
* Function: h_init							    *
* Parameters: size							    *
* Synopsis:								    *
* int size; 								    *
* Abstract:								    *
* 	Allocate the space of, and initialize a hash table. Returns the     *
*         address of the table to be used with other functions.		    *
* 									    *
\***************************************************************************/


struct h_table * h_init(int size)
{
   register struct h_table * rv = (struct h_table *) calloc (sizeof (* rv), 1);
   rv->h_size = size;
   rv->h_tab = (struct h_node **) calloc (sizeof (void *), size);
   return rv;
}


void h_destroy (struct h_table *h)
{
    struct h_node * p;
    int i;
    
    for (i = 0; i < h->h_size; i++) 
        while ((p = h->h_tab [i])) {
            h->h_tab [i]  = h->h_tab [i]->flink;
            free (p->h_name);
            free (p);
        }
    free (h->h_tab);
    free (h);
}

void h_destroyi (struct h_table *h)
{
    struct h_node * p;
    int i;
    
    for (i = 0; i < h->h_size; i++) 
        while ((p = h->h_tab [i])) {
            h->h_tab [i]  = h->h_tab [i]->flink;
            free (p);
        }
    free (h->h_tab);
    free (h);
}


void * h_get (struct h_table *tb, const char *s)
{
   int idx = hash (tb, s);		/* hash the string */
   register struct h_node * p;
   tb->hits++;
   tb->depth++;
   /* traverse the bucket */   
   for (p = tb->h_tab [idx]; p; p = p->flink, tb->depth++)
	if (! strcmp (p->h_name, s))
            return p->h_obj;
   return 0;
}

void * h_geti (struct h_table *tb, long s)
{
   int idx = hashi (tb, s);		/* hash the string */
   register struct h_node * p;
   tb->hits++;
   tb->depth++;
   /* traverse the bucket */
   for (p = tb->h_tab [idx]; p; p = p->flink, tb->depth++)	
	if (p->h_name == (void *) s)
            return p->h_obj;
   return 0;
}


void * h_first (struct h_table * tb) 
{
    tb->seaidx = 0;
    tb->seanode = 0;
    return h_next (tb);
}

void * h_next (struct h_table * tb) 
{
    struct h_node * rv;
    if (tb->seanode != 0) {
        rv = tb->seanode;
        tb->seanode = tb->seanode->flink;
        return rv->h_obj;
    }
    for (; tb->seaidx < tb->h_size; tb->seaidx++)
        if ((rv = tb->h_tab [tb->seaidx]) != 0) {
            tb->seanode = tb->h_tab [tb->seaidx]->flink;
            tb->seaidx++;
            return rv->h_obj;
        }
    return 0;
}
    


/*************************************************************************\
* 									  *
* Function: h_put							  *
* Parameters: tb, s, o							  *
* Synopsis:								  *
* struct h_table * tb;							  *
* register char * s;							  *
* long o; 								  *
* Abstract:								  *
* 	put an object (o) into the hash table (tb) to be referenced by	  *
*         string s.							  *
* 									  *
\*************************************************************************/


struct h_node * h_put (struct h_table *tb, const char *s, void * o)
{
   int idx = hash (tb, s);
   register struct h_node * p = (struct h_node *) malloc (sizeof (* p));
   p->flink = tb->h_tab [idx];		/* start linking it in */
   p->h_obj = o;			/* the object */
   p->h_name = strdup (s);
   tb->h_tab [idx] = p;			/* done! */
   return p;
}


struct h_node * h_puti (struct h_table *tb, long s, void * o)
{
   int idx = hashi (tb, s);
   register struct h_node * p = (struct h_node *) malloc (sizeof (* p));
   p->flink = tb->h_tab [idx];		/* start linking it in */
   p->h_obj = o;			/* the object */
   p->h_name = (void *) s;
   tb->h_tab [idx] = p;			/* done! */
   return p;
}


/**************************************************************************\
* 									   *
* Function: h_del							   *
* Parameters: tb, s							   *
* Synopsis:								   *
* struct h_table * tb;							   *
* register char * s; 							   *
* Abstract:								   *
* 	remove object "s" from the hash table "tb". returns 0 if found,	   *
*       -1 otherwise.							   *
* 									   *
\**************************************************************************/


int h_del (struct h_table *tb, const char *s)
{
   int idx = hash (tb, s);		/* hash the string */
   struct h_node * p, ** pp;
   for (p = tb->h_tab [idx], pp = &tb->h_tab [idx]; p;
        pp = &p->flink, p = p->flink)
       /* traverse the bucket */
       if (! strcmp (p->h_name, s)) {	/* dump the mess */
           * pp = p->flink;		/* previous gets next */
           free (p->h_name);
           free (p);
           return 0;
	}
   return -1;
}

int h_deli (struct h_table *tb, long s)
{
   int idx = hashi (tb, s);		/* hash the string */
   struct h_node * p, ** pp;
   for (p = tb->h_tab [idx], pp = &tb->h_tab [idx]; p;
        pp = &p->flink, p = p->flink)
       /* traverse the bucket */
       if (p->h_name == (void *)  s) {	/* dump the mess */
           * pp = p->flink;		/* previous gets next */
           free (p);
           return 0;
	}
   return -1;
}


/* internal function to hash a string into an index, we will use modulo
   table size for grins for starters */

static inline int hash (struct h_table *tb, const char *s)
{
  register long acc = 0, i = 0;
  for (; * s; s++) {
      acc ^= (* s << i) + i;
      if (i < 16)
          i++;
      else
          break;
  }
  return acc % tb->h_size;
}

static inline int hashi (struct h_table *tb, long s)
{
  return s % tb->h_size;
}


/*
 * analyse the contents of a hash table returning
 * the, minimum, maximum and average size of each
 * bucket
 */

static int maxbk;

void h_analyse (struct h_table * tb, int * sz, int * ptot, int * min, int * max, double * avg, int * nempty) 
{
    int i, n, tot = 0;
    struct h_node * p;
    
    * min = 0x7fffffff;
    * max = 0;
    * nempty = 0;
    for (i = 0; i < tb->h_size; i++) {
        /* count the hits. */
        for (p = tb->h_tab [i], n = 0; p; p = p->flink, n++);
        if (n < * min)
            * min = n;
        if (n > * max) {
            * max = n;
            maxbk = i;
        }
        tot += n;
        if (n == 0)
            * nempty += 1;
    }
    /* just do a simple minded average.
     * it would probably be nice to know about clustering
     * phenomenon too.
     */
    * avg = (double) tot / (double) tb->h_size ;
    * ptot = tot;
    * sz = tb->h_size;
}


void h_printstats (const char * name, struct h_table * tb, int first) 
{

    int mn, mx, sz, tot, nempty;
    double avg, hitdepth;
    const char * fmt = "%-7.7s%-4d    %-4d  %-4d  %-4d %-4d  %-6d  %.3f %.3f\n";
    if (first)
        printf ("table  buckets total min   max  empty hits    depth average\n");

    hitdepth = (double) tb->depth / (double) tb->hits;
    h_analyse (tb, &sz, &tot, &mn, &mx, &avg, &nempty);
    printf (fmt, name, sz, tot, mn, mx, nempty, tb->hits, hitdepth, avg);
#if 0
    struct h_node * p;
    printf (" most heinous bucket:\n");
    for (p = tb->h_tab [maxbk]; p; p = p->flink) 
        printf ("%s ", p->h_name);
    printf ("\n");
#endif
        
}
