/*
 *                 Copyright (c) Apr 19 10:16:52  MTCC
 *	Author: Michael Thomas
 *	Module: hcrawl.c
 *	Created: Sat Apr 19 10:16:52 1997
 *	Abstract:
 *	   web crawling interface
 */

/* Edit History: 
 */


#include "html.h"
#include "crawl.h"

struct flist * files, ** pfiles = &files;
struct flist * stack = 0;
URL baseurl;



/*
 * conditionally insert the given url into the file
 * list. some day it will pay attention to depth,
 * which controls how far we've traverse: rooted,
 * on site, and off site.
 */

static int slevel (int init) 
{
    struct flist * p, * last = stack;
    int i;
    if (stack == 0)
        return 0;
    else
        if (stack->stack == 0)
            return init;
    
    for (i = init, p = stack->stack; p; p = p->stack) {
        /* ignore dummied up leaf nodes. */
        if (p->visited == 2)
            continue;
        if (strcmp (last->parent, p->parent))
            i++;
        last = p;
    }
    return i;
}



static struct flist * insfile (char * url, char * parurl, int depth) 
{
    struct flist * fl, * p;
    int visited = 0, type;
    char * ext;
    URL u, parent, * par;

    if (parurl) {
        parseurl (parurl, &parent, 0);
        par = &parent;
    } else {
        par = &baseurl;
        parurl = url;
    }
    if (parseurl (url, &u, par) < 0 || strcmp (u.proto, "http"))
        return 0;
    /* ??? Should I or should I not traverse these. no for now. */
    if ((ext = strchr (u.file, '#')))
        * ext = 0;
    /* never traverse anywhere where the path is not a
       subset of the initial path. */

    /* if the parent is off site, we're done */
    if (strcmp (par->host, baseurl.host)) 
        type = CRL_EXTRAFULL;
    else {
        /* if this node is off site, or it's outside of the
           base url push it, but flag it as such */
        if (strcmp (u.host, baseurl.host) ||
            strncmp (u.path, baseurl.path, strlen (baseurl.path))) {
            type = CRL_EXTRAFIRST;
        } else {
            /* at this point we know we're intrasite, but 
               for formating reasons, we might choose to
               consider this site external */
            type = CRL_INTRASITE;
            if (strncmp (par->path, u.path, strlen (par->path))) {
                /* it's an on site, but not a proper subset of
                   the parent's path. if this is the first time
                   we've seen this node, push it as a normal node,
                   otherwise it's a leaf. */
                for (p = files; p; p = p->flink)
                    if (! strcmp (p->url, url)) {
                        type = CRL_EXTRAFIRST;
                        break;
                    }
            }
        }
    }
    if (type > depth)
        return 0;
    if (type == CRL_EXTRAFIRST)
        visited = 2;
    url = canonicalurl (&u);
    /* check to see if we've visited this node before, but
       only if it doesn't appear so */
    if (visited == 0) {
        for (p = files; p; p = p->flink) 
            if (! strcmp (p->url, url)) {
                if (depth > CRL_INTRASITE) {
                    visited = 1;
                    break;
                }
                free (url);
                return 0;
            }
    }
    fl = malloc (sizeof (* fl));
    fl->url = url;
    fl->flink = 0;
    fl->visited = visited;
    fl->type = type;
    fl->parent = strdup (parurl);
    fl->stack = stack;
    stack = fl;
    * pfiles = fl;
    pfiles = &fl->flink;
    return fl;
}



static 
void _crawl (htelm * p, char * parent, int depth) 
{
    char * f;

    for (; p && p->flink; p = p->flink);
    
    for (; p; p = p->blink) {
        if (p->type == H_ANCHOR && (f = htkgetval (p->tk, "href"))) 
            insfile (f, parent, depth);
        if (p->slink)
            _crawl (p->slink, parent, depth);
    }        
}


/*
 * crawl through the url space starting at url
 * the action routine is the user interface
 * which is called with a bunch of parameters,
 * but mainly the url of the file, and its
 * parsed contents. The action routine can
 * cut the crawl short by returning 0.
 */

#include <signal.h>

typedef void (*sighandler_t)(int);

void alrm (int ctx)
{
}



int hcrawl (char * url, int depth, int (* action) ()) 
{
    int dotrav;
    int init = 0;
    int flg = depth & 0xffff0000;
    
    depth &= 0xffff;
    if (parseurl (url, &baseurl, 0) < 0) 
        return -1;
    insfile (url, 0, depth);    
    while (stack) {
        struct crawlnode node;

        node.url = stack->url;
        node.parent = stack->parent;
        node.level = slevel (init);
        node.file = 0;
        node.sts = 0;
        /* a leaf visit forces the file to be read in even if
           it's off site. This is used for a broken link search.
         */
        if (((flg & CRL_LEAFVISIT) && stack->visited == 2) ||
            (stack->type == CRL_INTRASITE && stack->visited == 0)) {
            sighandler_t old;
            /* read it in */
            old  = signal (SIGALRM, alrm);
            alarm (30);
            node.file = htpreadurl (0, 0, node.url, &node.sts, node.type);
            alarm (0);
            signal (SIGALRM, old);
            /* release unless a full scale crawl; we just want status */
            if (stack->type != CRL_INTRASITE && stack->type != CRL_EXTRAFULL) {
                htprelease (node.file);
                node.file = 0;
            }
        }
        if (stack->visited) {
            if (depth > CRL_INTRASITE)
                dotrav = action (&node); 
            stack = stack->stack;
            if (node.file)
                htprelease (node.file);
            continue;
        }

        /* let the user have at it. */        
        dotrav = action (&node); 
        stack->visited = 1;
        stack = stack->stack;   /* pop the stack */
        if (stack == 0) init++;
        if (dotrav)
            _crawl (node.file, node.url, depth);
        htprelease (node.file);
    }
    return 0;
}



