/*
 *                 Copyright (c) Apr  4 13:19:14  MTCC
 *	Author: Michael Thomas
 *	Module: hform.c
 *	Created: Fri Apr  4 13:19:14 1997
 *	Abstract:
 *	   form manipulation functions.
 */

/* Edit History: 
 */

#include "html.h"

extern htelm * htcur, * hthead;

static int curform = 0;


/*
 * get the n'th form
 */

htelm * htpformn (htelm * h, int * n) 
{
    for (; h; h = h->flink) {
        htelm * rv;
        if (h->type == H_FORM && (h->tk->flags & TK_ISEND) == 0) {
            if (* n == 0)
                return h;
            * n = * n - 1;            
        }
        if (h->slink && (rv = htpformn (h->slink, n)))
            return rv;        
    }
    return 0;
}

htelm * htformn (int n) 
{
    int t = n;
    htelm * h;
    if (n < 0)
        return 0;
    h = htpformn (hthead, &t);
    if (h) {
        htcur = h;
        curform = n;
    }
    return h;    
}


htelm * htformprev () 
{
    return htformn (curform-1);
}

    

htelm * htformnext () 
{
    htelm * h;
    
    h = htformn (curform);
    if (h)
        curform++;
    return h;
}


void htgotoform (char * fname) 
{
    htelm * h;
    int n = 0;
    
    htgotomark (fname);
    if ((h = htpformn (htcur, &n)))
        htcur = h;    
}



/*
 * retrieve the field within the form/whatever pointed to in
 * h
 */

htelm * htpgotofield (htelm * h, char * field) 
{
    for (h = h->slink; h; h = h->flink) {
        char * p = htkgetval (h->tk, "name");
        if (p && ! strcasecmp (field, p))
            return h;
    }
    return 0;
}

htelm * htgotofield (char * field) 
{
    htelm * h = htpgotofield (htcur, field);
    if (h)
        htcur = h;
    return h;
}



char * htpgetfval (htelm * h, char * field, char * var) 
{
    h = htpgotofield (h, field);
    if (h == 0)
        return 0;    
    return htkgetval (h->tk, var);
}

char * htgetfval (char * field, char * var) 
{
    return htpgetfval (htcur, field, var);
}
    

        
void htpputval (htelm * h, char * var, char * val) 
{
    struct token * tk, * p, ** pp;
    
    if ((tk = htkckvar (h->tk, var))) {
        /* already present */
        if (tk->flink && (tk->flink->flags & TK_VAL)) {
            free (tk->flink->name);        
            tk->flink->name = strdup (val);
        } else {
            for (p = h->tk, pp = &h->tk; p; pp = &p->flink, p = p->flink)
                if (p == tk)
                    break;
            p = calloc (sizeof (* p), 1);
            p->flags = TK_VAL;
            p->name = strdup (var);
            * pp = p;
        }
    } else {
        for (p = h->tk, pp = &h->tk; p; pp = &p->flink, p = p->flink);
        p = calloc (sizeof (* p), 1);
        p->flags = TK_VAR;
        p->name = strdup (var);
        * pp = p;
        if (val) {
            pp = &p->flink;
            p = calloc (sizeof (* p), 1);
            p->flags = TK_VAL;
            p->name = strdup (val);
            * pp = p;
        }
    }
}

void htpputtag (htelm * h, char * var) 
{
    struct token * tk, * p, ** pp;

    /* only if not already present */
    if ((tk = htkckvar (h->tk, var)) == 0) {
        for (p = h->tk, pp = &h->tk; p; pp = &p->flink, p = p->flink);
        p = calloc (sizeof (* p), 1);
        p->flags = TK_VAR;
        p->name = strdup (var);
        * pp = p;
    }    
}



void htpdeltag (htelm * h, char * var) 
{
    struct token * tk, * p, ** pp;

    /* only if not already present */
    if ((tk = htkckvar (h->tk, var)) == 0) {
        for (p = h->tk, pp = &h->tk; p && p->flink; pp = &p->flink, p = p->flink)
            if (p == tk) {
                * pp = p->flink;
                free (p);
                return;
            }
        
    }    
}

void htputftag (char * field, char * var) 
{
    htelm * h = 0;
    if ((h = htpgotofield (h, field)) == 0)
        return;
    htpputtag (h, var);
}


void htputval (char * field, char * val) 
{
    htpputval (htcur, field, val);
}


void htpputfval (htelm * h, char * field, char * var, char * val)
{
    if ((h = htpgotofield (h, field)) == 0)
        return;
    htpputval (h, var, val);
}


void htputfval (char * field, char * var, char * val) 
{
    htpputfval (htcur, field, var, val);
}




static void _htbindform (hrec * r, htelm * h) 
{
    htelm * s, * new;
    char * p, * o, * val;

    /* don't traverse down a nested form */
    if (h->type == H_FORM)
        return;
    
    for (; h; h = h->flink) {
        /* if there is no name field, or if this hrec has
           nothing to say about the field, mosey along. */
        if (h->tk == 0 ||
            (p = htkgetval (h->tk, "name")) == 0 ||
            (val = hrisvis (r, p)) == 0) {
            if (h->slink)
                _htbindform (r, h->slink);
            continue;
        }
        switch (h->type) {
            case H_SELECT:
                for (s = h->slink; s; s = s->flink) {
                    if (s->type == H_OPTION) {
                        /* if there is a value field in the
                           option, use it. */
                        if ((o = htkgetval (s->tk, "value"))) {
                            if (! strcmp (o, val))
                                htpputtag (s, "selected");
                        } else {
                            /* else just compare the adjacent text */
                            if (s->flink && s->flink->type == H_TXT) {
                                /* don't consider a newline in the
                                   body to be significant. */
                                char * tmp = alloca (strlen (s->flink->body));
                                strcpy (tmp, s->flink->body);
                                    
                                trim (tmp);
                                if (! strcmp (tmp, val))
                                    htpputtag (s, "selected");
                            }
                        }                                
                    }
                }
                break;
            case H_TXTAREA:
                /* the car of a textarea is its payload */
                /* yes, this is squirrelly. the normal stuff
                   outputs *after* htcur rather than before.
                   it's a big 'ol tradeoff, so let's not go
                   there, but since we indent on a text area,
                   we don't want our text after the end tag
                   */
                new = htpaddelm (0, 0, val);
                new->flink = h->slink;
                new->ulink = h;
                new->blink = 0;
                new->slink = 0;
                if (h->slink)
                    h->slink->blink = new;
                h->slink = new;
                break;
            case H_CHECK:
            case H_RADIO:
                /* see if there is a name field for this input */
                if ((o = htkgetval (h->tk, "value")) && ! strcmp (o, val))
                    htpputtag (h, "checked");                
                break;
            case H_INPUT:
            case H_PASSWORD:
            case H_HIDDEN:
                /* see if there is a name field for this input */
                htpputval (h, "value", webquote (val));
                break;
            default:
                break;
        }
        if (h->slink)
            _htbindform (r, h->slink);
    }
    
}

/*
 *  bind the contents of a hrec to the current form
 *  you must be sitting on the current form for anything
 *  to happen
 *
 */
void htbindform (hrec * r) 
{
    if (htcur->type != H_FORM)
        return;
    _htbindform (r, htcur->slink);
    
}
