
/*
 *                 Copyright (c) Mar 17 14:53:27  MTCC
 *	Author: Michael Thomas
 *	Module: hrec.c
 *	Created: Mon Mar 17 14:53:27 1997
 *	Abstract:
 *	   html record binding services
 */

/* Edit History: 
 */

#include "html.h"

static int hrvrfybase ();
static settype (hrecelm * r, int fmt, int subfmt, int (* uvfy) ());


hrec * newhrec () 
{
    hrec * h = calloc (sizeof (* h), 1);
    h->htbl = h_init (23);
    return h;
}

void hrfree (hrec * hr) 
{}


int hraddf (hrec * h, const char * var, const char * fmt, ELLIPSE)
{
    va_list ap;
    char bf [512];
    VA_START (ap, fmt);
    vsprintf (bf, fmt, ap);
    va_end (ap);
    return hradd (h, var, bf);
}

    

int hradd (hrec * h, const char * var, const char * val)
{
    hrecelm * hr;

    if ((hr = h_get (h->htbl, var))) {
        hrecerr err;
        free (hr->val);
        hr->val = strdup (val);
        err.e = hr;
        err.name = (char *) var;
        if (hr->fmt == HRF_BINARY) 
            hr->subfmt = strlen (val);
        return err.e->vrfy (&err);
    } else {
        hr = calloc (sizeof (*hr), 1);
        hr->var = strdup (var);        
        hr->flags = 0;
        hr->val = strdup (val);
        h_put (h->htbl, var, hr);
        return settype (hr, HRF_TEXT, 0, 0);
    }
}


/* add a character buffer */

int hraddbf (hrec * h, const char * var, const char * val, int len) 
{
    hrecelm * hr;

    if ((hr = h_get (h->htbl, var))) {
        hrecerr err;
        free (hr->val);
        hr->val = malloc (len+1);
        if (len)
            memcpy (hr->val, val, len);
        hr->val [len] = 0;      /* bogus, yes, but its better than blowups */
        hr->cvtval.cval = (char *) val;
        err.e = hr;
        err.name = (char *) var;
        return err.e->vrfy (&err);
    } else {
        hr = calloc (sizeof (*hr), 1);
        hr->var = strdup (var);        
        hr->flags = 0;
        hr->val = malloc (len+1);
        if (len)
            memcpy (hr->val, val, len);
        hr->val [len] = 0;      /* bogus, yes, but its better than blowups */
        hr->cvtval.cval = (char*) val;
        hr->subfmt = len;
        h_put (h->htbl, var, hr);
        return settype (hr, HRF_CHAR, len, 0);
    }
}


/* add a binary buffer; basically the same as HRF_CHAR,
   but the datawidth is variable */

int hraddbin (hrec * h, const char * var, const char * val, int len) 
{
    hrecelm * hr;

    if ((hr = h_get (h->htbl, var))) {
        hrecerr err;
        free (hr->val);
        hr->val = malloc (len+1);
        if (len)
            memcpy (hr->val, val, len);
        hr->val [len] = 0;      /* bogus, yes, but its better than blowups */
        hr->cvtval.cval = (char *) val;
        hr->subfmt = len;                
        err.e = hr;
        err.name = (char *) var;
        return err.e->vrfy (&err);
    } else {
        hr = calloc (sizeof (*hr), 1);
        hr->var = strdup (var);        
        hr->flags = 0;
        hr->val = malloc (len+1);
        if (len)
            memcpy (hr->val, val, len);
        hr->val [len] = 0;      /* bogus, yes, but its better than blowups */
        hr->cvtval.cval = (char *) val;
        hr->subfmt = len;        
        h_put (h->htbl, var, hr);
        return settype (hr, HRF_BINARY, len, 0);
    }
}

void hraddint (hrec * h, const char * var, int ival)
{
    hrecelm * hr;
    char val [100];

    sprintf (val, "%d", ival);
    if ((hr = h_get (h->htbl, var))) {
        free (hr->val);
        hr->val = strdup (val);
    } else {
        hr = calloc (sizeof (*hr), 1);
        hr->var = strdup (var);        
        hr->flags = 0;
        hr->val = strdup (val);
        hr->fmt = HRF_INT;
        hr->vrfy = hrvrfybase;    
        h_put (h->htbl, var, hr);
    }
    hr->cvtval.ival = ival;
    hr->fmt = HRF_INT;
}



void hradddouble (hrec * h, const char * var, double dval)
{
    hrecelm * hr;
    char val [100];
    
    sprintf (val, "%f", dval);
    if ((hr = h_get (h->htbl, var))) {
        free (hr->val);
        hr->val = strdup (val);
    } else {
        hr = calloc (sizeof (*hr), 1);
        hr->var = strdup (var);        
        hr->flags = 0;
        hr->val = strdup (val);
        hr->vrfy = hrvrfybase;                
        h_put (h->htbl, var, hr);
            
    }
    hr->cvtval.dval = dval;
    hr->fmt = HRF_DOUBLE;
}



/*
 * add an element with typing info
 */

void hrtypeadd (hrec * h, const char * var, const char * val, int fmt, int subfmt, int (* uvfy) ())
{
    hrecelm * hr;

    if ((hr = h_get (h->htbl, var))) {
        free (hr->val);
        hr->val = strdup (val);
    } else {
        hr = calloc (sizeof (*hr), 1);
        hr->flags = 0;
        hr->val = strdup (val);
        hr->var = strdup (var);        
        h_put (h->htbl, var, hr);
    }
    settype (hr, fmt, subfmt, uvfy);
}


void hrsetflg (hrec * h, const char * var, int flags) 
{
    hrecelm * hr;
    
    if ((hr = h_get (h->htbl, var)) == 0) {
        hradd (h, var, "");
        hr = h_get (h->htbl, var);
    }
    if (hr)
        hr->flags |= flags;            
}

void hrclrflg (hrec * h, const char * var, int flags) 
{
    hrecelm * hr;
    
    if ((hr = h_get (h->htbl, var)) == 0) {
        hradd (h, var, "");
        hr = h_get (h->htbl, var);
    }
    if (hr)
        hr->flags &= ~flags;    
}



/*
 * verify that a given type is kosher handles the normal
 * cases, and passes off special user defined types to
 * their own handler. This also has the side effect of
 * setting the converted value field in the hrecelm 
 */

static int
hrvrfybase (hrecerr * e) 
{
    int rv;
    hrecelm * he = e->e;

    /* assume the best */
    rv = 0;
    e->errno = 0;
    e->errmsg = "";
    switch (he->fmt) {
        case HRF_TEXT:
            /* always hunky dory */
            he->cvtval.tval = he->val;
            break;
        case HRF_CHAR:
            if (strlen (he->val) <= he->subfmt)
                he->cvtval.tval = he->val;
            else {
                e->errmsg = "field too large";
                rv = -1;
            }
            he->cvtval.cval = he->val;
            break;
        case HRF_BINARY:
            /* always hunky dory */
            he->cvtval.cval = he->val;
            break;            
        case HRF_INT: 
        {
            char * fmt;
            switch (he->subfmt) {
                default:
                case HRSF_DEC:
                    fmt = "%d";
                    break;
                case HRSF_HEX:
                    fmt = "%x";
                    break;
                case HRSF_OCT:
                    fmt = "%o";
                    break;
            }
            if (sscanf (he->val, fmt, &he->cvtval.ival) != 1) {
                /* error */
                e->errmsg = "Invalid integer";
                rv = -1;
            }
            break;
        }
        case HRF_DOUBLE:
            /* prolly ought to have a few subfmt's here too... */
            if (sscanf (he->val, "%lf", &he->cvtval.dval) != 1) {
                e->errmsg = "Invalid floating point";
                rv = -1;
            }
            break;
        case HRF_DATE:
        case HRF_URL:
        case HRF_MONEY:
            rv = 0;
            break;
        case HRF_USER:
            return he->vrfy (e);
            break;
        default:
            rv = -1;             /* wtf? */
            break;

    }
    return rv;
}

    


/*
 * verify an hrec; optionally fills in a descriptor
 * with the offending element and error message.
 * returns -1 on error, >= 0 otherwise.
 * 
 * NOTE: in order to use converted values, you must
 *  verify the hrec first since the verifier has the
 *  side effect of converting the value.
 */

int hrvrfy (hrec * r, hrecerr * err) 
{
    hrecerr e;
    hrecelm * h;
    
    if (err == 0) 
        err = &e;

    /* traverse new, adding it into obeying nomerge flags */
    h = h_first (r->htbl);
    while (h != 0) {
        err->e = h;
        err->name = h->var;
        if (err->e->vrfy (err) < 0)
            return -1;
        h = h_next (r->htbl);
    }
    return 0;
}

/*
 * set a preexisting hrec element referred to by name
 */

int hrsettype (hrec * h, const char * var, int fmt, int subfmt, int (* uvfy) ()) 
{
    hrecelm * hr;
    if ((hr = h_get (h->htbl, var))) {
        settype (hr, fmt, subfmt, uvfy);
        return 0;
    }
    return -1;
}
      



/*
 * Set the type of a hrec element. These case be base
 * types, or user defined. If you don't specify a handler
 * for a non base type, it will generate an error.
 *
 * Note that you can always override
 * the base type verifier if you don't like what it does
 * (ie, say, make a text field require Shakespeare or somesuch)
 */


static int settype (hrecelm * r, int fmt, int subfmt, int (* uvfy) ()) 
{
    hrecerr err;
    r->fmt = fmt;
    r->subfmt = subfmt;
    r->vrfy = uvfy ? uvfy : hrvrfybase;
    err.e = r;
    err.name = "";
    return r->vrfy (&err);    
}

            

void hrappend (hrec * h, const char * var, const char * val, int flags)
{
    hrecelm * hr;
    /* if it already exists, append it onto the end */
    if ((hr = h_get (h->htbl, var))) {
        hr->val = realloc (hr->val, strlen (hr->val)+strlen (val)+2);
        strcat (hr->val, "\n");
        strcat (hr->val, val);
        hr->flags = flags;
    } else {
        hr = calloc (sizeof (*hr), 1);
        hr->flags = flags;
        hr->val = strdup (val);
        settype (hr, HRF_TEXT, 0, 0);
        h_put (h->htbl, var, hr);
    }
}


char * hrget (hrec * h, const char * var) 
{
    hrecelm * hr = h_get (h->htbl, var);
    if (hr)
        return hr->val;
    else
        return "";
}

char * hrgetbf (hrec * h, const char * var, int * plen) 
{
    hrecelm * hr = h_get (h->htbl, var);
    if (hr) {
        if (plen)
            * plen = hr->subfmt;
        return hr->cvtval.cval;
    } else {
        if (plen)
            * plen = 0;
        return "";
    }
}

/* for completeness */

char * hrgetbin (hrec * h, const char * var, int * plen)  
{
    return hrgetbf (h, var, plen);
}


int hrgetint (hrec * h, const char * var) 
{
    hrecelm * hr = h_get (h->htbl, var);
    if (hr) {
        /* should I convert type here too? or should
           I just keep up with the sort of lazy type
           checking that I'm going. hmm.
         */
        if (hr->fmt != HRF_INT) {
            hr->cvtval.ival = atol (hr->val);
        }
        return hr->cvtval.ival;
    }
    else
        return -1;
}

double hrgetdouble (hrec * h, const char * var) 
{
    hrecelm * hr = h_get (h->htbl, var);
    if (hr) {
        if (hr->fmt != HRF_DOUBLE) {
            hr->cvtval.dval = atof (hr->val);
        }
        return hr->cvtval.dval;
    } else
        return -1;
}

char * hrgetvis (hrec * h, const char * var) 
{
    hrecelm * hr = h_get (h->htbl, var);
    if (hr && (hr->flags & HR_HIDDEN) == 0)
        return hr->val;
    else
        return "";
}

char * hrisvis (hrec * h, const char * var) 
{
    hrecelm * hr = h_get (h->htbl, var);
    if (hr && (hr->flags & HR_HIDDEN) == 0)
        return hr->val;
    else
        return 0;
}

/* get the hrecelm itself. */

hrecelm * hrgeth (hrec * h, const char * var) 
{
    hrecelm * hr = h_get (h->htbl, var);
    return hr;
}

/*
 * find a match for field "var" in the list of
 * records starting at h, comparing for "val".
 */

hrec * hrfind (hrec * h, const char * var, const char * val) 
{
    for (; h; h = h->flink) {
        hrecelm * e = h_get (h->htbl, var);
        if (! strcmp (e->val, val))
            return h;
    }
    return 0;
}

hrecelm * hrfirst (hrec * r) 
{
    return h_first (r->htbl);
}


hrecelm * hrnext (hrec * r) 
{
    return h_next (r->htbl);
}


/*
 * merge new into old
 */

void hrmerge (hrec * old, hrec * new)
{
    hrecelm * p, * o;

    /* traverse new, adding it into obeying nomerge flags */
    p = h_first (new->htbl);
    while (p) {
        /* obey both sets of nomerge flags */
        if (p->flags & HR_NOMERGE) {
            p = h_next (new->htbl);            
            continue;    
        }
        
        o = h_get (old->htbl, p->var);
        if (o && (o->flags & HR_NOMERGE)) {
            p = h_next (new->htbl);                        
            continue;
        }
        /* update, adding if necessary */
        hradd (old, p->var, p->val);
        p = h_next (new->htbl);
    }
}

/* update a record. like merge, but update doesn't
   create new fields in old */

void hrupdate (hrec * old, hrec * new)
{
    hrecelm * p, * o;

    /* traverse new, adding it into obeying nomerge flags */
    p = h_first (new->htbl);
    while (p) {
        /* obey both sets of nomerge flags */
        if (p->flags & HR_NOMERGE) {
            p = h_next (new->htbl);            
            continue;    
        }
        
        o = h_get (old->htbl, p->var);
        if (o == 0 || (o->flags & HR_NOMERGE)) {
            p = h_next (new->htbl);                        
            continue;
        }
        /* update, the record */
        hradd (old, p->var, p->val);
        p = h_next (new->htbl);
    }
}

/* clone the hrec */

hrec * 
hrclone (hrec * old)
{
    hrecelm * p;
    hrec * new = newhrec ();
    
    /* traverse new, adding it into obeying nomerge flags */
    p = h_first (old->htbl);
    while (p) {
        /* obey both sets of nomerge flags */
        if (p->flags & HR_NOMERGE) {
            p = h_next (old->htbl);            
            continue;    
        }
        /* update, the record */
        hrtypeadd (new, p->var, p->val, p->fmt, p->subfmt, p->vrfy);
        p = h_next (old->htbl);
    }
    return new;
}


            
            

static hrec *    
readrec (FILE * db) 
{
    hrec * r = 0;
    char bf [256], * p;
    

    while (fgets (bf, 256, db) != NULL) {
        int l = strlen (bf);
        /* kill the stupid trailing nl */        
        if (l > 0 && bf [l-1] == '\n')
            bf [l-1]  = 0;
        
        /* toss comment lines */
        if (bf [0] == '#' || bf [0] == 0)
            continue;
        /* end of record */
        if (bf [0] == '\f') 
            return r;
        /* check the first word */
        if ((p = strchr (bf, ':')) == 0) {
            continue;
        }
        /* kill the colon and the leading white space */
        * p = 0;
        p++;
        while (* p && isspace (* p))
            p++;
        /* concatenate the strings if necessary */
        if (r == 0)
            r = newhrec ();        
        hrappend (r, bf, p, 0);
    }
    return r;
}

hrec * hrreadstrrec (hrec * r, const char * str) 
{
    char bf [256], last [256], * p;
    
    last [0] = 0;
    while (1) {
        int l;
        if ((p = strchr (str, '\n'))) {
            l = p - str;
            memcpy (bf, str, l);
            bf [l] = 0;
            str = p+1;         /* past th */
        } else {
            if (* str == 0)
                break;
            strcpy (bf, str);
            str += strlen (str);
        }
        /* end of record */
        if (bf [0] == '\f') 
            return r;
        /* check the first word */
        if ((p = strchr (bf, ':')) == 0) {
            continue;
        }
        /* kill the colon and the leading white space */
        * p = 0;
        p++;
        while (* p && isspace (* p))
            p++;
        /* concatenate the strings if necessary */
        if (r == 0)
            r = newhrec ();
        if (! strcmp (bf, last))
            hrappend (r, bf, p, 0);
        else
            hradd (r, bf, p);
        strcpy (last, bf);
    }
    return r;
}

hrec * hrreadfp (FILE * fp)
{
    hrec * h, * head = 0, * cur = 0;
    while ((h = readrec (fp))) 
        if (head == 0)
            head = cur = h;
        else {
            cur->flink = h;
            h->blink = cur;
            cur = h;
        }
    return head;
    
}

hrec * hrreadfile (const char * file) 
{
    FILE * fp = fopen (file, "r");
    hrec * h;

    if (fp == 0)
        return 0;
    
    h = hrreadfp (fp);
    fclose (fp);
    return h;
}

static void
dumprec (FILE * fp, hrec * h) 
{
    hrecelm * e;

    char * bf, * p, * l;

    e = h_first (h->htbl);
    while (e) {
        /* this rigamoroll is due to needing to split text
           fields into multiple lines */
        if (e->flags & HR_HIDDEN) {
            e = h_next (h->htbl);
            continue;
        }
        bf = alloca (strlen (e->val));
        strcpy (bf, e->val);            
        for (p = bf, l = p; * p; p++) 
            if (* p == '\n') {
                * p = 0;
                fprintf (fp, "%s: %s\n", e->var, l);
                l = p+1;
            }
        if (* l)
            fprintf (fp, "%s: %s\n", e->var, l);
        e = h_next (h->htbl);        
    }
    fprintf (fp, "\f\n");
}

int hrdumpstrrec (hrec * h, char * str, int len) 
{
    hrecelm * e;
    char * bf, * p, * l;

    e = h_first (h->htbl);
    while (e) {
            /* this rigamoroll is due to needing to split text
               fields into multiple lines */
        if (e->flags & HR_HIDDEN) {
            e = h_next (h->htbl);
            continue;
        }
        bf = alloca (strlen (e->val));
        strcpy (bf, e->val);
        for (p = bf, l = p; * p; p++) 
            if (* p == '\n') {
                * p = 0;
                len -= strlen (e->var)+strlen (l) + 3;
                if (len < 3)
                    return -1;
                sprintf (str, "%s: %s\n", e->var, l);
                str += strlen (str);
                l = p+1;
            }
        if (* l) {
            len -= strlen (e->var)+strlen (l) + 3;
            if (len < 3)
                return -1;                
            sprintf (str, "%s: %s\n", e->var, l);
            str += strlen (str);
        }
        e = h_next (h->htbl);        
    }
    if (len < 3)
        return -1;
    strcat (str, "\f\n");
    return 0;
}


int hrwritefile (const char * file, hrec * recs) 
{
    FILE * fp = fopen (file, "w");

    if (fp == 0)
        return -1;
    
    for (; recs; recs = recs->flink)
        dumprec (fp, recs);
    fclose (fp);
    return 0;
}


/*
 * read a post command's contents into a hrec
 */

hrec * hrgetorpost (hrec * h, char * bf, int asize)
{
    char * p, * field, * val;

    if (h == 0)
        h = newhrec ();
    field = bf;
    val = "";
    for (p = bf; p < bf+asize; p++) {
        if (* p == '&') {
            char * f, * v;
            * p = 0;             /* which will null terminate the val */
            f = alloca (strlen (field)+1);
            v = alloca (strlen (val)+1);            
            unwebify (field, f);
            unwebify (val, v);
            hradd (h, f, v);
            field = p+1;
        }
        if (* p == '=') {
            * p = 0;
            val = p+1;
        }
        if (* p == '\n')
            * p = 0;
        if (* p == '+')
            * p = ' ';
        
    }
    /* get the last dribble */
    * p = 0;
    if (* field) {
        char * f, * v;
        f = alloca (strlen (field)+1);
        v = alloca (strlen (val)+1);        
        unwebify (field, f);
        unwebify (val, v);
        hradd (h, f, v);
    }
    return h;    
}





hrec * hrreadget (hrec * h, int ac, char ** av) 
{
    char * p = getenv ("QUERY_STRING");
    /* for reasons I profess to not understand,
       QUERY_STRING seems to always be accurrate.
       I've seen Apache send a null ac,av where
       query_string is correct. use it in preference.
     */
    if (p)
        return hrgetorpost (h, p, strlen (p));
    
    else if (ac > 1) { 
       if (h == 0)
            h = newhrec ();
        av++;
        ac--;
        while (ac-- > 0) {
            hrgetorpost (h, * av , strlen (* av));
            av++;
        }
    } 
    return h ? h : newhrec ();
}



hrec * hrreadpost (hrec * h) 
{
    char * p, * bf;
    int size, asize = 0, msize, rv;
    
    if ((p = getenv ("CONTENT_LENGTH")))
        size = atol (p);
    else
        size = 512;
    msize = size+1;    
    p = bf = malloc (size+1);
    /* read the form */
    while ((rv = read (0, p, size)) > 0) {
        size -= rv;
        if (size <= 0) {
            int t = p-bf;
            msize += 512;
            bf = realloc (bf, msize);
            /* reposition p */
            p = bf+t;
        }
        p += rv;
        asize += rv;
    }
    * p = 0;
    h = hrgetorpost (h, bf, asize);
    free (bf);
    return h;
}




hrec * hrgetparams (hrec * h, int ac, char ** av) 
{
    char * p = getenv ("REQUEST_METHOD");
    
    if (p && ! strcasecmp (p, "POST"))
        return hrreadpost (h);
    else
        return hrreadget (h, ac, av);
}


void hrlink (hrec ** list, hrec * new) 
{
    new->flink = * list;
    if (* list)
        (*list)->blink = new;
    * list = new;
            
}

void hrdelink (hrec ** list, hrec * old) 
{
    /* delete the item */
    if (old->flink) 
        old->flink->blink = old->blink;
    if (old->blink) 
        old->blink->flink = old->flink;
    if (old == * list)
        * list = old->flink;
}


void hrdump (hrec * r) 
{
    hrecelm * h;

    h = hrfirst (r);

    while (h) {
        printf (" %s: %s\n", h->var, h->val);
        h = hrnext (r);
    }
    
}
