/*
 *                 Copyright (c) Mar 20 13:27:39  MTCC
 *	Author: Michael Thomas
 *	Module: hsql.c
 *	Created: Thu Mar 20 13:27:39 1997
 *	Abstract:
 *	   interface routines to msql
 */

/* Edit History: 
 */



#include "html.h"

#ifdef WANTMSQL


#include <msql.h>

int hminit (char * host, char * db) 
{
    int sock;
    
    if (host == 0 || * host == 0)
        host = "localhost";
    
    sock = msqlConnect (host);
    if (sock < 0)
        return -1;
    if (msqlSelectDB (sock, db) < 0)
        return -1;
    return sock;
}



/*
 * find a particular row in the named table and fill a hrec
 * with the results
 */

int msqlQueryf (int sock, char * fmt, ...) 
{
    va_list ap;
    char bf [1024];
    
    VA_START (ap, fmt);
    vsnprintf (bf, 1024, fmt, ap);
    va_end (ap);
    return msqlQuery (sock, bf);
}


hrec * hmfind (int sock, hrec * hr, char * table, char * field, char * val) 
{
    char bf [512], * p;
    m_result * r;
    m_row row;
    int rv, col = 0;
    hrecelm * e;
    
    /* build the query string */
    p = bf;
    sprintf (p, "select ");
    p += strlen (p);
    e = h_first (hr->htbl);
    while (e) {
        if (e->flags & HR_HIDDEN) {
            e = h_next (hr->htbl);
            continue;
        }
        sprintf (p, "%s,", e->var);
        p += strlen (p);
        col++;
        e = h_next (hr->htbl);        
    }
    /* kill the trailing comma */
    *--p = 0;
    if (* field) 
        sprintf (p, " from %s where %s='%s'", table, field, slashify (val));
     else
        sprintf (p, " from %s", table);
    if ((rv = msqlQuery (sock, bf)) <= 0)
        return 0;
    /* if more than one row, use the first. */
    r = msqlStoreResult ();
    row = msqlFetchRow (r);
    col = 0;
    /* fill the hrec with the new data */
    e = h_first (hr->htbl);
    while (e) {
        if (e->flags & HR_HIDDEN) {
            e = h_next (hr->htbl);
            continue;
        }
        
        if (row [col])
            hradd (hr, e->var, row [col++]);
        else {
            hradd (hr, e->var, "");
            col++;
        }
        e = h_next (hr->htbl);        
    }
    msqlFreeResult (r);    
    return hr;
}



int hmfetchrowsf (int sock, hrec * hr, char * table, void ** pctx, char * fmt, ...) 
{
    va_list ap;
    char bf [1024], * p;
    m_row row;
    int rv = 0, col = 0;
    hrecelm * e;
    m_result ** ctx = (m_result **) pctx;
    
    /* build the query string */
    if (* ctx == 0) {
        p = bf;
        sprintf (p, "select ");
        p += strlen (p);
        e = h_first (hr->htbl);
        while (e) {
            if (e->flags & HR_HIDDEN) {
                e = h_next (hr->htbl);                
                continue;
            }
            sprintf (p, "%s,", e->var);
            p += strlen (p);
            col++;
            e = h_next (hr->htbl);
        }
        /* kill the trailing comma */
        *--p = 0;
        sprintf (p, " from %s ", table);
        p += strlen (p);
        if (fmt) {    
            VA_START (ap, fmt);
            vsnprintf (p, 1024, fmt, ap);
            va_end (ap);
        }            
        if ((rv = msqlQuery (sock, bf)) <= 0)
            return rv;
        * ctx = msqlStoreResult ();
    }
    
    if ((row = msqlFetchRow (* ctx)) == 0) {
        msqlFreeResult (* ctx);
        * ctx = 0;
        return 0;
    }
    col = 0;
    /* fill the hrec with the new data */
    e = h_first (hr->htbl);
    while (e) {
        char * v;
            
        if (e->flags & HR_HIDDEN) {
            e = h_next (hr->htbl);
            continue;
        }
        if (row [col])
            v = row [col++];
        else {
            v = "";
            col++;
        }
        /* type conversion is dealt with by hradd */
        hradd (hr, e->var, v);
        e = h_next (hr->htbl);
    }
    return 1;
}

int hmfetchrows (int sock, hrec * hr, char * table, void ** ctx) 
{
    return hmfetchrowsf (sock, hr, table, ctx, 0);
}


int hmmax (int sock, char * table, char * field) 
{
    int rv = 0;
    m_result * r;
    m_row row;

    if (msqlQueryf (sock, "select %s from %s", field, table) <= 0)
        return -1;
    r = msqlStoreResult ();
    while ((row = msqlFetchRow (r))) {
        int new = atol (row [0]);
        if (new > rv)
            rv = new;
    }
    msqlFreeResult (r);    
    return rv;
}


char * slashify (char * str) 
{
    static char rv [2048];
    char * p;
    
    p = rv;
    while (* str) {
        if (* str == '\'')
            * p++ = '\\';
        * p++ = * str++;
    }
    * p = 0;
    return rv;
}


/*
 * update a record given the hrec. if it's not found,
 * one will be created.
 *
 * FIXME: more intelligence at buffer overflow!!!
 */

#define UPDINS  0
#define INSONLY 1
#define UPDONLY 2


static int _hmupdate (int sock, hrec * hr, char * table, char * qual, int cmd) 
{
    
    char bf [2048], * p;
    hrecelm * e;
    int rv;
    
    
    /* build the query string */
    if (cmd != INSONLY) {
        p = bf;
        sprintf (p, "update %s set ", table);
        p += strlen (p);

        e = h_first (hr->htbl);
        while (e) {
            if (e->flags & HR_HIDDEN) {
                e = h_next (hr->htbl); 
                continue;
            }
            switch (e->fmt) {
                case HRF_INT:
                    sprintf (p, "%s=%ld,", e->var, e->cvtval.ival);
                    break;
                case HRF_DOUBLE:
                    sprintf (p, "%s='%f',", e->var, e->cvtval.dval);
                    break;
                default:
                    sprintf (p, "%s='%s',", e->var, slashify (e->val));
                    break;
            }
            p += strlen (p);
            e = h_next (hr->htbl);
        }
        /* kill the trailing comma */
        *--p = 0;
        sprintf (p, " %s", qual);
        if ((rv = msqlQuery (sock, bf)) > 0)
            return rv;
        if (cmd == UPDONLY)
            return rv;
    }
    p = bf;
    sprintf (p, "insert into %s (", table);
    p += strlen (p);
    /* yes this is a little lame traversing the hash table
       for the row twice, but it shouldn't be all that
           big, so it's easier to think of this way. I'm not
           the one who came up with the lame inconsistancies
           in SQL's syntax after all.
     */
    e = h_first (hr->htbl);
    while (e) {
        if (e->flags & HR_HIDDEN) {
            e = h_next (hr->htbl); 
            continue;
        }
        sprintf (p, "%s,", e->var);
        p += strlen (p);
        e = h_next (hr->htbl);
    }
    /* kill the trailing comma */
    *--p = 0;
    sprintf (p, ") values (");
    p += strlen (p);
    e = h_first (hr->htbl);
    while (e) {    
        if (e->flags & HR_HIDDEN) {
            e = h_next (hr->htbl);            
            continue;
        }
        switch (e->fmt) {
            case HRF_INT:
                sprintf (p, "%ld,", e->cvtval.ival);
                break;
            case HRF_DOUBLE:
                sprintf (p, "%f,", e->cvtval.dval);
                break;
            default:
                sprintf (p, "'%s',", slashify (e->val));
                break;
        }            
        p += strlen (p);
        e = h_next (hr->htbl);
    }
    p [-1] = ')';
    rv = msqlQuery (sock, bf);        
    return rv;        
}

int hmupdate (int sock, hrec * hr, char * table, char * var, char * val) 
{
    char qual [512];
    if (var) 
        snprintf (qual, 512, "where %s='%s'", var, slashify (val));
    else
        qual [0] = 0;
    return _hmupdate (sock, hr, table, qual, UPDINS);
}


int hminsert (int sock, hrec * hr, char * table) 
{
    return _hmupdate (sock, hr, table, 0, INSONLY);
}


int hmupdatef (int sock, hrec * hr, char * table, char * fmt, ...) 
{
    va_list ap;
    char qual [1024];
    VA_START (ap, fmt);
    vsnprintf (qual, 1024, fmt, ap);
    va_end (ap);
    return _hmupdate (sock, hr, table, qual, UPDINS);
}



int hmdelete (int sock, char * table, char * field, char * val) 
{
    if (field)
        return msqlQueryf (sock, "delete from %s where %s='%s'",
                           table, field, slashify (val));
    else
        return msqlQueryf (sock, "delete from %s", table);    
}


int hmdeletef (int sock, char * table, char * fmt, ...)
{
    va_list ap;
    char bf [1024];

    VA_START (ap, fmt);
    sprintf (bf, "delete from %s ", table);
    vsnprintf (bf+strlen (bf), 1024-strlen (bf), fmt, ap);
    va_end (ap);
    return msqlQuery (sock, bf);
}

    
#endif
