/*
 *              Copyright (c) Apr  4 13:18:04  MTCC
 *	Author: Michael Thomas
 *	Module: htable.c
 *	Created: Fri Apr  4 13:18:04 1997
 *	Abstract:
 *	   table manipulation functions.
 */

/* Edit History: 
 */

#include "html.h"

extern htelm * htcur, * hthead;


static void
dumppseudotag (htelm * h, char * name, char * omit)
{
    struct token * tk;
    
    htfprintf ("<%s ", name); 
    for (tk = h->tk->flink; tk; tk = tk->flink) {
        if (tk->flags & TK_ISEND)
            htfprintf ("/%s", tk->name);
        else if (tk->flags & TK_VAL) {
            if (tk->flags & TK_HASQ)
                htfprintf ("=\"%s\"", tk->name);
            else
                htfprintf ("=\"%s\"", tk->name);
        } else {
            /* strip the name field here. */
            if (! strcmp (tk->name, omit)) {
                if (tk->flink && tk->flink->flags & TK_VAL)
                    tk = tk->flink;
                continue;
            }
            htfprintf ("%s", tk->name);
        }
        if (tk->flink && (tk->flink->flags & TK_VAL) == 0)
            htfprintf (" ");                
    }
    htfprintf (">\n");
    htflush ();
}



/*
 * create a new table starting at htcur from the
 * template pointed at in tb
 */

void httablestart (htelm * tb) 
{
    if (tb) {
        dumppseudotag (tb, "table", "name");
        /* dump out any <th>'s */
        for (tb = htpgototag (tb, "th"); tb;
             tb = htpgototag (tb->flink, "th")) {
            dumppseudotag (tb, "th", "name");
            _htpdumpfile (tb->slink, -1, htfprintf);
            htflush ();
        }    
    } else
        htprintf ("<table>");
}




/*
 * tricky: bind a hrec to a particular table row
 * this is accomplished by using the table cells
 * which have fake name fields in them which
 *
 */

void htbindtable (htelm * tb, hrec * r,  int (* callback) (char *, char *, hrec *))
{
    htelm * row = 0; 
    htelm * td = 0;
    htelm * svpos;
    int insat;

    /* if a tb is specified, use it */
    if (tb) {
        td = htpgototag (tb, "td");
        row = htpgototag (tb, "tr");
    }
    
    /* if no table heading, just splat the tr out
       the way it comes from the hash table. */
    
    /* dump the row descriptor */
    if (row)
        dumppseudotag (row, "tr", "name");
    else {
        htprintf ("<tr>\n");
    }
    if (td) {
        for (; td; td = htpgototag (td->flink, "td")) {
            char * p = htkgetval (td->tk, "name");
            insat = 1;
            svpos = htcur;
            dumppseudotag (td, "td", "name");
            if (p) {
                if (callback) {
                    insat = callback (p, hrgetvis (r, p), r);
                } else                
                    htprintf ("%s", hrgetvis (r, p));
            }
            /* dump any stuff between <td></td> */
            switch (insat) {
                case 0:
                    break;
                case -1:
                    htcur = svpos;
                    /* fall through */
                case 1:
                    _htpdumpfile (td->slink, H_TBCELL, htfprintf);
                    break;
            }
            htflush ();
            htprintf ("</td>\n");
        }
    } else {
        hrecelm * e;

        e = h_first (r->htbl);
        while (e) {
            insat = 1;
            svpos = htcur;
            if (e->flags & HR_HIDDEN) {
                e = h_next (r->htbl);
                continue;   
            }
            htprintf ("<td>");
            if (callback)
                insat = callback (e->var, e->val, r);
            else                    
                htprintf ("%s", e->val);
            /* dump any stuff between <td></td> */
            switch (insat) {
                case 0:
                    break;
                case -1:
                    htcur = svpos;
                    /* fall through */
                case 1:
                    _htpdumpfile (td->slink, H_TBCELL, htfprintf);
                    break;
            }

            htprintf ("</td>");
            e = h_next (r->htbl);            
        }
    }
    htprintf ("</tr>\n");
}

/* find the length of a table definition's row */

int httablerowlen (htelm * tb) 
{
    htelm * td;
    int i;
    

    for (i = 0, td = htpgototag (tb, "td");
         td; td = htpgototag (td->flink, "td"), i++);
    return i;
}


void httableend ()
{
    htprintf ("</table>");
}


htelm * htgettabledef (char * name) 
{
    htelm * h;
    /* ??? does this do recursion correctly? prolly not. */
    for (h = htpgototag (hthead,"$table"); h;
         h = htpgototag (h->flink, "$table")) {
        char * p = htkgetval (h->tk, "name");
        if (p && ! strcmp (name, p))
            return h;
    }
    return 0;
}
