/*
 * Common Web interface routines for cgi stuff.
 * 970208/mat
 */


#include "html.h"


void preHTML (void)
{
    htprintf ("Content-type: text/html\n<$mark mime-headers>\n");
}

void preTEXT (void)
{
    htprintf ("Content-type: text/plain\n<$mark mime-headers>\n");
}



void startHTML (const char * title, const char * base)
{
    static started = 0;

    if (started)
        return;
    htprintf ("<html><head><base href=%s>", base ? base : "/");    
    htprintf ("<title>%s</title></head>\n", title);
    
    htprintf ("<BODY>\n");    
    htprintf ("<center><h1>%s</h1></center>", title);
    started++;
}

void endHTML (void)
{
    htprintf ("<p></body></html>\n");
    htdumpfile ();
    
}

void formHTML (const char * method, const char * action, const char * parm) 
{
    if (parm)
        htprintf ("<form method=%s action=%s?%s>\n", method, action, parm);
    else
        htprintf ("<form method=%s action=%s>\n", method, action);        
}


void endformHTML (void)
{
    htprintf ("</form>");
}


void selectHTML (const char * label, const char * name, const char ** vec, const char * selected) 
{
    htprintf ("%s<select name=\"%s\">\n", label, name);
    while (* vec) {
        if (! strcmp (selected, * vec))
            htprintf ("<option selected>%s\n", * vec);
        else
            htprintf ("<option>%s\n", * vec);
        vec++;
    }
    htprintf ("</select>\n");
}

void optionHTML (const char * val, const char * txt, int selected)
{

    if (val) 
        htprintf ("<option value=\"%s\"%s>%s",
                  webquote (val), selected ? " selected" : "", txt);
    else
        htprintf ("<option %s>%s",
                  selected ? " selected" : "", txt);        
}    


void radioHTML (const char * label, const char * name, const char ** vec, const char * selected) 
{
    htprintf ("%s", label);
    while (* vec) {
        if (! strcmp (selected, * vec))
            htprintf ("<input type=radio name=\"%s\" value=\"%s\" checked>%s \n",
                    name,  webquote (* vec), * vec);
        else
            htprintf ("<input type=radio name=\"%s\" value=\"%s\">%s \n",
                    name, webquote (* vec), * vec);
        vec++;
    }    
}

void checkHTML (const char * label, const char * name, const char * valtag, const char * val) 
{
    htfprintf ("%s<input type=checkbox name=\"%s\" value=\"%s\" ",
            label, name, webquote (valtag));
    if (! strcmp (val, valtag))
        htfprintf ("checked>");
    else
        htfprintf (">");
    htflush ();
}


void inputHTML (const char * label, const char * name, const char * value,
           enum itype type, int size, int maxsize) 
{
    htfprintf ("%s<input name=\"%s\" value=\"%s\" ",
            label, name, webquote (value));
    switch (type) {
        case reset:
            htfprintf ("type=reset ");
            break;
        case submit:
            htfprintf ("type=submit ");
            break;
        case hidden:
            htfprintf ("type=hidden ");
            break;
        case check:
            htfprintf ("type=check ");
            break;
        case input:
            htfprintf ("size=%d maxsize=%d", size, maxsize);
            break;
        default:
            break;
    }
    htfprintf (">\n");
    htflush ();
}

void textHTML (const char * label, const char * name, const char * val, int cols, int rows) 
{
    htprintf ("%s<textarea name=\"%s\" cols=%d rows=%d>%s</textarea>\n",
            label, name, cols, rows, webquote (val));
}

void aHTML (const char * url, const char * txt) 
{
    htprintf ("<a href=\"%s\">", url);
    if (txt && * txt)
        htprintf ("%s</a>", txt);
}

void endaHTML (void)
{
    htprintf ("</a>");
}

void dl (int on)
{
    if (on)
        htprintf ("<dl>");
    else
        htprintf ("</dl>");
}

void dt (void)
{
    htprintf ("<dt>");
}

void dd (void)
{
    htprintf ("<dd>");
}


void pre (int on)
{
    if (on)
        htprintf ("<pre>");
    else
        htprintf ("</pre>");
}



void center (int on) 
{
    if (on)
        htprintf ("<center>");
    else
        htprintf ("</center>");
}


void br (int n) 
{
    while (n-- > 0)
        htprintf ("<br>");
}

void par (void)
{
    htprintf ("<p>\n");
}

void bq (int n) 
{
    if (n > 0)
        while (n-- > 0)
            htprintf ("<blockquote>");
    else
        while (n++ < 0)
            htprintf ("</blockquote>");
}

void hr (void)
{
    htprintf ("<hr>");
}


void hl (int n) 
{
    if (n > 0)
        htprintf ("<h%d>", n);
    else
        htprintf ("</h%d>", -n);
}



void tableHTML (int on, int border) 
{
    if (on)
        htprintf ("<table border=%d>\n", border);
    else
        htprintf ("</table>\n");
}


void tr (int on) 
{
    if (on)
        htprintf ("<tr>");
    else
        htprintf ("</tr>");
}

void td (int on) 
{
    if (on)
        htprintf ("<td>");
    else
        htprintf ("</td>");
}


void th (const char * string) 
{
    htprintf ("<th>%s</th>", string);
}

void li (void) 
{
    htprintf ("<li>");
}
