/*
 * various html defines.
 */

#ifndef h_html
#define h_html

#include "config.h"
#include "hash.h"

enum itype { input, submit, hidden, selectbar, check, radio, reset, text
};

struct varval 
{
    char * name;
    int type;
    int ( * fn) ();
    
};

struct token 
{
    char * name;
    char flags;
#define TK_TAG 1
#define TK_ISEND 2
#define TK_VAR  4
#define TK_VAL  8
#define TK_HASQ 0x10
    struct token * flink;
};


typedef struct HTELM
{
    int type;
#define H_TXT     0
#define H_MISCTAG 1
#define H_FORM    2
#define H_SELECT  3
#define H_OPTION  4
#define H_TXTAREA 5
#define H_INPUT   6
#define H_RADIO   7
#define H_CHECK   8
#define H_HIDDEN  9
#define H_SUBMIT  10
#define H_RESET   11
#define H_PASSWORD 12
#define H_TABLE   13
#define H_TBHDR   14
#define H_TBROW   15
#define H_TBCELL  16
#define H_ANCHOR  17
#define H_PSSTART 1000
#define H_MARKER  1000          /* pseudo types, marker (gotomark) */
#define H_DATE    1001          /* insert date */
#define H_TABLEDEF 1002
#define H_IF      1003
#define H_INC     1004
#define H_SET     1005
#define H_ELSE     1006
#define H_NOTAG 0xffff
    struct HTELM  * flink;      /* forward, backward, and sublinks */
    struct HTELM  * blink;
    struct HTELM  * slink;
    struct HTELM  * ulink;
    char * body;                /* the text body */
    int len;                     /* current length */
    int alen;                   /* allocation length */
    int lineno;
    struct token * tk;
} htelm;

htelm * htpreadfile (),
      * htreadfile (),
      * htpreadstr (),
    * htgotomark ();

char * htkgetval ();
struct token * htkckvar (), * htkgetword ();


    
typedef struct HREC
{
    struct h_table * htbl;
    struct HREC * flink;
    struct HREC * blink;
} hrec;

typedef struct HRECELM 
{
    char * var;                 /* the variable this element represents */
    char * val;                 /* char string value of the element */
    short flags;                /* various flags dealing with binding, etc. */
#define HR_HIDDEN 1
#define HR_NOMERGE 2
#define HR_NODUMP  4
    unsigned short fmt;         /* on input, elements can be verified */
#define HRF_TEXT   0            /* text, unbound string array */
#define HRF_CHAR   1            /* character array (bound by dimension) */
#define HRF_INT    2            /* integer */
#define HRF_DOUBLE 3            /* floating point */
#define HRF_DATE   4            /* canonical date */
#define HRF_URL    5            /* a valid URL */
#define HRF_MONEY  6            /* money type, (really a float) */
#define HRF_BINARY 7            /* binary data, subfmt contains length */
#define HRF_USER   0x8000       /* user defined */
    long subfmt;                /* either a dimension or qualifier (date style, say)*/
#define HRSF_DEC   0            /* int types */
#define HRSF_HEX   1
#define HRSF_OCT   2
    int (* vrfy) ();            /* function that verfies input */
    union                       /* the converted value */
    {
        long ival;              /* integer value */
        double dval;            /* float value */
        char * tval;            /* text value (same as val above) */
        void * uval;            /* user value -- prolly some struct */
        char * cval;            /* char/binary value */
    } cvtval;
} hrecelm;


typedef struct {
    hrecelm * e;
    char * name;
    int errno;
    char * errmsg;
} hrecerr;


typedef struct 
{
    char proto [20];
    char host [256];
    int port;
    char path [1024];
    char file [1024];
} URL;
    

/* these must be included to get the prototype to trabajaba */


#include "html_proto.h" 
#include "hparse_proto.h"
#include "hutil_proto.h"
#include "hrec_proto.h"
#include "hsql_proto.h"
#include "hdbm_proto.h"
#include "htable_proto.h"
#include "hform_proto.h"
#include "hlogin_proto.h"
#include "hcond_proto.h"
#include "hcrawl_proto.h"
#include "sock_proto.h"

#endif
