
/*
 *                 Copyright (c) Mar 11 16:43:50  MTCC
 *	Author: Michael Thomas
 *	Module: hutil.c
 *	Created: Tue Mar 11 16:43:50 1997
 *	Abstract:
 *	   utility function for html digester
 */

/* Edit History: 
 */

#include "html.h"

htelm * htcur = 0, * hthead = 0;


/*
 * there is method to the naming madness here:
 * The ht functions track the concept of a
 * current pointer, whereas htp functions have
 * no idea about this and always demand a list
 * to traverse
 */

static htelm * 
hteol (htelm * h) 
{
    if (h->slink)
        return hteol (h->slink);    
    for (; h->flink; h = h->flink);
    return h;
}

/* add a file to the current page */

htelm * htreadfile (const char * file) 
{
    htelm *pend, * h = htpreadfile (htcur, &pend, file);

    if (h == 0)
        return 0;
    if (hthead == 0)
        hthead = h;
    htcur = pend;
    return h;
}

htelm * htreadurl (const char * file, int * sts, char * type) 
{
    htelm *pend, * h = htpreadurl (htcur, &pend, file, sts, type);

    if (h == 0)
        return 0;
    if (hthead == 0)
        hthead = h;
    htcur = pend;
    return h;
}


htelm * htreadfiletext (const char * file) 
{
    htelm *pend, * h = htpreadfiletext (htcur, &pend, file);

    if (h == 0)
        return 0;
    if (hthead == 0)
        hthead = h;
    htcur = pend;
    return h;
}


/*
 * dump the contents to stdout
 */

void htdumpfile () 
{
    htpdumpfile (hthead);
}

void htdumptree () 
{
    htpdumptree (hthead, 0);
}


void htrelease () 
{
    htprelease (hthead);
    htcur = 0;
    hthead = 0;
}



/*
 * add a markup element after h. this useful in congunction
 * with markers.
 */

htelm * htpaddelm (htelm * h, htelm **pend, const char * txt) 
{
    htelm * p;
    if ((p = htpreadstr (h, pend, txt)) == 0)
        return 0;
    return p;
}

htelm * htaddelm (const char * txt) 
{
     htelm * pend, * h = htpaddelm (htcur, &pend, txt);
     if (hthead == 0)
         hthead = h;
     if (h)
         htcur = pend;
     return h;
}



/*
 * goto a particular named tag including meta tags 
 */

htelm * htpgototag (htelm * h, const char * tag) 
{
    htelm * p;
    
    for (; h; h = h->flink) {
        if (h->type != H_TXT) {
            if (h->tk && ! strcasecmp (h->tk->name, tag))
                return h;
        }
        if (h->slink && (p = htpgototag (h->slink, tag)))
            return p;
    }
    return 0;
}

htelm * htgototag (const char * tag) 
{
    htelm * h = htpgototag (hthead, tag);
    if (h)
        htcur = h;
    return h;    
}



htelm * htpgotomark (htelm * h, const char * mark) 
{
    htelm * p;
    
    for (; h; h = h->flink) {
        if (h->type == H_MARKER) {
            struct token * p = htkgetword (h->tk, 1);
            if (! strcmp (p->name, mark)) {
                return h/*->flink*/;
            }            
        }
        /* traverse the tree */
        if (h->slink && (p = htpgotomark (h->slink, mark)))
            return p;
    }
    
    return 0;
}

htelm * htgotomark (const char * mark) 
{
    htelm * h = htpgotomark (hthead, mark);
    if (h)
        htcur = h;
    return h;    
}


htelm * htpgotoline (htelm * h, int line) 
{
    htelm * p, * lastp = 0;
    
    for (; h; h = h->flink) {
        if (h->lineno == line)
            return h;
        if (h->slink && (p = htpgotoline (h->slink, line)))
            return h;
        /* we've gone past */
        if (h->lineno > line)
            if (h->blink)
                return h->blink;
            else                /* last ditch */
                return h;
        lastp = h;
    }
    /* last, last ditch */
    return lastp;    
}

            

htelm * htgotoline (int line) 
{
    htelm * h = htpgotoline (hthead, line);
    if (h)
        htcur = h;
    return h;    
}

    
int htprintf (const char * fmt, ELLIPSE) 
{
    va_list ap;
    char bf [2048];
    int rv;
    
    
    VA_START (ap, fmt);
    rv = vsprintf (bf, fmt, ap);
    va_end (ap);
    htaddelm (bf);
    return rv;
}

int hthdrprintf (const char * fmt, ELLIPSE) 
{
    va_list ap;
    htelm * h;
    char bf [1024];
    int rv;
    
    VA_START (ap, fmt);
    rv = vsprintf (bf, fmt, ap);
    va_end (ap);
    h = htpgotomark (hthead, "mime-headers");
    if (h)
        htpaddelm (h, 0, bf);
    return rv;
}


/*
 * like htprintf, but keeps appending to an internal buffer
 * until htflush () is called. This is mainly because the
 * parser is too stupid to do anything intelligent with
 * separate calls like: htprintf ("<blah"); htprintf(">");
 */

static char * htbf = 0;
static int htbfsz = 0;

int htfprintf (const char * fmt, ELLIPSE) 
{
    va_list ap;
    int len = 0, rv;

    if (htbf == 0) {
        htbf = malloc (1024);
        htbfsz = 1024;
    } else if ((len = strlen (htbf))+512 >= htbfsz) {
        htbfsz += 512;
        htbf = realloc (htbf, htbfsz);
    }    
    VA_START (ap, fmt);
    rv = vsnprintf (htbf+len, htbfsz-len, fmt, ap);
    va_end (ap);
    return rv;
}

void htflush () 
{
    if (htbf) {
        htaddelm (htbf);
        if (htbfsz > 1024) {
            htbf = realloc (htbf, 1024);
            htbfsz = 1024;
        }
        * htbf = 0;
    }    
}


int htpprintf (htelm * h, const char * fmt, ELLIPSE) 
{
    va_list ap;
    htelm * pend;
    char bf [1024];
    int rv;
    
    
    VA_START (ap, fmt);
    rv  = vsprintf (bf, fmt, ap);
    va_end (ap);
    htpaddelm (h, &pend, bf);
    return rv;
}


htelm * htgetcur () 
{
    return htcur;
}


htelm *
htgetpage () 
{
    return hthead;
}

void hteof () 
{
    for (; htcur && htcur->flink; htcur = htcur->flink);
}

void htrewind () 
{
    htcur = hthead;
}

void hterase () 
{
    htrelease ();
}



/*
 *  return a vector of strings split apart
 *  using sep as the delimiter ala perl.
 */

char ** split (const char * str, char sep) 
{
    const char * p;
    char ** rv;
    int n = 1;

    for (p = str; * p; p++)
        if (* p == sep)
            n++;
    if (n == 1)
        return 0;
    rv = calloc (n, sizeof (* rv));
    n = 0;
    for (p = str; * p; p++) {
        if (* p == sep) {
            rv [n] = calloc (p-str+1, 1);
            memcpy (rv [n], str, p-str);
            str = p+1;
            n++;
        }
    }
    if (p != str) {
        rv [n] = calloc (p-str+1, 1);
        memcpy (rv [n], str, p-str);
    }
    return rv;
}

/*
 * truncate the last byte of a string ala perl
 */

void chop (char * str) 
{
    if (* str)
        str [strlen (str)-1] = 0;
}


void trim (char * str)
{
    char * p;

    for (p = str+strlen (str)-1; p > str; p--)
        if (isspace (* p))
            * p = 0;
        else
            break;
}

    
/* unwebify stupid %xx into normal characters */

void unwebify (const char * word, char * bf) 
{
    char hval [3];
    int val;
    
    while (* word) {
        /* escaped character */
        if (* word == '\\') {
            word++;
            if (* word == 0)
                break;
            * bf++ = * word++;
        } else if (* word == '%') {
            word++;
            hval [0] = * word++;
            hval [1] = * word++;
            hval [2] = 0;
            sscanf (hval, "%x", &val);
            * bf++ = val;
        } else
            * bf++ = * word++;
    }
    * bf = 0;
}

char *
webify (const char * str) 
{
    char * rv = malloc (strlen (str) * 3 + 1), * p;
    
    for (p = rv; * str; str++) {
        /* be extremely conservative. I guess. */
        if (* str <= ' ' || * str >= 0177 ||
            * str == '&' || * str == '\"' ||
            * str == '<' || * str == '>' ||
            * str == '=' || * str == '?') {
            sprintf (p, "%%%02x", * str);
            p += 3;
        } else
            * p++ = * str;
    }
    * p = 0;
    
    return rv;
}

char *
webquote (const char * str) 
{
    static char * rv;
    static int sz = 0;
    char * p;

    if (sz < strlen (str)*6+1) {
        if (rv)
            free (rv);
        sz = strlen (str)*6+1;
        rv = malloc (sz);
    }    
    for (p = rv; * str; str++) {
        /* be extremely conservative. I guess. */
        if (* str <= ' ' || * str >= 0177 ||
            * str == '&' || * str == '\"' ||
            * str == '<' ||* str == '>') {
            sprintf (p, "&#%03d;", * str);
            p += 6;
        } else
            * p++ = * str;
    }
    * p = 0;
    
    return rv;
}


static void stripdotdot (char * path) 
{
    char * p;
    for (p = path; * p;) {
        if (* p == '.') {
            if (p [1] == '/') {
                /* useless ./ */
                strcpy (p, p+2);
            } else if (p [1] == '.' && p [2] == '/') {
                char * t;
                if (p-1 >= path && p [-1] == '/')
                    t = p-2;
                else
                    t = p-1;
                for (; t > path; t--)
                    if (* t == '/')
                        break;
                if (t >= path)
                    strcpy (t, p+2);
                p = t;
            } else
                p++;
        } else if (p [0] == '/' && p [1] == '/') {
            /* gobble up the // in  /foo//bar/baz */
            strcpy (p, p+1);
            p++;
        }
        else
            p++;
    }
}



/*
 * parse a URL into its components. I really should write
 * a regular expression parser for this.
 */

int parseurl (const char * url, URL * out, URL * base) 
{
    
    char nurl [1024], * p;
    char host [256];
    
    if (url == 0)
        url = "/";
    out->port = 80;

    /* special case of relative path name: use base url */    
    if (strchr (url, ':') == 0 && base) {
        if (* url == '/')
            sprintf (nurl, "%s://%s%s", base->proto, base->host, url);
        else {
            if (* url == 0 || * url == '#')
                sprintf (nurl, "%s://%s%s%s%s", base->proto, base->host, base->path,
                     base->file, url);
            else
                sprintf (nurl, "%s://%s%s%s", base->proto, base->host, base->path,
                     url);
            url = nurl;
        }
    } else if (* url == 0)
        url = "/";    
    if ((p = strchr (url, ':'))) {
        strncpy (out->proto, url, p-url);
        out->proto [p-url] = 0;
        url = p+1;
    } else
        strcpy (out->proto, "http");
    if (! strncmp (url, "//", 2)) {
        url += 2;
        /* optional port number? */
        if ((p = strchr (url, ':'))) {
            sscanf (p+1, "%d", &out->port);
            strncpy (out->host, url, p-url);
            out->host [p-url] = 0;
            if ((p = strchr (url, '/')))
                url = p;
            else
                url += strlen (url);
        } else if ((p = strchr (url, '/'))) {
            strncpy (out->host, url, p-url);
            out->host [p-url] = 0;
            url = p;
        } else {
            strcpy (out->host, url);
            /* goto end of string. */
/*            url += strlen (url);*/
            url = "/";
        }        
    } else {
        gethostname (host, 256);
        strcpy (out->host, host);
    }
    strcpy (out->path, url);
    stripdotdot (out->path);    /* strip any ./ and ../ in the path now */

    /* deal with parameters */
    if ((p = strchr (out->path+1, '?')) == 0) {
        /* place p at the last non null char. not guaranteed
           to deal with zero, but the for loop checks against
           that. */
        p = out->path+strlen (out->path)-1;
    } else
        p--;
    
    /* sort of a strrchr to find the last '/' in the real path*/
    for (; p >= out->path+1; p--)
        if (* p == '/')
            break;
    if (* p != '/' || p == out->path)
        p = 0;        
    
    /* /path/... or /path */            
    if (p) {
        p++;
         strcpy (out->file, p);
         * p = 0;
    } else {
        /* if /~mike or /file.html */
        if (out->path [1] != '~') {    
            strcpy (out->file, out->path+1);
            out->path [1] = 0;
        } else {
            strcat (out->path, "/");
            out->file [0] = 0;
        }        
    }
    /* this is extremely sketchy. it should be something, but
       that something is a function of the HTTPD, not us. Punt
       for now; it means we won't always distinguish
       /foo/bar/ from /foo/bar/index.htm or somesuch. oh well.
     */
    if (out->file [0] == 0)
        strcpy (out->file, "index.html");
    if (strpbrk (out->path, " \t\n") || strpbrk (out->file, " \t\n"))
        return -1;
    return 0;
}

       
            

char * canonicalurl (URL * u) 
{
    char * rv;

    if (! strcmp (u->proto, "http") || ! strcmp (u->proto, "ftp")) {
        rv = malloc (strlen (u->proto)+3+strlen (u->host)+7+1+
                     strlen (u->path)+1+ strlen (u->file)+1);
        if (u->port == 80)
            sprintf (rv, "%s://%s%s%s", u->proto, u->host, u->path, u->file);
        else
            sprintf (rv, "%s://%s:%d%s%s", u->proto, u->host, u->port, u->path, u->file);
    } else {
        rv = malloc (strlen(u->proto)+1+strlen(u->file)+1);
        sprintf (rv, "%s:%s", u->proto, u->path);
    }
    return rv;
}

    
