/*
 *                 Copyright (c) Apr 19 10:16:52  MTCC
 *	Author: Michael Thomas
 *	Module: hcrawl.c
 *	Created: Sat Apr 19 10:16:52 1997
 *	Abstract:
 *	   web crawling interface
 */

/* Edit History: 
 */

package mtcc.libht;

import mtcc.libht.*;
import java.lang.*;
import java.io.*;
import java.net.*;
import java.util.*;


class flist 
{
    String url;
    flist flink;
    flist stack;
    String parent;
    int visited;
    int type;
}



public abstract class hcrawl
{
    // accessible fields for the abstract action routine
    String cururl;
    String  curparent;
    String curcontent;
    int curlevel;
    hparse curfile;
    int cursts;
    
private
    flist files, pfiles;
private
    flist stack = null;
private
    hURL baseurl;
public static final int INTRASITE = 0;         /* visit intrasite nodes once only */
public static final int INTRAFULL = 1;         /* view the entire tree, intrasite; some nodes
                                   may be visited more than once */
public static final int EXTRAFIRST = 2;        /* touch the first extranode site */
public static final int EXTRAFULL = 3;         /* loose the hounds */

public static final int LEAFVISIT = 0x10000;   /* normally we don't visit leaves outside of us */

    // empty constructor
public
    hcrawl () 
    {}



private
    int slevel (int init) 
    {
        flist p, last = stack;
        int i;
        if (stack == null)
            return 0;
        else
            if (stack.stack == null)
                return init;
    
        for (i = init, p = stack.stack; p != null; p = p.stack) {
            /* ignore dummied up leaf nodes. */
            if (p.visited == 2)
                continue;
            if (! last.parent.equals (p.parent))
                i++;
            last = p;
        }
        return i;
    }

/*
 * conditionally insert the given url into the file
 * list. some day it will pay attention to depth,
 * which controls how far we've traverse: rooted,
 * on site, and off site.
 */
    

private
    flist insfile (String url, String parurl, int depth) 
    {
        flist fl, p;
        int rv = 0, visited = 0, type;
        hURL u, par;
        int ext;

        if (parurl != null) {
            par = new hURL (parurl, null);
        } else {
            par = baseurl;
            parurl = url;
        }
        
        u = new hURL (url, par);
        if (! u.proto.equals ("http"))
            return null;

        /* ??? Should I or should I not traverse these. no for now. */
        if ((ext = u.file.indexOf ('#')) >= 0)
            u.file = u.file.substring (0, ext);
        

        /* figure out what type of node url is. */
        if (! par.host.equals (baseurl.host))
            type = EXTRAFULL;
        else {
            /* if this node is off site, or it's outside of the
               base url push it, but flag it as such */
            if (! u.host.equals (baseurl.host) ||
                ! u.path.regionMatches (0, baseurl.path, 0, baseurl.path.length ())) {
                type = EXTRAFIRST;
            } else {
                /* at this point we know we're intrasite, but 
                   for formating reasons, we might choose to
                   consider this site external */
                type = INTRASITE;
                if (! par.path.regionMatches (0, u.path, 0, par.path.length ())) {
                    /* it's an on site, but not a proper subset of
                       the parent's path. if this is the first time
                       we've seen this node, push it as a normal node,
                       otherwise it's a leaf. */
                    for (p = files; p != null; p = p.flink)
                        if (p.url.equals (url)) {
                            type = EXTRAFIRST;
                            break;
                        }
                }
                // after all is said and done, scripts are always
                // considered "extrafirst", ie note it, but don't
                // traverse it.
                if (u.parameters != null)
                    type = EXTRAFIRST;
            }
        }    
        if (type > depth)
            return null;
        if (type == EXTRAFIRST)
            visited = 2;
        url = u.toString ();
        /* check to see if we've visited this node before, but
           only if it doesn't appear so */
        if (visited == 0) {
            for (p = files; p != null; p = p.flink) 
                if (p.url.equals (url)) {
                    if (depth > INTRASITE) {
                        visited = 1;
                        break;
                    }
                    return null;
                }
        }
        fl = new flist ();
        fl.url = url;
        fl.flink = null;
        fl.visited = visited;
        fl.type = type;
        fl.parent = parurl;
        fl.stack = stack;
        stack = fl;

        if (pfiles == null) {
            files = fl;
            pfiles = fl;
        } else {
            pfiles.flink = fl;
            pfiles = fl;
        }
        return fl;
    }


private
    void _crawl (htelm p, String parent, int depth) 
    {
        String f;
        flist fl;

        for (; p != null && p.flink != null; p = p.flink);
        for (; p != null; p = p.blink) {
            if (p.type == htelm.H_ANCHOR &&
                (f = p.tk.getval ("href")) != null)
                insfile (f, parent, depth);
            if (p.slink != null)
                _crawl (p.slink, parent, depth);
        }        
    }


    // the action routines must be implemented by the caller

    abstract int action ();
    

    /*
 * crawl through the url space starting at url
 * the action routine is the user interface
 * which is called with a bunch of parameters,
 * but mainly the url of the file, and its
 * parsed contents. The action routine can
 * cut the crawl short by returning 0.
 */

    void alrm (int ctx)
    {
    }


public
    int crawl (String url, int depth) 
    {
        hparse p = new hparse ();
        int dotrav;
        int init = 0;
        int flg = depth & 0xffff0000;
    
        depth &= 0xffff;
        baseurl = new hURL (url, null);
        insfile (url, null, depth);    
        while (stack != null) {            
            cururl = stack.url;
            curparent = stack.parent;
            curlevel = slevel (init);
            curfile = null;
            cursts = 0;
            curcontent = "content/unknown";
            
            /* a leaf visit forces the file to be read in even if
               it's off site. This is used for a broken link search.
               */
            if (((flg & LEAFVISIT) != 0 && stack.visited == 2) ||
                (stack.type == INTRASITE && stack.visited == 0)) {

                /* read it in */
//                old  = signal (SIGALRM, alrm);
//                alarm (30);
                try {
                    p.parseURL (cururl);
                    curfile = p;                    
                } catch (IOException e) {
                    System.out.println ("shit happened "+e.toString());
                    
                    p.status = 500;
                    p.contenttype = "content/unknown";                    
                }
                curcontent = p.contenttype;
                cursts = p.status;                
                
//                alarm (0);
//                signal (SIGALRM, old);
                /* release unless a full scale crawl; we just want status */
                if (stack.type != INTRASITE && stack.type != EXTRAFULL) {
                    p.erase ();
                    curfile = null;
                } 

            }
            if (stack.visited != 0) {
                if (depth > INTRASITE)
                    dotrav = action (); 
                stack = stack.stack;
                p.erase ();
                continue;
            }

            /* let the user have at it. */        
            dotrav = action (); 
            stack.visited = 1;
            stack = stack.stack;   /* pop the stack */
            if (stack == null) init++;
            if (dotrav != 0)
                _crawl (p, cururl, depth);
            p.erase ();
        }
        return 0;
    }



}
