
/*
 *                 Copyright (c) Mar 11 16:43:50  MTCC
 *	Author: Michael Thomas
 *	Module: hdoc.java
 *	Created: Tue Mar 11 16:43:50 1997
 *	Abstract:
 *	   utility function for html digester
 */

/* Edit History: 
 */

package mtcc.libht;

import mtcc.libht.*;
import java.lang.*;
import java.io.*;





class hdoc extends htelm
{
    StringBuffer htbf = null;
    public htelm head;
    public htelm cur;
    public int status;
    public String contenttype;
    public String curfile;
    protected int lineno = 1;    
   

    // the constructor 
public
    hdoc () 
    {
        super (H_TXT, 1);
        this.curfile = "Document Root";
        this.head = this;
        this.cur = this;
    } 

/*
 * add a markup element after h. this useful in congunction
 * with markers.
 */
public
    void addString (String txt) 
    {
        hparse h = new hparse ();

        h.ulink = cur.ulink;
        try {
            h.parseString (txt);
            // absorb all errors, I guess.
        } catch (IOException e) {
            System.out.println ("adds: "+e.toString());
            
            return;
        }

        // now splice it into the parent node
        // don't bother linking it in if it didn't
        // produce anything.
        if (h.cur != null) {
            if (cur.flink != null) {                
                h.cur.flink = cur.flink;
                cur.flink.blink = h.cur;
            }
            cur.flink = h;        
            h.blink = cur;
            h.ulink = cur.ulink;
            this.cur = h.cur;
        }        
    }

public
    void addFile (String txt) throws IOException
    {
        hparse h = new hparse ();
        h.ulink = cur.ulink;
        h.parseFile (txt);

        // now splice it into the parent node
        // don't bother linking it in if it didn't
        // produce anything.
        if (h.cur != null) {
            if (cur.flink != null) {
                h.cur.flink = cur.flink;
                cur.flink.blink = h.cur;
            }
            cur.flink = h;        
            h.blink = cur;
            h.ulink = cur.ulink;
            this.cur = h.cur;            
        }        
    }

public
    void addURL (String txt) throws IOException
    {
        hparse h = new hparse ();

        h.ulink = cur.ulink;
        h.parseURL (txt);

        // now splice it into the parent node
        // don't bother linking it in if it didn't
        // produce anything.
        if (h.cur != null) {
            if (cur.flink != null) {
                h.cur.flink = cur.flink;
                cur.flink.blink = h.cur;
            }
            cur.flink = h;        
            h.blink = cur;
            h.ulink = cur.ulink;
            this.cur = h.cur;
        }        
    }
    


/*
 * goto a particular named tag including meta tags 
 */
public
    htelm gototag (String tag) 
    {
        return gototag (this.head, tag);
    }
    
public
    htelm gototag (htelm h, String tag) 
    {
        htelm p;        
        
        for (; h != null; h = h.flink) {
            if (h.type != htelm.H_TXT) {
                if (h.tk != null && tag.equalsIgnoreCase (h.tk.name))
                    return h;
            }
            if (h.slink != null && (p = gototag (h.slink, tag)) != null)
                return p;
        }
        return null;
    }



public
    htelm gotomark (htelm h, String mark) 
    {
        htelm p;
        
        for (; h != null; h = h.flink) {
            if (h.type == htelm.H_MARKER) {
                token t = h.tk.getword (1);
                if (t != null && mark.equals (t.name)) {
                    return h;
                }            
            }
            /* traverse the tree */
            if (h.slink != null && (p = gotomark (h.slink, mark)) != null)
                return p;
        }
        
        return null;
    }

public
    htelm gotomark (String  mark) 
    {
        htelm h = gotomark (this.head, mark);
        if (h != null)
            this.cur = h;
        return h;    
    }

public
    htelm gotoline (htelm h, int line) 
    {
        htelm p, lastp = null;
        
        for (; h != null; h = h.flink) {
            if (h.lineno == line)
                return h;
            if (h.slink != null && (p = gotoline (h.slink, line)) != null)
                return h;
            /* we've gone past */
            if (h.lineno > line)
                if (h.blink != null)
                    return h.blink;
                else                /* last ditch */
                    return h;
            lastp = h;
        }
        /* last, last ditch */
        return lastp;    
    }

            

    htelm gotoline (int line) 
    {
        htelm h = gotoline (this.head, line);
        if (h != null)
            this.cur = h;
        return h;    
    }

    //
    // c'est la grande snorte, I know, calling these xxprintf
    // but it most hystorical.
    
public
    void printf (String str) 
    {
        addString (str);
    }

public
    void hdrprintf (String str) 
    {
        htelm h;
        htelm sv = this.cur;
        
        h = gotomark (this.head, "mime-headers");
        if (h != null)
            this.addString (str);
        this.cur = sv;
    }


/*
 * like htprintf, but keeps appending to an internal buffer
 * until htflush () is called. This is mainly because the
 * parser is too stupid to do anything intelligent with
 * separate calls like: htprintf ("<blah"); htprintf(">");
 */

public
    void fprintf (String str) 
    {
        if (htbf == null)
            htbf = new StringBuffer ();
        htbf.append (str);
    }

public
    void flush () 
    {
        if (htbf != null) {
            addString (htbf.toString ());
            htbf.setLength (0);
        }    
    }



public
    htelm getcur () 
    {
        return this.cur;
    }


public
    htelm getpage () 
    {
        return this.head;
    }

public
    void eof () 
    {
        for (; this.cur != null && this.cur.flink != null;
             this.cur = this.cur.flink);
    }

public
    void rewind () 
    {
        this.cur = this.head;
    }
    
public
    void erase () 
    {
        this.head = this;
        this.cur = this;
        this.flink = this.slink = this.ulink = null;
    }
    


/*
 *  return a vector of strings split apart
 *  using sep as the delimiter ala perl.
 */

public static
    String [] split (String str, char sep) 
    {
        String rv [];
        int n = 0, i, lasti = 0;

        for (i = 0; i < str.length(); i++)
            if (str.charAt (i) == sep)
                n++;
        if (n == 0)
            return null;
        rv = new String [n];
        n = 0;
        for (i = 0; i < str.length (); i++) {
            if (str.charAt (i) == sep) {
                if (i > lasti) {
                    rv [n++] = str.substring (lasti, i-1);
                    lasti = i+i;
                } else
                    rv [n++] = "";                
            }
        }
        return rv;
    }

/*
 * truncate the last byte of a string ala perl
 */

public static
    String chop (String str) 
    {   
        return str.substring (0, str.length()-1);
    }
    

    
/* unwebify stupid %xx into normal characters */

public static
    String unwebify (String word) 
    {
        int i;
        StringBuffer bf = new StringBuffer ();
        int len = word.length ();
        char ch;
    
        for (i = 0; i < len;) {
            ch = word.charAt (i);
            /* escaped character */
            if (ch == '\\') {
                i++;
                if (i >= len)
                    break;
                bf.append (ch);
            } else if (ch == '%') {
                bf.append ((char) Integer.parseInt (word.substring (i+1, i+3), 16));
                i += 3;
            } else {
                bf.append (ch);
                i++;
            }
        }
        return bf.toString ();
    }

public static
    String webify (String str) 
    {
        StringBuffer rv = new StringBuffer ();
        String s;
        int i;
        char ch;
    
        for (i = 0; i < str.length (); i++) {
            ch = str.charAt (i);
            /* be extremely conservative. I guess. */
            if (ch <= ' ' || ch >= 0177 ||
                ch == '&' || ch == '\"' ||
                ch == '<' || ch == '>' ||
                ch == '=' || ch == '?') {
                // this is so damn gross. where the fuck is sprintf???
                rv.append ('%');
                s = Integer.toHexString ((int) ch);
                if (s.length () < 2)
                    rv.append ('0');
                rv.append (s);
            } else
                rv.append (ch);
        }
        return rv.toString ();
    }

public static
    String webquote (String str) 
    {
        StringBuffer rv = new StringBuffer ();
        int i;
        char ch;

        for (i = 0; i < str.length (); i++) {
            ch = str.charAt (i);
        
            /* be extremely conservative. I guess. */
            if (ch <= ' ' || ch >= 0177 ||
                ch == '&' || ch == '\"' ||
                ch == '<' || ch == '>') {
                // really should be %03d, I think 
                rv.append ("&#"+Integer.toString((int) ch)+";");
            } else
                rv.append (ch);
        }
        return rv.toString ();
    }

}
