/*
 *                 Copyright (c) Apr  4 13:19:14  MTCC
 *	Author: Michael Thomas
 *	Module: hform.c
 *	Created: Fri Apr  4 13:19:14 1997
 *	Abstract:
 *	   form manipulation functions.
 */

/* Edit History: 
 */
package mtcc.libht;

import java.lang.*;
import java.io.*;
import mtcc.libht.*;

public class hform extends hparse
{
    
    static int curform = 0;

    /*
     * get the n'th form
     */

private
    htelm _formn (htelm h, int n []) 
    {
        htelm rv;

        for (; h != null; h = h.flink) {
            if (h.type == H_FORM && (h.tk.flags & token.TK_ISEND) == 0) {
                if (n [0] == 0)
                    return h;
                n [0]--;
            }
            if (h.slink != null && (rv = _formn (h.slink, n) )!= null)
                return rv;        
        }
        return null;
    }

public
    htelm formn (int n) 
    {
        int pn [] = new int [1];
        htelm h;
        
        if (n < 0)
            return null;
        pn [0] = n;
        h = _formn (head, pn);
        if (h != null) {
            cur = h;
            curform = n;
        }
        return h;    
    }

public
    htelm formprev () 
    {
        return formn (curform-1);
    }

    
public
    htelm formnext () 
    {
        htelm h;
    
        h = formn (curform);
        if (h != null)
            curform++;
        return h;
    }

public
    htelm gotoform (String fname) 
    {
        htelm h;
        int pn [] = new int [1];
        pn [0] = 0;
        gotomark (fname);
        if ((h = _formn (cur, pn)) != null)
            cur = h;
        return h;
    }



/*
 * retrieve the field within the form/whatever pointed to in
 * h
 */
public
    htelm gotofield (String field) 
    {
        htelm h;
        String p;
        
        for (h = this.cur.slink; h != null; h = h.flink) {
            p = h.tk.getval ("name");
            if (p != null && field.equalsIgnoreCase (p))
                return h;
        }
        return null;
    }



    String getval (String field, String var) 
    {
        htelm h;
        
        h = gotofield (field);
        if (h == null)
            return null;    
        return h.tk.getval (var);
    }

    

        
public
    void putval (htelm h, String var, String val) 
    {
        token tk, p, pp;
    
        if ((tk = h.tk.ckvar (var)) != null) {
            /* already present */
            if (val == null)
                return;
            if (tk.flink != null && (tk.flink.flags & token.TK_VAL) != 0) {
                tk.flink.name = val;
            } else {
                // splice into the cdr of the list
                pp = new token ();
                pp.flags = token.TK_VAL;
                pp.name = val;
                pp.flink = tk.flink;
                tk.flink = pp;
            }
        } else {
            // both missing
            p = new token ();
            p.flags = token.TK_VAR;
            p.name = var;
            if (val != null) {
                pp = new token ();
                pp.flags = token.TK_VAL;
                pp.name = val;
                p.flink = pp;
            } else
                pp = p;
            // now splice in after the first element
            pp.flink = h.tk.flink;
            h.tk.flink = p;
        }
    }


public
    void puttag (htelm h, String var) 
    {
        putval (h, var, null);
    }


public
    void deltag (htelm h, String var) 
    {
        token tk, p, pp;

        /* only if not already present */
        if ((tk = h.tk.ckvar (var)) != null) {        
            for (p = h.tk; p.flink != null; p = p.flink)
                if (p.flink == tk) {
                    p.flink = p.flink.flink;
                    return;
                }
        }
    }
public
    void putftag (String field, String var) 
    {
        htelm h;
        if ((h = gotofield (field)) == null)
            return;
        putval (h, var, null);
    }


public
    void putfval (String field, String var, String val)
    {
        htelm h;
        if ((h = gotofield (field)) == null)
            return;
        putval (h, var, val);
    }



    private void _bindform (htelm h, hrec r) 
    {
        htelm s;
        hparse newh;
        String p, o, val;

        /* don't traverse down a nested form */
        if (h.type == H_FORM)
            return;
    
        for (; h != null; h = h.flink) {
            /* if there is no name field, or if this hrec has
               nothing to say about the field, mosey along. */
            if (h.tk == null ||
                (p = h.tk.getval ("name")) == null ||
                (val = r.isvis (p)) == null) {
                if (h.slink != null)
                    _bindform (h.slink, r);
                continue;
            }
            switch (h.type) {
                case H_SELECT:
                    for (s = h.slink; s != null; s = s.flink) {
                        if (s.type == H_OPTION) {
                            /* if there is a value field in the
                               option, use it. */
                            if ((o = s.tk.getval ("value")) != null) {
                                if (o.equals (val))
                                    puttag (s, "selected");
                            } else {
                                /* else just compare the adjacent text */
                                if (s.flink != null && s.flink.type == H_TXT) {
                                    /* don't consider a newline in the
                                       body to be significant. */
                                    if (val.equals (s.flink.body.toString ().trim ()))
                                        puttag (s, "selected");
                                }
                            }                                
                        }
                    }
                    break;
                case H_TXTAREA:
                    /* the car of a textarea is its payload */
                    s = cur;
                    cur = h.slink;
                    addString (val);
                    cur = s;                    
                    break;
                case H_CHECK:
                case H_RADIO:
                    /* see if there is a name field for this input */
                    if ((o = h.tk.getval ("value")) != null
                        && o.equals (val)) {
                        puttag (h, "checked");
                    }
                    break;
                case H_INPUT:
                case H_PASSWORD:
                case H_HIDDEN:
                    /* see if there is a name field for this input */
                    putval (h, "value", hdoc.webquote (val));
                    break;
                default:
                    break;
            }
            if (h.slink != null)
                _bindform (h.slink, r);
        }
    
    }

/*
 *  bind the contents of a hrec to the current form
 *  you must be sitting on the current form for anything
 *  to happen
 *
 */
public
    void bindform (hrec r) 
    {
        if (this.cur.type != H_FORM) {
            return;
        }
        _bindform (this.cur.slink, r);
    
    }
}
