/*
 *                 Copyright (c) Apr  7 13:32:17  MTCC
 *	Author: Michael Thomas
 *	Module: hlogin.c
 *	Created: Mon Apr  7 13:32:17 1997
 *	Abstract:
 *	   login related functions
 */

/* Edit History: 
 */

#include "html.h"

#ifdef WANTLOGIN

struct cgilogin
{
    char * seed;                 /* the encrypted password seed */
    long timeout;               /* when this record expires */
    char * facility;            /* the facility (program) */
    int sock;                   /* database socket */
};



static struct cgilogin us;


hrec * newloginrec () 
{
    hrec * h = newhrec ();
    long tmo;
    char * host;

    time (&tmo);
    if ((host = getenv ("REMOTE_HOST")) == 0)
        host = "";
    hradd (h, "sessid", "");
    hradd (h, "user", "");
    hradd (h, "pwd", "");
    hradd (h, "facility", us.facility);
    hraddint (h, "timeout", us.timeout + tmo);
    hradd (h, "host", host);
    hraddint (h, "hits", 1);
    return h;
}


/* connect up to the login database engine. currently this
 * uses msql, but the database needs of this puppy are
 * pretty minimal. A relational database does, of course,
 * give the ability to do cute things with statistical
 * analysis and that sort of cruft -- ie site hit analysis --
 * so it's probably not as much overkill as you think.
 */

int hinitlogin (char * server, char * facility, char * seed, int deftmo) 
{
    us.facility = facility;
    us.timeout = deftmo;
    us.seed = seed;
#if (LOGINTYPE == 0)
    us.sock = msqlConnect (server);
    if (us.sock < 0)
        return -1;
    if (msqlSelectDB (us.sock, LOGINDB) < 0)
        return -1;
#endif
#if (LOGINTYPE == 1)
    us.sock = hdbminit (LOGINDBFILE);
#endif
    return us.sock;
}

/*
 * lookup the encrypted password in a file ala /etc/passwd.
 * you can use /etc/passwd but other files don't need to
 * have the other fields. Returns 0 if found, -1 otherwise.
 */

int hpwdfile (char * file, char * user, char * pwd) 
{
    char bf [132];
    char * p, * p2;
    int rv = -1;
    
    FILE * fp = fopen (file, "r");
    while (fgets (bf, 132, fp) != NULL) {
        if (bf [0] == '#')
            continue;
        if ((p = strchr (bf, ':')) == 0)
            continue;
        * p++ = 0;
        if (strcmp (user, bf))
            continue;     
        if (p2 = strchr (p, ':'))
            * p2 = 0;
        strcpy (pwd, p);
        rv = 0;
        break;
    }
    fclose (fp);
    return rv;
}

/* the session identifier could convievably be something
 * as simple as the time of day, or a constantly incrementing
 * counter or something like that, but it ought to be
 * *somewhat* hard to guess. Of course, if people are really
 * worried about a hacker spoofing a session id, they should
 * be using SSL or the like so you can't just look at the
 * stream with a sniffer, but knowing that lots of cruft
 * just ain't, we should make it at least a little bit non
 * intuitive what it will be.
 */
/* fix me!!! */

static hrec * newsession (hrec * h) 
{
    static char bf [20];
    long tm;
    hrec * seq = newhrec ();
    

    hraddint (seq, "_seq", 0);
#if (LOGINTYPE == 0)
    /* this takes advantage of a msql feature called a sequence.
     * for other databases, this might be different, but worst
     * is having to do it by hand.
     */
    hmfind (us.sock, seq, LOGINDBTABLE, "", "");
#endif
#if (LOGINTYPE == 1)
    /* need to reproduce this... */
#endif
    time (&tm);
    hraddf (h, "sessid", "%s%d.%d",  us.facility, tm, hrgetint (seq, "_seq"));
    hrfree (seq);
    return h;
}


/*
 * login to an existing account. returns the session identifier
 */

hrec * hlogin (char * uname, char * pwd, char * encpwd) 
{
    char * pass = crypt (pwd, encpwd); /* uses seed in the encrypted */
    /* if they specified a password, and there  isn't one, that's an error */
    if (* pwd && * encpwd == 0)
        return 0;    
    /* if the encrypted password is null, don't check */
    if (* encpwd == 0 || ! strcmp (pass, encpwd)) {
        hrec * h = newloginrec ();
        /* must call newsession first so the db doesn't whine. */
        if (newsession (h) == 0)
            return 0;
        hradd (h, "user", uname);
        hradd (h, "pwd", encpwd);
#if (LOGINTYPE == 0)        
        if (hmupdate (us.sock, h, LOGINDBTABLE, "sessid", hrget (h, "sessid")) < 0) {
            hrfree (h);
            return 0;
        }
#endif
#if (LOGINTYPE == 1)
        if (hdbmupdate (hrget (h, "sessid"), h) == 0) {
            hrfree (h);
            return 0;
        }
#endif        
        return h;
    } else
        return 0;
}


/*
 * create a new login for this user if the two
 * passwords match. Returns the session id
 * and the encrypted password.
 */

hrec * hnewlogin (char * uname, char * pwd, char * pwd2, char * pencpwd)
{
    if (strcmp (pwd, pwd2))
        return 0;
    if (* pwd)
        strcpy (pencpwd, crypt (pwd, us.seed));
    else
        strcpy (pencpwd, "");
    return hlogin (uname, pwd, pencpwd);
}





/*
 * login using an admin password providing an override
 * mechanism.
 */

hrec * hadminlogin (char * uname, char * pwd, char * admin) 
{
    /* this should be the same as newlogin, but the
       admin pwd's aren't currently encrypted, so
       we'll punt for the time being. */
    if (strcmp (pwd, admin))
        return 0;
    else {
        hrec * h = newloginrec ();
        if (newsession (h) == 0)
            return 0;
        
        hradd (h, "user", uname);
        hradd (h, "pwd", pwd);
#if (LOGINTYPE == 0)
        if (hmupdate (us.sock, h, LOGINDBTABLE, "sessid", hrget (h, "sessid")) < 0) {
            hrfree (h);
            return 0;
        }
#endif
#if (LOGINTYPE == 1)
        if (hdbmupdate (hrget (h, "sessid"), h) < 0) {
            hrfree (h);
            return 0;
        }
#endif
        return h;
    }
}


void hloginbindall (hrec * in)
{
    hrec * new = newhrec ();
    htrewind ();
    hradd (new, "sessid", hrget (in, "sessid"));
    while (htformnext ())
        htbindform (new);
    hrfree (new);
}



/*
 * check to see if the session id is valid in the current
 * form.
 */

hrec * hvrfysession (hrec * in, char * user) 
{
    char * host = getenv ("REMOTE_HOST");
    hrec * hr;
    long tm;
    
    if (host == 0)
        host = "";    
    /* check to see if there is a session id and whether
       it matches anything in our database */
#if (LOGINTYPE == 0)
    hr = newloginrec ();
    if (hmfind (us.sock, hr, LOGINDBTABLE, "sessid", hrget (in, "sessid")) == 0) {
        hrfree (hr);
        return 0;
    }
#endif
#if (LOGINTYPE == 1)
    hr = newhrec ();
    if (hdbmfind (hrget (in, "sessid"), hr) == 0) {
        hrfree (hr);
        return 0;
    }

#endif
    if (strcmp (hrget (hr, "host"), host)) {
        hrfree (hr);
        return 0;
    }
    if (strcmp (hrget (hr, "facility"), us.facility)) {
        hrfree (hr);
        return 0;
    }
    
    time (&tm);
    if (hrgetint (hr, "timeout") < tm) {
        hrfree (hr);
        return 0;
    }
    if (user && strcmp (hrget (hr, "user"), user)) {
        hrfree (hr);
        return 0;
    }
    hraddint (hr, "hits", hrgetint (hr, "hits")+1);
    /* increment the hit count */
#if (LOGINTYPE == 0)
    if (hmupdate (us.sock, hr, LOGINDBTABLE, "sessid", hrget (hr, "sessid")) <= 0)
        ;                       /* uh, not sure. not the user's fault */
#endif
#if (LOGINTYPE == 1)
    if (hdbmupdate (hrget (hr, "sessid"), hr) <= 0)
        ;                       /* uh, not sure. not the user's fault */    
#endif
    return hr;        
}


/*
 * purge stale sessions
 */

int hloginpurge (int all) 
{
    long tm;
    time (&tm);
#if (LOGINTYPE == 0)
    if (all)
        return msqlQueryf (us.sock, "delete from login");
    else
        return msqlQueryf (us.sock, "delete from login where timeout < %d", tm);
#endif
#if (LOGINTYPE == 1)
    {
        void * ctx = 0;
        hrec * hr = newhrec ();
        int rv = 0;

        hradd (hr, "sessid", "");
        hraddint (hr, "timeout", 0);
        while (hdbmfetchrows (hr, &ctx, 0, 0)) {
            if (all || hrgetint (hr, "timeout") < tm)
                if (hdbmdelete (hrget (hr, "sessid")) < 0)
                    return -1;
                else {
                    rv++;
                }
        }
        hrfree (hr);
        return rv;
    }
#endif
}

#endif
