/*
 *                 Copyright (c) Mar 20 13:27:39  MTCC
 *	Author: Michael Thomas
 *	Module: hsql.java
 *	Created: Thu Mar 20 13:27:39 1997
 *	Abstract:
 *	   interface routines to msql
 */

/* Edit History: 
 */


package mtcc.libht;

import mtcc.libht.*;
import mtcc.msql.*;
import java.io.*;

public class hmsql extends hrec implements hdb
{
    msql mdb;
    protected msqlResult ctx = null;
    StringBuffer bf = new StringBuffer ();

    // the constructor, sets of the database
public
    hmsql (String host, String db) throws IOException
    {
        if (host == null || host.length () == 0)
            host = "localhost";
        mdb = new msql (host, db);
    }

    // basically a way to clone an already open
    // connection
public
    hmsql (hmsql old) 
    {
        mdb = old.mdb;
    }


public
    int find (String table, String field, String val) throws IOException
    {
        msqlResult r;
        String row [];
        int rv, col = 0;
        hrecelm e;

        bf.setLength (0);
        /* build the query string */
        bf.append ("select ");
        e = (hrecelm) this.first ();
        while (e != null) {
            if ((e.flags & hrec.HR_HIDDEN) != 0) {
                e = (hrecelm) this.next ();
                continue;
            }           

            bf.append (e.var+",");
            e = (hrecelm) next ();
        }
        /* kill the trailing comma */
        bf.setLength (bf.length()-1);
        bf.append (" from "+table+" ");
        if (field != null)
            bf.append (" where "+field+"='"+slashify (val)+"'");

        if ((rv = mdb.Query (bf.toString ())) == 0)
            return 0;
        /* if more than one row, use the first. */
        r = mdb.GetResult ();
        row = r.FetchRow ();
        col = 0;
        /* fill the hrec with the new data */
        e = (hrecelm) first ();
        while (e != null) {
            if ((e.flags & hrec.HR_HIDDEN) != 0) {
                e = (hrecelm) next ();
                continue;
            }
            if (row [col] != null)
                add (e.var, row [col++]);
            else {
                add (e.var, "");
                col++;
            }
            e = (hrecelm) next ();
        }
        return rv;
    }

public
    int firstrow (String table, String qual) throws IOException
    {
        hrecelm e;
        
        /* build the query string */
        bf.setLength (0);
        bf.append ("select ");
        e = (hrecelm) first ();
        while (e != null) {
            if ((e.flags & hrec.HR_HIDDEN) != 0) {
                e = (hrecelm) next ();
                continue;
            }
            bf.append (e.var+",");
            e = (hrecelm) next ();
        }
        /* kill the trailing comma */
        bf.setLength (bf.length ()-1);
        bf.append (" from "+table+" ");
        if (qual != null)
            bf.append (qual);
        if (mdb.Query (bf.toString ()) == 0)
            return 0;
        ctx = mdb.GetResult ();
        return nextrow ();
        
    }
    

public
    int nextrow ()
    {
        String row [];
        int rv = 0, col = 0;
        String v;
        hrecelm e;
    
        if ((row = ctx.FetchRow ()) == null) {
            ctx = null;
            return 0;
        }
        /* fill the hrec with the new data */
        e = (hrecelm) first ();
        while (e != null) {
            if ((e.flags & hrec.HR_HIDDEN) != 0) {
                e = (hrecelm) next ();
                continue;
            }
            if (row [col] != null)
                v = row [col++];
            else {
                v = "";
                col++;
            }
            /* type conversion is dealt with by hradd */
            add (e.var, v);
            e = (hrecelm) next ();
        }
        return 1;
    }


public
    int max (String table, String field) throws IOException
    {
        int rv = 0, n;
        msqlResult r;
        String row [];

        if (mdb.Query ("select "+field+" from "+table) <= 0)
            return -1;
        r = mdb.GetResult ();
        while ((row = r.FetchRow ()) != null) {
            n = Integer.parseInt (row [0]);
            if (n > rv)
                rv = n;
        }
        return rv;
    }

public
    String slashify (String str) 
    {
        StringBuffer rv = new StringBuffer ();
        int i;

        for (i = 0; i < str.length (); i++) {
            if (str.charAt (i) == '\'')
                rv.append ('\\');
            rv.append (str.charAt (i));
        }
        return rv.toString ();
    }


/*
 * update a record given the hrec. if it's not found,
 * one will be created.
 *
 */

protected static final int UPDINS = 0;
protected static final int INSONLY = 1;
protected static final int UPDONLY = 2;

protected 
    int _update (String table, String qual, int cmd) throws IOException
    {
        int rv;
        hrecelm e;
    
        bf.setLength (0);
        /* build the query string */
        if (cmd != INSONLY) {
            bf.append ("update "+table+" set ");
            e = (hrecelm) first ();
            while (e != null) {            
                if ((e.flags & hrec.HR_HIDDEN) != 0) {
                    e = (hrecelm) next ();
                    continue;
                }
                bf.append (e.var+"=");
                
                switch (e.fmt) {
                    case hrecelm.HRF_INT:
                        bf.append (""+Integer.toString (e.ival)+",");
                        break;
                    case hrecelm.HRF_DOUBLE:
                        bf.append (""+Double.toString (e.dval)+",");
                        break;
                    default:
                        bf.append ("'"+slashify (e.val)+"',");
                        break;
                }
                e = (hrecelm) next ();
            }
            bf.setLength (bf.length ()-1);
            bf.append (" "+qual);
            if ((rv = mdb.Query (bf.toString ())) > 0)
                return rv;
            if (cmd == UPDONLY)
                return rv;
        }
        bf.setLength (0);
        bf.append ("insert into "+table+ " (");
        /* yes this is a little lame traversing the hash table
           for the row twice, but it shouldn't be all that
           big, so it's easier to think of this way. I'm not
           the one who came up with the lame inconsistancies
           in SQL's syntax after all.
           */
        e = (hrecelm) first ();
        while (e != null) {
            if ((e.flags & hrec.HR_HIDDEN) != 0) {
                e = (hrecelm) next ();
                continue;
            }
            bf.append (e.var+",");
            e = (hrecelm) next ();
        }
        /* kill the trailing comma */
        bf.setLength (bf.length () - 1);
        bf.append (") values (");
        e = (hrecelm) first ();
        while (e != null) {        
            if ((e.flags & hrec.HR_HIDDEN) != 0) {
                e = (hrecelm) next ();
                continue;
            }
            
            switch (e.fmt) {
                case hrecelm.HRF_INT:
                    bf.append (Integer.toString (e.ival)+",");
                    break;
                case hrecelm.HRF_DOUBLE:
                    bf.append (Double.toString (e.dval)+",");
                    break;
                default:
                    bf.append ("'"+slashify(e.val)+"',");
                    break;
            }
            e = (hrecelm) next ();
        }
        bf.setLength (bf.length ()-1);
        bf.append (')');
        rv = mdb.Query (bf.toString ());
        return rv;        
    }
    
public
    int update (String table, String var, String val) throws IOException
    {
        String qual;
        qual = "where "+var+"='"+slashify(val)+"'";
        return _update (table, qual, UPDINS);
    }

public
    int update (String table, String var, int val) throws IOException
    {
        String qual;
        qual = "where "+var+"="+Integer.toString (val);
        return _update (table, qual, UPDINS);
    }
public
    int update (String table, String var, double val) throws IOException
    {
        String qual;
        qual = "where "+var+"="+Double.toString (val);
        return _update (table, qual, UPDINS);
    }

public
    int update (String table, String qual) throws IOException
    {
        return _update (table, qual, UPDINS);
    }


public
    int insert (String table) throws IOException
    {
        return _update (table, null, INSONLY);
    }




public
    int delete (String table, String field, String val) throws IOException
    {
        return mdb.Query ("delete from "+table+" where "+field+"='"+
                            slashify (val)+"'");
    }

public
    int delete (String table, String field, int val) throws IOException
    {
        return mdb.Query ("delete from "+table+" where "+field+"="+
                            Integer.toString (val));
    }

public
    int delete (String table) throws IOException
    {
        return mdb.Query ("delete from "+table);
    }
    



}

