

/*
 *                Copyright (c) Apr 10 19:15:10  MTCC
 *	Author: Michael Thomas
 *	Module: hparse.c
 *	Created: Thu Apr 10 19:15:10 1997
 *	Abstract:
 *	   main parsing routines. Though this module doesn't
 *         parse every entity, it's easy to extend to the
 *         elements which are interesting.
 */

/* Edit History: 
 */


package mtcc.libht;

import mtcc.libht.*;
import java.lang.*;
import java.io.*;
import java.net.*;
import java.util.*;


class hparseException extends IOException 
{
    hparse hp;
    String msg;
public
    hparseException (String s, hparse h) 
    {
        super (s);
        hp = h;
        msg = s;
    }
public
    String toString () 
    {
        return msg + "\n  line " + hp.lineno + " in file " + hp.curfile;
    }

}


public class hparse extends hdoc 
{
    

    private int ishtml;

public hparse () 
    {
        super ();
        ishtml = 1;
    }

public hparse (int parsetype) 
    {
        super ();
        ishtml = parsetype;
    }
    
    

public int
    parseString (String str)
    throws hparseException, IOException    
    {
        int rv;
        DataInputStream d;

        curfile = new String (str);
        lineno = 1;
        d = new DataInputStream (new StringBufferInputStream (str));
        
        rv = htread (d);
        return rv;
    }


public int
    parseFile (String file)
    throws hparseException, IOException
    {
        FileInputStream f;
        DataInputStream d;
        int rv;

        f = new FileInputStream (file);
        d = new DataInputStream (f);
        curfile = new String (file);
        lineno = 1;
        rv = htread (d);
        f.close ();
        return 0;
    }


/* read the given URL  */

public int
    parseURL (String url)
    throws hparseException, IOException
    {
        URL u;
        URLConnection c;
        int rv = 0;        
        InputStream inp;
        DataInputStream d;
        String bf;

        u = new URL (url);
        
        curfile = new String (url);
        c = u.openConnection ();
        inp = c.getInputStream ();        
        d = new DataInputStream (inp);
        // parse the headers
        bf = d.readLine ();
        contenttype = c.getContentType ();
        
//         bf = bf.substring (bf.indexOf (' ')+1, bf.length ());
//         bf = bf.substring (0, bf.indexOf (' ')).trim ();
//         status = Integer.parseInt (bf);
//         while ((bf = d.readLine ()) != null) {
//             bf = bf.trim ();
//             if (bf.length () == 0)
//                 break;
//             bf = bf.toLowerCase ();
//             if (bf.startsWith ("content-type:")) 
//                 contenttype = bf.substring (14, bf.length()).trim ();
//         }
            
//         /* if specifically html parse as such */        
        if (contenttype.equals ("text/html")) {
            ishtml = 1;
            rv = htread (d);
        }
        else if (contenttype.startsWith ("text/", 5)) {
            /* otherwise, if some other type of text, turn off html part */
            ishtml = 0;
            rv = htread (d);
        }
        /* else ignore it altogether. */
        inp.close ();
        return rv;
    }
    

protected int
    htread (DataInputStream fp)
    throws hparseException, IOException, FileNotFoundException
    {
        int rv, i = 0;
        char ch;
        htelm h, newh;
        int inquote = 0;
        int len = 0;
        byte bf [] = new byte [512];
        
        

        /* always splice in a new node */
        lineno = 1;
        
        h = this;
        this.lineno = 1;
        this.type = H_TXT;
        if (head == null)
            head = h;
        else {
            if (cur == null)
                cur = head;
        }

        // note: h here is really just "cur", but, we'll
        // use it cuz it's probably faster to access than a
        // class field.
        while (true) {
            if (i >= len) {
                try {
                    len = fp.read (bf);
                } catch (EOFException e) {
                    break;
                }
                catch (IOException e) {
                    break;
                }
                if (len <= 0)
                    break;
                i = 0;
            }
            ch = (char) bf [i++];
                            
            /* fudge the character to be something uninteresting
               if we're not parsing html */
            switch (ishtml != 0 ? ch : ' ') {
                case '<':
                    if (inquote != 0) {
                        h.addc (ch);
                        break;
                    }
                    /* start a new tag, only if necessary */
                    if (h.type != H_TXT) {
                        throw new hparseException
                            ("spurious open tag around '"+
                             h.blink.body.toString (), this);
                    }
                    if (h.type != H_TXT || h.len != 0) {
                        newh = new htelm (H_MISCTAG, lineno);
                        /* inherit uplink */
                        newh.ulink = h.ulink;
                        if (h.flink != null) {
                            newh.flink = h.flink;
                            newh.flink.blink = newh;
                        }                    
                    h.flink = newh;
                    newh.blink = h;
                    h = newh;
                } else
                    h.type = H_MISCTAG;                
                h.addc (ch);                
                break;                
            case '>':
                h.addc (ch);
                if (inquote != 0)
                    break;
                if (h.type != H_MISCTAG) {
                    if (h.blink != null)
                        throw new hparseException
                            ("spurious '>' around '"+ h.blink.body, this);
                    else
                        throw new hparseException                        
                            ("spurious '>'", this);
                    // not reached. break;
                }
                /* parse the tag, see if we should push/pop context */
                newh = new htelm (H_TXT, lineno);

                // we need to update cur so that any embedded includes
                // land in the right place.
                cur = h;
                rv = token.parsetag (h, this);
                h = cur;
                
                if (rv != 0) {
                    
                    if (rv > 0) {
                        /* start a newh list off of slink */
                        h.slink = newh;
                        newh.ulink = h;
                    } else {
                        if (h.ulink == null) {
                            throw new
                                hparseException ("unexpected close tag "+h.body, this);
                        }
                        /* pop */
                        h = h.ulink;
                        if (h.flink != null) {
                            newh.flink = h.flink;
                            newh.flink.blink = newh;
                        }
                        h.flink = newh;
                        newh.blink = h;
                        newh.ulink = h.ulink;
                    }
                } else {
                    if (h.flink != null) {
                        newh.flink = h.flink;
                        newh.flink.blink = newh;
                    }                    
                    h.flink = newh;
                    newh.blink = h;
                    newh.ulink = h.ulink;
                }
                
                h = newh;                
                break;
            default:
                if (ch == '\n' || ch == '\f') {
                    lineno++;
                    /* seal off the current tag */
                    if (h.type == H_TXT) {   
                        newh = new htelm (H_TXT, lineno);
                        if (h.flink != null) {
                            newh.flink = h.flink;
                            newh.flink.blink = newh;
                        }                    
                        h.flink = newh;
                        newh.blink = h;
                        newh.ulink = h.ulink;
                        h = newh;
                    }
                }
                if (h.type != H_TXT && ch == '\"')
                    inquote ^= 1;
                h.addc (ch);
                break;
            }
        }
        cur = h;
        return 0;
    }

    // override this if you want to redirect output
public
    void dumper (String s) 
    {
        System.out.print (s);
    }
    

public
    void dumpfile () 
    {
        dumpfile (this.head, H_NOTAG);
    }
public    
    void dumpfile (htelm h, int endtag) 
    {
        token tk;
        htelm p;
        
        while (h != null) {
            if (h.type == endtag) {
                break;
            }
            
            /* text and unknown goobers get emitted as is */
            if (h.type == H_TXT || h.tk == null) 
                dumper (h.body.toString ());
            else if (h.type < H_PSSTART) {
                dumper ("<");
                for (tk = h.tk; tk != null; tk = tk.flink) {
                    if ((tk.flags & token.TK_ISEND) != 0)
                        dumper ("/"+ tk.name);
                    else if ((tk.flags & token.TK_VAL) != 0) {
                        if ((tk.flags & token.TK_HASQ) != 0)
                            dumper ("=\""+tk.name+"\"");
                        else {
                            /* variable expansion? */
                            if (tk.name.charAt (0) == '$') {
                                String v = hcond.hgetenv (tk.name.substring (1));
                                if (v != null)
                                    dumper ("=\""+v+"\"");
                                else
                                    dumper ("=???");
                            } else                        
                                dumper ("="+tk.name);
                        }
                        
                    } else 
                        dumper (tk.name);
                    if (tk.flink != null && (tk.flink.flags & token.TK_VAL) == 0)
                        dumper (" ");                
                }
                dumper (">");
            } else {
                switch (h.type) {
                    case H_DATE: 
                    {
                        dumper (new Date().toString ());
                        break;
                    }
                    case H_TABLEDEF:
                        /* don't go down here. */
                        h = h.flink;
                        continue;
                    case H_IF:
                        if ((h.tk.flags & token.TK_ISEND) != 0)
                            break;
                        if (hcond.hcond (h) == 0) {
                            /* search the next level down foran else */
                            for (p = h.slink; p != null; p = p.flink)
                                if (p.type == H_ELSE)
                                    break;
                            if (p != null)
                                dumpfile (p.flink, H_NOTAG);
                            h = h.flink;
                        } else {
                            dumpfile (h.slink, H_ELSE);
                            h = h.flink;
                        }
                        // note that we don't exec the bottom
                        //of this loop.
                        continue;                        
                    default:
                        break;                    
                }
            }
            /* do not consider endtags of inferior branches */
            if (h.slink != null)
                dumpfile (h.slink, H_NOTAG);        
            h = h.flink;
        }
    }

    
private void
    indent (int n) 
    {
        int i;
        
        n *= 4;        
        for (i = 0; i < n; i++)
            System.out.print (" ");
    }
    
public 
    void dumptree (htelm h, int lvl) 
    {

        int n;
        
        if (lvl == 0) {
            h = this.head;
        }

        
        while (h != null) {
            if (h instanceof hparse) {
                hparse t = (hparse) h;
                System.out.println ("File "+t.curfile);
            }
            indent (lvl);
            System.out.print ("#"+ h.lineno + " ");
            if (h.type == htelm.H_TXT || h.tk == null) {
                System.out.println ("txt: '"+h.body.toString()+"'");
                
            } else if (h.type < htelm.H_PSSTART) {
                token tk;
                System.out.print ("tag: "+h.type+" <");        
                for (tk = h.tk; tk != null; tk = tk.flink) {
                    if ((tk.flags & token.TK_ISEND) != 0)
                        System.out.print ("/"+tk.name);
                    else if ((tk.flags & token.TK_VAL) != 0){
                        if ((tk.flags & token.TK_HASQ) != 0)
                            System.out.print ("=\""+tk.name+"\"");
                        else
                            System.out.print ("=" + tk.name);
                    } else {
                        System.out.print (tk.name);
                    }
                    if (tk.flink != null && (tk.flink.flags & token.TK_VAL) == 0)
                        System.out.print (" ");
                }
                System.out.println (">");
            } else {
                switch (h.type) {
                    case htelm.H_DATE: 
                    {
                        System.out.println ("meta: "+ new Date ().toString());
                        break;
                    }
                    default:
                        System.out.println ("meta: " + h.body.toString ());
                        break;
                }
            }
            if (h.slink != null) {
                dumptree (h.slink, lvl+1);
            }
            h = h.flink;
        }
        
    }
    
    
}


