
/*
 *                 Copyright (c) Mar 17 14:53:27  MTCC
 *	Author: Michael Thomas
 *	Module: hrec.java
 *	Created: Mon Mar 17 14:53:27 1997
 *	Abstract:
 *	   html record binding services
 */

/* Edit History: 
 */


package mtcc.libht;

import mtcc.libht.*;
import java.lang.*;
import java.io.*;


class hrecelm 
{
    String var;                 // copy of the variable name
    String val;                 /* char string value of the element */
    short flags;                /* various flags dealing with binding, etc. */
    short fmt;                  /* on input, elements can be verified */
public static final short HRF_TEXT   = 0;  /* text, unbound string array */
public static final short HRF_CHAR   = 1;  /* character array (bound by dimension) */
public static final short HRF_INT    = 2;  /* integer */
public static final short HRF_DOUBLE = 3;  /* floating point */
public static final short HRF_DATE   = 4;  /* canonical date */
public static final short HRF_URL    = 5;  /* a valid URL */
public static final short HRF_MONEY  = 6;  /* money type, (really a float) */
public static final short HRF_USER   = 0x4000;/* user defined */
    short subfmt;               /* either a dimension or qualifier (date style, say)*/
public static final short HRSF_DEC =  10;            /* int types */
public static final short HRSF_HEX =  16;
public static final short HRSF_OCT =  8;
    int ival;               /* integer value */
    double dval;            /* float value */
    String tval;            /* text value (same as val above) */
    Object uval;            /* user value -- prolly some struct */
    String errmsg;
    int errno;

    // simple constructor 
public
    hrecelm (String vr, String vl) 
    {
        var = vr;
        val = vl;
    }

}


public class hrec extends hashtable
{
    hrec flink;
    hrec blink;
    // these should rightfully be in hrecelm, but too bad.
    public static final int HR_HIDDEN =  1;
    public static final int HR_NOMERGE = 2;
    public static final int HR_NODUMP  = 4;


public
    hrec ()
    {
        super (23);        
    }

    hrec (int size) 
    {
        super (size);
    }


    // It would have been nice to override the class methods
    // hashtable,but java seems to not allow methods to
    // have the same signature except for return value.
    // groan.
public
    String gets (String var) 
    {
        hrecelm hr = (hrecelm) super.get (var);
        if (hr != null)
            return hr.val;
        else
            return "";
    }

public
    int getint (String var) 
    {
        hrecelm hr = (hrecelm) super.get (var);
        if (hr != null) {
            /* should I convert type here too? or should
               I just keep up with the sort of lazy type
               checking that I'm going. hmm.
               */
            if (hr.fmt != hrecelm.HRF_INT) {
                try {
                    hr.ival = Integer.parseInt (hr.val);
                } catch (NumberFormatException n) {
                    hr.ival = -1;
                }                
            }
            return hr.ival;            
        }
        else
            return -1;
    }
        
public
    double getdouble (String var) 
    {
        hrecelm hr = (hrecelm) super.get (var);
        if (hr != null) {
            /* should I convert type here too? or should
               I just keep up with the sort of lazy type
               checking that I'm going. hmm.
               */
            if (hr.fmt != hrecelm.HRF_DOUBLE) {
                try {
                    hr.dval = Double.valueOf (hr.val).doubleValue ();
                } catch (NumberFormatException n) {
                    hr.dval = -1;
                }                
            }
            return hr.dval;            
        }
        else
            return -1;
    }
public
    String getvis (String var) 
    {
        hrecelm hr = (hrecelm) super.get (var);
        if (hr != null && (hr.flags & HR_HIDDEN) == 0)
            return hr.val;
        else
            return "";
    }
public
    String isvis (String var) 
    {
        hrecelm hr = (hrecelm) super.get (var);
        if (hr != null && (hr.flags & HR_HIDDEN) == 0)
            return hr.val;
        else
            return null;
    }
    
    

public 
    void add (String var, String val)
    {
        hrecelm hr;

        if ((hr = (hrecelm) super.get (var)) != null) {
            hr.val = new String (val);
            // this needs to be done so that types
            // get appropriately converted.
            vrfybase(hr);
        } else {
            hr = new hrecelm (var, val);
            super.put (var, hr);
            settype (hr, hrecelm.HRF_TEXT, (short) 0);
        }
    }
public
    void add  (String var, int val) 
    {
        hrecelm hr;

        if ((hr = (hrecelm) super.get (var)) != null) {
            hr.val = ""+val;
        } else {
            hr = new hrecelm (var, ""+val);
            super.put (var, hr);
        }
        hr.ival = val;
        hr.fmt = hrecelm.HRF_INT;
    }
    
public
    void add  (String var, double val) 
    {
        hrecelm hr;

        if ((hr = (hrecelm) super.get (var)) != null) {
            hr.val = ""+val;
        } else {
            hr = new hrecelm (var, ""+val);
            super.put (var, hr);
        }
        hr.dval = val;
        hr.fmt = hrecelm.HRF_DOUBLE;
    }

public
    void append (String var, String val)
    {
        hrecelm hr;
        /* if it already exists, append it onto the end */
        if ((hr = (hrecelm) super.get (var)) != null) {
            hr.val = hr.val + "\n" + val;
        } else {
            hr = new hrecelm (var, val);
            hr.flags = 0;
            settype (hr, hrecelm.HRF_TEXT, (short) 0);            
            super.put (var, hr);
        }
    }


public
    void setflg (String var, int flags) 
    {
        hrecelm hr;
    
        if ((hr = (hrecelm) super.get (var)) == null) {
            add (var, "");
            hr = (hrecelm) super.get (var);
        }
        if (hr != null)
            hr.flags |= flags;            
    }

public
    void clrflg (String var, int flags) 
    {
        hrecelm hr;
    
        if ((hr = (hrecelm) super.get (var)) == null) {
            add (var, "");
            hr = (hrecelm) super.get (var);
        }
        if (hr != null)
            hr.flags &= ~flags;            
    }





/*
 * verify an hrec; optionally fills in a descriptor
 * with the offending element and error message.
 * returns null if it's ok, or the offending hrecelm
 * 
 */
public
    hrecelm vrfy () 
    {
        hrecelm hr;
        int i;
    
        /* traverse new, adding it into obeying nomerge flags */

        hr = (hrecelm) first ();        
        while (hr != null) {
            if (vrfybase (hr) < 0)
                return hr;
            hr = (hrecelm) next ();
        }
        return null;
    }

/*
 * set a preexisting hrec element referred to by name
 */
public 
    void settype (String var, short fmt, short subfmt) 
    {
        hrecelm hr;
        if ((hr = (hrecelm) super.get (var)) != null) {
            settype (hr, fmt, subfmt);
        }
    }
      

    /*
     * Set the type of a hrec element. These case be base
     * types, or user defined. If you don't specify a handler
     * for a non base type, it will generate an error.
     *
     * Note that you can always override
     * the base type verifier if you don't like what it does
     * (ie, say, make a text field require Shakespeare or somesuch)
     */    
public
    void settype (hrecelm hr, short fmt, short subfmt) 
    {
        hr.fmt = fmt;
        hr.subfmt = subfmt;
        vrfybase (hr);    
    }

    /*
     * verify that a given type is kosher handles the normal
     * cases, and passes off special user defined types to
     * their own handler. This also has the side effect of
     * setting the converted value field in the hrecelm 
     */
    
/*
 * find a match for field "var" in the list of
 * records starting at h, comparing for "val".
 */
public
    hrec find (String var, String val) 
    {
        hrec h;
    
        for (h = this; h != null; h = h.flink) {
            String s = h.gets (var);
            if (s == null)
                continue;
            if (s.equals (val))
                return h;
        }
        return null;
    }

/*
 * merge new into old
 */

public
    void merge (hrec newr)
    {
        hrecelm hr, o;
        
        hr = (hrecelm) newr.first ();
        while (hr != null) {
            /* obey both sets of nomerge flags */
            if ((hr.flags & HR_NOMERGE) != 0) {
                hr = (hrecelm) newr.next ();
                continue;
            }
            o = (hrecelm) super.get (hr.var);
            if (o != null && (o.flags & HR_NOMERGE) != 0) {
                hr = (hrecelm) newr.next ();
                continue;
            }
            /* update, adding if necessary */
            add (hr.var, hr.val);
            hr = (hrecelm) newr.next ();            
        }
    }

public
    hrec link (hrec list, hrec newr) 
    {
        newr.blink = null;
        newr.flink = list;
        if (list != null)
            list.blink = newr;
        return newr;            
    }

public
    hrec delink (hrec list, hrec old) 
    {
        /* delete the item */
        if (old.flink != null) 
            old.flink.blink = old.blink;
        if (old.blink != null) 
            old.blink.flink = old.flink;
        if (old == list)
            return list.flink;
        else
            return list;
    }
            
    //
    // These routines should probably have their own
    // class, but whatever.


   /*
    * read a post command's contents into an hrec
    */

public
    void getorpost (String bf)
    {
        char ch;
        int rv;
        int i;
        StringBuffer field = new StringBuffer ();
        StringBuffer val = new StringBuffer ();
        StringBuffer cur = field;

        for (i = 0; i < bf.length (); i++) {
            ch = bf.charAt (i);
            
            switch (ch) {
                case '&':
                    add (hdoc.unwebify (field.toString ()),
                         hdoc.unwebify (val.toString ()));
                    val.setLength (0);
                    field.setLength (0);
                    cur = field;
                    break;                    
                case '=':
                    cur = val;
                    break;
                case '\n':
                    break;
                case '+':
                    ch = ' ';
                default:
                    cur.append (ch);
            }
        
        }
        /* get the last dribble */
        if (field.length () > 0) {
            add (hdoc.unwebify (field.toString ()),
                 hdoc.unwebify (val.toString ()));
        }
    }




public
    void readget (String av []) 
    {
        String p = System.getProperty ("QUERY_STRING");
        int ac = av.length;
        int i;
    
        /* for reasons I profess to not understand,
           QUERY_STRING seems to always be accurrate.
           I've seen Apache send a null ac,av where
           query_string is correct. use it in preference.
           */
        if (p != null)
            getorpost (p);
        else 
            for (i = 0; i < ac; i++)
                getorpost (av [i]);
    }



public
    void readpost () 
    {
        StringBuffer bf = new StringBuffer ();
        int rv;

        /* read the form */
        try {
            while ((rv = System.in.read ()) > 0)
                bf.append ((char) rv);
        } catch (IOException e) {
           
        }
        getorpost (bf.toString ());
    }


    
public
    void getparams (String av []) 
    {
        String p = System.getProperty ("REQUEST_METHOD");
        if (p != null && p.equalsIgnoreCase ("POST"))
            readpost ();
        else
            readget (av);
    }

    
public
    int vrfybase (hrecelm hr) 
    {
        int rv;

        /* assume the best */
        rv = 0;
        hr.errno = 0;
        hr.errmsg = "";
        switch (hr.fmt) {
            case hrecelm.HRF_TEXT:
                /* always hunky dory */
                hr.tval = hr.val;
                break;
            case hrecelm.HRF_CHAR:
                if (hr.val.length () <= hr.subfmt)
                    hr.tval = hr.val;
                else {
                    hr.errmsg = "field too large";
                    rv = -1;
                }
                break;
            case hrecelm.HRF_INT: 
            {
                try {
                    hr.ival = Integer.parseInt (hr.val);
                } catch (NumberFormatException ne) {
                    hr.errmsg = "Invalid integer";
                    rv = -1;
                }
                break;
            }
            case hrecelm.HRF_DOUBLE:
                try {
                    hr.dval = (double) Double.valueOf (hr.val).doubleValue();
                } catch (NumberFormatException ne) {
                    hr.errmsg = "Invalid double value";
                    rv = -1;
                }                
                break;
            case hrecelm.HRF_DATE:
            case hrecelm.HRF_URL:
            case hrecelm.HRF_MONEY:
                rv = 0;
                break;
            default:
                rv = -1;             /* wtf? */
                break;

        }
        return rv;
    }

    

}
