/*
 *                 Copyright (c) May 24 17:22:29  MTCC
 *	Author: Michael Thomas
 *	Module: hrfile.java
 *	Created: Sat May 24 17:22:29 1997
 *	Abstract:
 *	   file read/writing routines with hrecs
 
 */

/* Edit History: 
 */

class hrfile extends hrec 
{
    


static hrec *    
readrec (FILE * db) 
{
    hrec * r = 0;
    hrecelm * h;
    char bf [256], * p;
    

    while (fgets (bf, 256, db) != NULL) {
        int l = strlen (bf);
        /* kill the stupid trailing nl */        
        if (l > 0 && bf [l-1] == '\n')
            bf [l-1]  = 0;
        
        /* toss comment lines */
        if (bf [0] == '#' || bf [0] == 0)
            continue;
        /* end of record */
        if (bf [0] == '\f') 
            return r;
        /* check the first word */
        if ((p = strchr (bf, ':')) == 0) {
            continue;
        }
        /* kill the colon and the leading white space */
        * p = 0;
        p++;
        while (* p && isspace (* p))
            p++;
        /* concatenate the strings if necessary */
        if (r == 0)
            r = newhrec ();        
        hrappend (r, bf, p, 0);
    }
    return r;
}

hrec * hrreadstrrec (hrec * r, char * str) 
{
    hrecelm * h;
    char bf [256], last [256], * p;
    
    last [0] = 0;
    while (1) {
        int l;
        if (p = strchr (str, '\n')) {
            l = p - str;
            memcpy (bf, str, l);
            bf [l] = 0;
            str = p+1;         /* past th */
        } else {
            if (* str == 0)
                break;
            strcpy (bf, str);
            str += strlen (str);
        }
        /* end of record */
        if (bf [0] == '\f') 
            return r;
        /* check the first word */
        if ((p = strchr (bf, ':')) == 0) {
            continue;
        }
        /* kill the colon and the leading white space */
        * p = 0;
        p++;
        while (* p && isspace (* p))
            p++;
        /* concatenate the strings if necessary */
        if (r == 0)
            r = newhrec ();
        if (! strcmp (bf, last))
            hrappend (r, bf, p, 0);
        else
            hradd (r, bf, p);
        strcpy (last, bf);
    }
    return r;
}

hrec * hrreadfp (FILE * fp)
{
    hrec * h, * head = 0, * cur;
    while (h = readrec (fp)) 
        if (head == 0)
            head = cur = h;
        else {
            cur->flink = h;
            h->blink = cur;
            cur = h;
        }
    return head;
    
}

hrec * hrreadfile (char * file) 
{
    FILE * fp = fopen (file, "r");
    hrec * h;

    if (fp == 0)
        return 0;
    
    h = hrreadfp (fp);
    fclose (fp);
    return h;
}

static
dumprec (FILE * fp, hrec * h) 
{
    int i;
    hrecelm * e;
    struct h_node * n;
    
    for (i = 0; i < h->htbl->h_size; i++)
        for (n = h->htbl->h_tab [i]; n; n = n->flink) {
            char * bf, * p, * l;
            /* this rigamoroll is due to needing to split text
               fields into multiple lines */
            e = n->h_obj;
            if (e->flags & HR_HIDDEN)
                continue;
            bf = alloca (strlen (e->val));
            strcpy (bf, e->val);            
            for (p = bf, l = p; * p; * p++) 
                if (* p == '\n') {
                    * p = 0;
                    fprintf (fp, "%s: %s\n", n->h_name, l);
                    l = p+1;
                }
            if (* l)
                fprintf (fp, "%s: %s\n", n->h_name, l);                
        }
    fprintf (fp, "\f\n");
}

int hrdumpstrrec (hrec * h, char * str, int len) 
{
    int i;
    hrecelm * e;
    struct h_node * n;
    
    for (i = 0; i < h->htbl->h_size; i++)
        for (n = h->htbl->h_tab [i]; n; n = n->flink) {
            char * bf, * p, * l;
            /* this rigamoroll is due to needing to split text
               fields into multiple lines */
            e = n->h_obj;
            if (e->flags & HR_HIDDEN)
                continue;
            bf = alloca (strlen (e->val));
            strcpy (bf, e->val);
            for (p = bf, l = p; * p; * p++) 
                if (* p == '\n') {
                    * p = 0;
                    len -= strlen (n->h_name)+strlen (l) + 3;
                    if (len < 3)
                        return -1;
                    sprintf (str, "%s: %s\n", n->h_name, l);
                    str += strlen (str);
                    l = p+1;
                }
            if (* l) {
                len -= strlen (n->h_name)+strlen (l) + 3;
                if (len < 3)
                    return -1;                
                sprintf (str, "%s: %s\n", n->h_name, l);
                str += strlen (str);
            }
        }
    if (len < 3)
        return -1;
    strcat (str, "\f\n");
    return 0;
}


hrwritefile (char * file, hrec * recs) 
{
    FILE * fp = fopen (file, "w");

    if (fp == 0)
        return -1;
    
    for (; recs; recs = recs->flink)
        dumprec (fp, recs);
    fclose (fp);
    return 0;
}
}
