/*
 *                 Copyright (c) May 24 18:45:21  MTCC
 *	Author: Michael Thomas
 *	Module: htable.java
 *	Created: Sat May 24 18:45:21 1997
 *	Abstract:
 *	   hashing primitives.
 */

/* Edit History: 
 */

/* Edit History: 
 */

package mtcc.libht;

import java.lang.*;
import mtcc.libht.*;

class h_node 
{
    h_node flink;		/* the forward bucket link */
    String h_name;
    Object h_obj;

    // empty constructor 
public 
    h_node (String s) 
    {
        h_name = s;
    }    
};


public class htable
{
    
    public int h_size;
    public int hits;
    public int depth;

    // various statistics produced by h_analyze
    int sz, tot, min, max, nempty, maxbk;
    double avg, hitdepth;

    
    
    private h_node seanode;
    private int seaidx;
    private h_node h_tab [];    
    
public
    htable (int size) 
    {
        h_size = size;
        h_tab = new h_node [size];
    }


public
    Object get (String s)
    {
        int idx = hash (s);		/* hash the string */
        h_node p;
        hits++;
        depth++;
        /* traverse the bucket */
        for (p = h_tab [idx]; p != null; p = p.flink, depth++)	
            if (p.h_name.equals (s))
                return p.h_obj;
        return null;
    }


public
    void put (String s, Object o)
    {
        int idx = hash (s);
        h_node p = new h_node (s);
        
        p.flink = h_tab [idx];		/* start linking it in */
        p.h_obj = o;			/* the object */
        h_tab [idx] = p;			/* done! */
    }


public
    void del (String s)
    {
        int idx = hash (s);		/* hash the string */
        h_node p, pprev = null;
        
        for (p = h_tab [idx]; p != null;  pprev = p, p = p.flink)
            /* traverse the bucket */
            if (p.h_name.equals(s)) {
                if (pprev != null) 
                    pprev.flink = p.flink;
                else
                    h_tab [idx] = p.flink;
                return;
            }
        return;
    }


    // traversal functions.
    
public
    h_node first () 
    {
        seaidx = 0;
        seanode = null;
        return next ();
    }

public
    h_node next () 
    {
        if (seanode != null && seanode.flink != null) {
            seanode = seanode.flink;
            return seanode;
        }
        for (; seaidx < h_size; seaidx++)
            if (h_tab [seaidx] != null) {
                seanode = h_tab [seaidx];
                return seanode;
            }
        return null;
    }
    
            
    

/* internal function to hash a string into an index, we will use modulo
   table size for grins for starters */

private
    int hash (String s)
    {
        int acc = 0, i = 0;
        int len = s.length (), j;
  
        for (j = 0; j < len; j++) {
            acc ^= (s.charAt (j) << i) + i;
            if (i < 16)
                i++;
            else
                break;
        }
        return acc % h_size;
    }


/*
 * analyse the contents of a hash table returning
 * the, minimum, maximum and average size of each
 * bucket
 */


public
    void h_analyse ()
    {
        int i, n;
        h_node p;
    
        min = Integer.MAX_VALUE;
        max = 0;
        nempty = 0;
        tot = 0;
        
        for (i = 0; i < h_size; i++) {
            /* count the hits. */
            for (p = h_tab [i], n = 0; p != null; p = p.flink, n++);
            if (n < min)
                min = n;
            if (n > max) {
                max = n;
                maxbk = i;
            }
            tot += n;
            if (n == 0)
                nempty++;
        }
        /* just do a simple minded average.
         * it would probably be nice to know about clustering
         * phenomenon too.
         */
        avg = (double) tot / (double) h_size ;
        sz = h_size;
    }

public
    void h_printstats (String name, int first) 
    {
        h_node p;
//        char * fmt = "%-7.7s%-4d    %-4d  %-4d  %-4d %-4d  %-6d  %.3f %.3f\n";
        if (first != 0)
            System.out.println
                ("table  buckets total min   max  empty hits    depth average");

        hitdepth = (double) depth / (double) hits;
        h_analyse ();
        // Fucking Java and its lack of formating. ick.
        System.out.print (name);
        System.out.print (" "+sz);
        System.out.print (" "+tot);
        System.out.print (" "+min);
        System.out.print (" "+max);
        System.out.print (" "+nempty);
        System.out.print (" "+hits);
        System.out.print (" "+hitdepth);
        System.out.println (" "+avg);        
    }

}

