/*
 *                 Copyright (c) Jun 20 14:20:51  MTCC
 *	Author: Michael Thomas
 *	Module: base.c
 *	Created: Fri Jun 20 14:20:51 1997
 *	Abstract:
 *	   base driver module for mtccbase
 */

/* Edit History: 
 */

#include "db.h"


/* open db dbname for access */

mtccdb *
mtccOpenDB (char * dbname) 
{
    mtccdb * db = calloc (sizeof (* db), 1);
    db->idx = h_init (23);
    db->frecs = h_init (23);
    db->fullrec = newhrec ();
    db->flat = flatinit (db);    /* eventually this could be changed to
                                   have a different flat db driver, like
                                   say a network version switched on the
                                   dbname */
    /* call the physical layer */
    if (db->flat->open (dbname, db) == 0) {
        hrfree (db->fullrec);
        h_destroy (db->frecs);
        h_destroy (db->idx);
        return null;
    }
    db->cleanrec = hrclone (db->fullrec);    
    return db;
}


/* create DB dbname.
   note: this doesn't open the db for access */

mtccCreateDB (char * dbname, hrec * r) 
{
    mtccdb * db = calloc (sizeof (* db), 1);
    int rv;
    
    db->idx = h_init (23);
    db->frecs = h_init (23);
    db->fullrec = newhrec ();
    db->flat = flatinit (db);     
    rv = db->flat->create (dbname, db, r);    
    hrfree (db->fullrec);
    h_destroy (db->frecs);
    h_destroy (db->idx);
    free (db);
    return rv;
}



mtccCloseDB (mtccdb * db) 
{
    closeidx (db);          /* close all secondary idx */
    db->flat->close (db);
    free (db->flat);
    hrfree (db->fullrec);
    h_destroy (db->frecs);
    h_destroy (db->idx);
    free (db);
    return 1;
}



/* kill DB dead. */

mtccDeleteDB (char * db)
{}


/* create an index of type itype of field 'field' */

mtccCreateIndex (mtccdb * db, char * field, int itype) 
{}

/* select a record(s) within the database */

mtccSelect (mtccdb * db, hrec * dset, scriteria * s) 
{
    dbindex * idx;
    
    idx = findidx (db, s);
    if (db->ctx) {
        db->ctx->idx->release (db, db->ctx);
        db->ctx = 0;
    }
    db->ctx = idx->select (db, idx, s);
    if (db->ctx != null)
        return mtccNext (db, dset);
    return 0;
}


/* start the traversal of the database, optionally using
   field (and its key) to guide the traversal along */

mtccFirst (mtccdb * db, hrec * dset) 
{
    dbindex * idx;
    idx = findidx (db, null);
    if (db->ctx) {
        db->ctx->idx->release (db, db->ctx);
        db->ctx = 0;
    }
    db->ctx = idx->first (db, idx);
    if (db->ctx != null)
        return mtccNext (db, dset);
    return 0;
    
}


/* get the next record of multi record search */

mtccNext (mtccdb * db, hrec * dset) 
{
    u32 rv;
    if (db->ctx == null || db->ctx->idx == null)
        return 0;
    rv = db->ctx->idx->next (db, db->ctx, dset);
    return rv;
}


/* update the record where field == val */

mtccUpdate (mtccdb * db, hrec * dset, scriteria * s) 
{
    dbindex * idx;
    int rv, n = 0;
        
    idx = findidx (db, s);
    if (db->ctx) {
        db->ctx->idx->release (db, db->ctx);
        db->ctx = 0;
    }
    /* first select the dataset */
    db->ctx = idx->select (db, idx, s);
    if (db->ctx != null) {
        while (idx->next (db, db->ctx, db->fullrec) > 0) {
            /* merge their dataset into the current record */
            hrupdate (db->fullrec, dset);
            rv = idx->update (db, db->ctx, db->fullrec);
            if (rv < 0)
                return rv;
            n += rv;
        } 
    }
    return n;
}


/* insert the dataset dset into the database */

mtccInsert (mtccdb * db, hrec * dset) 
{
    dbindex * idx;
    int rv;
    
    idx = findidx (db, null);
    rv = idx->insert (db, dset);
    return rv;
}



mtccDelete (mtccdb * db, scriteria * s) 
{
    dbindex * idx;
    int rv, n = 0;
    
    idx = findidx (db, s);    
    if (db->ctx) {
        db->ctx->idx->release (db, db->ctx);
        db->ctx = 0;
    }
    /* first select the dataset */
    db->ctx = idx->select (db, idx, s);
    if (db->ctx != null) {
        while (idx->next (db, db->ctx, 0) > 0) {
            rv = idx->delete (db, db->ctx);
            if (rv < 0)
                return rv;
            n++;
        } 
    }
    return n;
}



hrec *
mtccDefRec (mtccdb * db) 
{
    return hrclone (db->cleanrec);
}
