/*
 *                 Copyright (c) Jun 20 13:46:11  MTCC
 *	Author: Michael Thomas
 *	Module: db.h
 *	Created: Fri Jun 20 13:46:11 1997
 *	Abstract:
 *	   the main header file for mtccbase
 */

/* Edit History: 
 */

#ifndef h_mtccdb
#define h_mtccdb

#include <html.h>
#include "search.h"

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned long u32;
typedef int boolean;
#define null 0                  /* this makes java happy later */
#define true 1
#define false 0
#define BLKSIZE 32              /* in the layout file */
#define BLKMASK (BLKSIZE-1)
#define MAXFIELD 128            /* this should be enough. */

/*
 * the layout file contains a master block,
 * the fields, followed by data file deletion records.
 */

typedef struct 
{
    u32 fields;                 /* where the fields start */
    u32 drecs;                  /* where the deletion records start */
    u32 ndrecs;                 /* and their respective sizes */
    u16 nfields;
    u16 dbversion;              /* version of this database */
#define MTCCBASEVERSION 0       /* version 0 */
    u16 flags;
#define SUPPRESSIDX 1           /* suppress indicies */
    u8 pad [14];                /* pad to 32 bytes */
} mblock;


typedef struct
{
    u8 name [24];              /* the name of the field */
    u32 len;                    /* its length */
    u8 type;
    /* NOTE:
     *   these mimic hrf_xx for sanity, though not above 128.
     */    
#define F_TEXT   0                /* a bound char array */
#define F_CHAR   1                /* a text field, null terminated */
#define F_INT    2                /* a 4 byte integer */
#define F_REAL   3                /* a 8 byte double floating point */
#define F_BIN    7                /* a counted, though variable binary array */
#define F_IDX    128              /* start of index types */
#define F_FLAT   128              /* default flat file "index: */
#define F_BTREE  129              /* a btree */
#define F_HASH   130              /* a hash table */
    u8 pad [3];                 /* pad to 32 bytes */
} fieldrec;


typedef struct 
{
    u32 loc;
#define DRECDELETED 0xffffffff
#define LOCDELETED  0x80000000
    u32 len;
} delrec;

#define DRECSIZE 8
#define DRECINC 128             /* the increment size for realloc */
#define LENSIZE 4

#define scriteria clause

typedef struct 
{
    struct DBINDEX * idx;
    u32 currec;                 /* address of current record */
    u32 nextrec;                /* address of the next record */
    void * ctx;
    scriteria * s;              /* any search context */
} searchctx;

/* the search criteria */



typedef struct DBINDEX
{
    int ifile;                  /* file descriptor for the index */
    fieldrec * field;           /* pointer to descriptor */
    u8 type;                    /* type of index */
    void * idxdata;             /* attached data specific to the index */
    /* dispatch routines */    
    searchctx * (* select) ();    
    searchctx * (* first) ();
    int (* next) ();
    int (* update) ();
    int (* insert) ();
    int (* delete) ();
    
    int (* create) ();          /* create this index */
    int (* open) ();            /* open this index */
    int (* close) ();           /* close the index */
    int (* release) ();         /* release the search context */
} dbindex;




/* the main structure of an open database */

typedef struct 
{
    mblock m;                   /* the cached mblock */
    int dfile;                  /* the datafile descriptor */
    int lfile;                  /* the layout file */
    struct h_table * idx;       /* a hash table of the indicies */
    int nidx;                   /* and the number of them. */
    delrec * drecs;             /* an internal array of delrecs */
    int ndrecs;                 /* array size (m.ndrecs gives actual number) */
    struct h_table * frecs;     /* hash table of pointers to fieldrecs */
    hrec * fullrec;             /* an hrec containing an entire db rec */
    hrec * cleanrec;            /* an hrec containing an entire db rec */
    dbindex * flat;             /* there is always a "flat" index */
    searchctx * ctx;            /* current search context */
    u32 eof;                    /* where end of file is */
} mtccdb;


    
#endif
