/*
 *                 Copyright (c) Jun 20 15:02:41  MTCC
 *	Author: Michael Thomas
 *	Module: flat.c
 *	Created: Fri Jun 20 15:02:41 1997
 *	Abstract:
 *	   simple flat file "index" routines. This be de'fault.
 *         these routines actually implement the database
 *         datafile manipulation routines (unlike real indicies
 *         which just operate on their index.
 */

/* Edit History: 
 */

#include "db.h"
#include <fcntl.h>

static delrec flatalloc (), flatrealloc ();
static void flatfree ();

static searchctx *
flatselect (mtccdb * db, dbindex * idx, scriteria * sc) 
{
    searchctx * s = malloc (sizeof (* s));
    s->idx = idx;
    s->currec = 0;
    s->nextrec = 0;
    s->ctx = null;
    s->s = sc;
    
    return s;
}


static searchctx *
flatfirst (mtccdb * db, dbindex * idx) 
{
    searchctx * s = malloc (sizeof (* s));
    s->idx = idx;
    s->currec = 0;
    s->nextrec = 0;
    s->ctx = null;
    s->s = null;
    
    return s;
}


static int
flatnext (mtccdb * db, searchctx * ctx, hrec * r) 
{
    u8 curbf [LENSIZE], * bf;
    u32 cur;
    int rv;
    u32 next;


    next = ctx->currec = ctx->nextrec;
    if (ctx->currec >= db->eof)
        return 0;
    lseek (db->dfile, ctx->currec, 0);
    do {
        /* read the record length */
        if ((rv = readfully (db->dfile, curbf,  LENSIZE)) < LENSIZE) {
            return 0;
        }
        cur = vtohl (curbf);
        /* make certain it's not deleted */
        if (cur & LOCDELETED) {
            next += cur & (LOCDELETED-1);
            lseek (db->dfile, next, 0);
            continue;
        }
        /* this be payload */
        bf = alloca (cur);
        if (readfully (db->dfile, bf, cur-LENSIZE) < 0)
            return -1;
        unpack (db, bf, db->fullrec);            
        /* validate currec too */
        if (seamatch (ctx->s, r) == false) {
            next += cur;
            lseek (db->dfile, next, 0);
            continue;
        }        
        ctx->currec = next;
        ctx->nextrec = next+cur;
        if (r != null)
            hrupdate (r, db->fullrec);
        return 1;
        
    } while (1);
    return cur;
}

    
        
        
static
flatupdate (mtccdb * db, searchctx * ctx, hrec  * r) 
{
    int rv;
    u8 curbf [LENSIZE];
    u32 cur, len;
    u8 * bf;
    delrec d, nd;

    if (ctx->ctx == null) 
        ctx->ctx = h_init (1023);
    /* don't go beyond end of file */
    if (ctx->currec >= db->eof)
        return 0;
    /* don't re-traverse nodes which moved around */
    if (h_geti (ctx->ctx, ctx->currec) != null)
        return 0;
    lseek (db->dfile, ctx->currec, 0);
    if ((rv = readfully (db->dfile, curbf,  LENSIZE)) < LENSIZE) {
        if (rv == 0)
            return 0;
        else
            return -1;           
    }
    cur = vtohl (curbf);
    /* if it's already deleted, something's definitely fucked */
    if (cur & LOCDELETED)
        return -1;
    
    d.loc = ctx->currec;
    d.len = cur;
    /* try to realloc the region */
    bf = pack (db, r, &len);    
    nd = flatrealloc (db, &d, len);
    if (nd.loc == -1) {
        free (bf);
        return 0;
    }
    /* if this node's location changed, note it so we
       don't traverse it later */
    if (nd.loc != ctx->currec)
        h_puti (ctx->ctx, nd.loc, "");
    
    if ((rv  = writefully (db->dfile, bf, len)) < len) {
        free (bf);
        return rv;
    }
    /* update the new length */
    ctx->nextrec = d.loc + d.len;
    free (bf);
    return 1;
}


static
flatinsert (mtccdb * db, hrec * r) 
{
    u32 len;
    u8 * bf;
    int rv;
    delrec d;

    bf = pack (db, r, &len);
    /* allocate the region */
    d = flatalloc (db, len);
    if (d.loc == -1) {
        free (bf);
        return 0;
    }
    if ((rv  = writefully (db->dfile, bf, len)) < len) {
        free (bf);
        return rv;
    }
    free (bf);
    return 1;
}

static
flatdelete (mtccdb * db, searchctx * ctx) 
{
    int rv;
    u8 curbf [LENSIZE];
    u32 cur;
    delrec d;
    
    lseek (db->dfile, ctx->currec, 0);
    if ((rv = readfully (db->dfile, curbf,  LENSIZE)) < LENSIZE) {
        if (rv == 0)
            return 0;
        else
            return -1;           
    }
    cur = vtohl (curbf);
    /* if it's already deleted, something's fishy but
       ignore it */
    if (cur & LOCDELETED)
        return 1;
    d.loc = ctx->currec;
    d.len = cur;
    flatfree (db, &d);
}

    

static
flatupdidx () 
{}

/*
 * the flat index create is special in that it
 * creates the actual database; note that this doesn't
 * actually open the database for access; open does
 * that.
 */

static
flatcreate (char * name, mtccdb * db, hrec * r) 
{
    char bf [256];
    hrecelm * h;
    fieldrec fr;
    mblock m;
    delrec d;
    int n = 0;

    memset (&m, 0, BLKSIZE);
    memset (&fr, 0, BLKSIZE);
    
    sprintf (bf, "%s.lo", name);
    if ((db->lfile =  creat (bf, 0777)) < 0)
        goto error;
    sprintf (bf, "%s.dat", name);
    if ((db->dfile =  creat (bf, 0777)) < 0)
        goto error;
    if (lseek (db->lfile, BLKSIZE, 0) < 0)
        goto error;
    h = hrfirst (r);
    while (h != null) {
        if (n >= MAXFIELD-1)
            goto error;
        strcpy (fr.name, h->var);
        switch (h->fmt) {
            case HRF_TEXT:
                fr.len = 0;
                fr.type = F_TEXT;
                break;
            case HRF_CHAR:
                fr.len = h->subfmt;
                fr.type = F_CHAR;
                break;
            case HRF_BINARY:
                fr.len = 0;
                fr.type = F_BIN;
                break;                
               
            case HRF_INT:
                fr.len = 4;
                fr.type = F_INT;
                break;
            case HRF_DOUBLE:
                fr.len = 8;
                fr.type = F_REAL;
                break;
                
        }
        n++;
        
        if (write (db->lfile, &fr, BLKSIZE) < 0)
            goto error;
        h = hrnext (r);
    }
    if (lseek (db->lfile, 0, 0) < 0)
        goto error;
    m.fields = BLKSIZE;         /* past the mblock */
    m.nfields = n;
    m.drecs = MAXFIELD * BLKSIZE;       /* where the delete records start */
    m.ndrecs = 1;
    if (writefully (db->lfile, &m, BLKSIZE) < 0)
        goto error;
    /* create one invalidated drec to make things copescetic */
    d.loc = DRECDELETED;
    d.len = 0;
    lseek (db->lfile, m.drecs, 0);
    writefully (db->lfile, &d, DRECSIZE);
    close (db->lfile);
    close (db->dfile);
    return 1;
  error:
    /* ignore errors on closes */
    close (db->lfile);
    close (db->dfile);    
    return -1;
}

/* complete the base level flat file open on the database.
 */

static
int flatopen (char * name, mtccdb * db) 
{
    char bf [256];
    hrec * r = db->fullrec;
    fieldrec * fr;
    int i;

    sprintf (bf, "%s.lo", name);
    db->lfile = open (bf, O_RDWR);
    if (db->lfile < 0)
        goto error;
    sprintf (bf, "%s.dat", name);
    db->dfile = open (bf, O_RDWR);
    if (db->dfile < 0)
        goto error;

    readfully (db->lfile, &db->m, BLKSIZE);
    lseek (db->lfile, db->m.fields, 0);
    for (i = 0; i < db->m.nfields; i++) {
        fr = malloc (sizeof (fieldrec));        
        if (readfully (db->lfile, fr, BLKSIZE) < 0)
            goto error;
        switch (fr->type) {
            case F_TEXT:
                hradd (r, fr->name, "");
                h_put (db->frecs, fr->name, fr);
                break;                
            case F_CHAR:
                hraddbf (r, fr->name, "", fr->len);
                h_put (db->frecs, fr->name, fr);
                break;
            case F_INT:
                hraddint (r, fr->name, 0);
                h_put (db->frecs, fr->name, fr);
                break;
            case F_REAL:
                hradddouble (r, fr->name, 0.0);
                h_put (db->frecs, fr->name, fr);
                break;
            case F_BIN:
                hraddbin (r, fr->name, "", 0);
                h_put (db->frecs, fr->name, fr);
                break;
            case F_HASH:
            case F_BTREE:
                break;
            default:
                break;
        }
    }
    /* read the deletion records */
    if (db->m.ndrecs > 0) {
        db->drecs = malloc (DRECSIZE * db->m.ndrecs);
        db->ndrecs = db->m.ndrecs;
        lseek (db->lfile, db->m.drecs, 0);
        readfully (db->lfile, db->drecs, db->m.ndrecs*DRECSIZE);
    } else {
        /* just create one bogus, deallocated drec */
        db->drecs = malloc (DRECSIZE);
        db->ndrecs = 1;
        db->drecs [0].loc = DRECDELETED;
    }
    db->eof = lseek (db->dfile, 0, 2);
    return 1;
  error:
    close (db->lfile);
    close (db->dfile);
    return 0;
}

/* close access */
static
flatclose (mtccdb * db) 
{
    close (db->dfile);
    close (db->lfile);
}

/* release local search context */

static
flatrelease (mtccdb * db, searchctx * ctx) 
{
    if (ctx->ctx)
        h_destroyi (db->ctx);
}


/* main entry point which creates this class */

dbindex *
flatinit (mtccdb * db) 
{
    dbindex * rv = malloc (sizeof (* rv));
    rv->ifile = db->dfile;
    rv->select = flatselect;
    rv->first = flatfirst;
    rv->next = flatnext;
    rv->update = flatupdate;
    rv->insert = flatinsert;
    rv->delete = flatdelete;
    rv->create = flatcreate;
    rv->open = flatopen;
    rv->release = flatrelease;
    rv->close = flatclose;
    return rv;
}


/* local allocation routines */

static inline upddrec (mtccdb * db, int i) 
{
    int rv;
    u8 dloc [LENSIZE];
    delrec d;

    /* put in endian neutral format */
    htovl (db->drecs [i].loc, &d.loc);
    htovl (db->drecs [i].len, &d.len);
    
    lseek (db->lfile, db->m.drecs+((i)*DRECSIZE), 0);
    rv = writefully (db->lfile, &d, DRECSIZE);
    if (rv != DRECSIZE)
        return -1;
    /* now, update the datafile version */
    lseek (db->dfile, db->drecs [i].loc, 0);
    
    htovl (db->drecs [i].len | LOCDELETED, dloc);
    rv = writefully (db->dfile, dloc, LENSIZE);
    if (rv != LENSIZE)
        return -1;
    printf ("updd: %d = (%d,%d)\n", i, db->drecs [i].loc,db->drecs [i].len);
    
    return rv;
}

static inline updmblock (mtccdb * db) 
{
        
    lseek(db->lfile, 0, 0);
    return writefully (db->lfile, &db->m, BLKSIZE);
}


static delrec
flatalloc (mtccdb * db, u32 size) 
{
    int i;
    delrec d;

    /* round up to the next 32 byte block */
    if ((size & BLKMASK) != 0)
        size += BLKSIZE - (size & BLKMASK);
    /* do a first fit algorithm */
    for (i = 0; i < db->m.ndrecs; i++) 
        if (db->drecs [i].len >= size) {
            /* do we need to split the area? */
            d = db->drecs [i];            
            if (db->drecs [i].len == size) {
                db->drecs [i].loc = DRECDELETED;
                upddrec (db, i);
                lseek (db->dfile, d.loc, 0);
                return d;
            }
            /* else need to split it */
            d.loc = db->drecs [i].loc;
            d.len = size;
            db->drecs [i].len -= size;
            db->drecs [i].loc += size;
            upddrec (db, i);
            lseek (db->dfile, d.loc, 0);
            return d;
        }
    /* need to allocate from the end of the file */
    d.loc = lseek (db->dfile, 0, 2);
    d.len = size;
    db->eof = d.loc+d.len;
    return d;
}

/* reallocate a region.
   note that oldr might be modified to reflect the proper
   pointer to the next region */

static delrec
flatrealloc (mtccdb * db, delrec * oldr, u32 size) 
{

    delrec d;
    u32 end;
    int i;
    
    if ((size & BLKMASK) != 0)
        size += BLKSIZE - (size & BLKMASK);    
    /* is the new region smaller? */
    if (oldr->len >= size) {
        if (oldr->len == size) {
            lseek (db->dfile, oldr->loc, 0);
            return * oldr;
        }
        /* this will create a new drec, most likely */
        d.loc = oldr->loc + size;
        d.len = oldr->len - size;
        flatfree (db, &d);
        d.len = size;
        d.loc = oldr->loc;
        /* reflect the new size in oldr */
        oldr->len = size;
        lseek (db->dfile, d.loc, 0);
        return d;
    }
    /* next, check to see if there is an adjacent region
       free which will fulfill our request */
    end = oldr->loc+oldr->len;
    for (i = 0; i < db->m.ndrecs; i++) 
        if (db->drecs [i].loc == end) {
            /* no fit? */
            if (oldr->len + db->drecs [i].len < size)
                break;
            /* fits exactly */
            if (oldr->len + db->drecs [i].len == size) {
                d.loc = oldr->loc;
                d.len = size;
                db->drecs [i].loc = DRECDELETED;
                upddrec (db, i);
                lseek (db->dfile, d.loc, 0);
                /* update the length to update its new size */
                oldr->len = size;
                return d;
            }
            /* split it */
            d.loc = oldr->loc;
            d.len = size;
            db->drecs [i].loc += size - oldr->len;
            db->drecs [i].len -= size - oldr->len;
            upddrec (db, i);
            lseek (db->dfile, d.loc, 0);
            /* update the length to update its new size */
            oldr->len = size;
            return d;
        }
    /* else free the old, and allocate a new region */
    flatfree (db, oldr);
    return flatalloc (db, size);
    
}


static void
flatfree (mtccdb * db, delrec * d) 
{
    u32 end = d->loc+d->len;
    int s = -1, e = -1, avail = -1;
    int i;
    
    
    /* given a region, check its adjacencies */
    for (i = 0; i < db->m.ndrecs; i++) 
        if (db->drecs [i].loc + db->drecs [i].len == d->loc) {
            s = i;
            if (e != -1)
                break;
        } else if (db->drecs [i].loc == end) {
            e = i;
            if (s != -1)
                break;
        } else if (db->drecs  [i].loc == DRECDELETED) {
            avail = i;
        }
    /* 4 cases: both, one or the other, or neither */
    if (s != -1 && e != -1) {
        db->drecs [s].len += d->len + db->drecs [e].len;
        db->drecs [e].loc = DRECDELETED;
        upddrec (db, e);
        upddrec (db, s);
        /* it's completely changed, reflect that */
        * d = db->drecs [s];
    } else if (s != -1) {
        db->drecs [s].len += d->len;
        upddrec (db, s);
        /* update d too */
        * d = db->drecs [s];
    } else if (e != -1) {
        db->drecs [e].loc = d->loc;
        db->drecs [e].len += d->len;
        upddrec (db, e);
        /* and update d to reflect reality */
        d->len = db->drecs [e].len;
    } else {
        if (avail == -1) {
            /* add a new drec at the end of the file */
            db->m.ndrecs++;
            if (db->ndrecs < db->m.ndrecs) {
                db->ndrecs += DRECINC;
                db->drecs = realloc (db->drecs, db->ndrecs * DRECSIZE);            
            }
            /* update internal cache */
            db->drecs [db->m.ndrecs-1] = * d;
            upddrec (db, db->m.ndrecs-1);
            updmblock (db);
        } else {
            /* just use an available slot */
            db->drecs [avail] = * d;
            upddrec (db, avail);
        }
            
        
    }

}

