/*
 *                 Copyright (c) Jul  2 11:01:03  MTCC
 *	Author: Michael Thomas
 *	Module: hashidx.c
 *	Created: Wed Jul  2 11:01:03 1997
 *	Abstract:
 *	   hashing index routines
 */

/* Edit History: 
 */

#define TSIZE 8192
#define BKSIZE 8

#define BKOFF(hb, off) ((off+1)*(hb->bksize*4))

typedef struct
{
    u32 flink;                  /* next bucket */
    u32 load;                   /* loading factor of buckets */
    u16 tsize;                  /* table size */
    u8  bksize;                 /* bucket size */
    u8  ntabs;                  /* total number of tables */
    u8 pad [BKSIZE*4 - (3*4)];
} hashhead;


typedef struct 
{
    long cells [BKSIZE-1];
    long flink;
} bucket;

typedef struct 
{
    hashhead hh [];             /* cached hash header block */
    int ntabs;                  /* number of levels */
    u16 bksize;                 /* more convenient bucket and table size */
    u16 tsize;
} hblk;



/*
 * the layout of the hash file is straightforward,
 * though twofold: at the root is a hashtable which
 * contains a hashhead followed by HSIZE-1 buckets
 * where each bucket is BKSIZE. When a bucket overflows,
 * we do one of two things: create a new hash table
 * or create a bucket extension, which is based on
 * the loading factor of the table.
 */



static
hashselect (mtccdb * db, dbindex * idx, scriteria * sc) 
{}
    


/* hashing functions return an uninteresting first/next
 * pattern, so just call the flat file functions to
 * avoid our overhead.
 */

static
hashfirst (mtccdb * db, dbindex * idx) 
{
    return db->flat->first (db, idx);
    
}


static
hashnext (mtccdb * db, searchctx * ctx, hrec * r)
{}

static
hashupdate () 
{}


static
hashinsert (mtccdb * db, dbindex * idx, hrec * r) 
{
    hblk * hb = idx->idxdata;
    bucket * bk;
    
    switch (idx->field->type) {
        case F_TEXT:
        case F_CHAR:
        case F_BIN:
            bk = hashs (hrget (r, idx->field->name));
            break;
        case F_REAL:
            bk = hashd (hrgetdouble (r, idx->field->name));
            break;
        case F_INT:
            bk = hashi (hrgetint (r, idx->field->name));
            break;
    }
    /* now that it's hashed, find a slot for it */
    bk = alloca (sizeof (bucket)*hb->bksize);
    
    for (i = 0; i < hb->ntabs; i++) {
        /* read the bucket's contents */
        lseek (idx->ifile, BKOFF(hb, bk), 0);
        if (readfully (idx->ifile, bk, sizeof(bucket)*hb->bksize) <= 0)
            return -1;
        for (j = 0; j < BKSIZE-1; j++) {
            if (
        }
        
    }
}


static
hashdelete () 
{}

static hashcreate (char * dbname, mtccdb * db, dbindex * idx) 
{
    char bf [256];
    bucket b;
    hashhead h;
    
    sprintf (bf, "%s.hash", dbname);
    idx->ifile = creat (bf, 0777);
    if (idx->ifile < 0)
        return -1;
    idx->type = F_HASH;
    b = {0};
    h = {0};
    /* create the first table */
    h.ntabs = 1;
    h.tsize = TSIZE;
    h.blksize = BKSIZE;
    writefully (idx->ifile, &h, sizeof (h));
    
    for (i = 1; i < HSIZE; i++)
        write (idx->ifile, &b, sizeof (b));
    return 1;
}


static hashopen (char * dbname, mtccdb * db, dbindex * idx) 
{
    char bf [256];
    hashhead h;
    hblk * hb;
    int i, n;
    
    sprintf (bf, "%s.hash", dbname);
    idx->ifile = open (bf, O_RDWR);
    if (idx->ifile < 0)
        return -1;
    idx->type = F_HASH;
    idx->idxdata = malloc (sizeof (hashblk));
    /* read in the hash header blocks */
    if (readfully (idx->ifile, &h, sizeof (h)) <= 0) {
        close (idx->ifile);
        return -1;
    }
    hb = malloc (sizeof (* hb));
    n = h.ntabs;
    hb->hh = malloc (sizeof (hashhead)*n);
    hb->hh [0] = h;
    idx->idxdata = hb;
    hb->tsize = hb->hh [0].tsize;
    hb->bksize = hb->hh [0].bksize;
    
    for (i = 1; i < n; i++) {
        lseek (idx->ifile, hb->hh [i-1].flink, 0);
        if (readfully (idx->ifile, &hb->hh [i], sizeof (hashhead)) <= 0) {
            free (hb->hh);
            free (hb);
            close (idx->ifile);
            return -1;
        }
    }
    return 1;
}


static hashclose (mtccdb * db, dbindex * idx)
{
    close (idx->ifile);
}


static hashrelease () 
{}


hashinit () 
{}
