/*
 *                 Copyright (c) Jun 27 15:38:38  MTCC
 *	Author: Michael Thomas
 *	Module: seactx.c
 *	Created: Fri Jun 27 15:38:38 1997
 *	Abstract:
 *	   stuff for creating and manipulating search
 *         contexts
 */

/* Edit History: 
 */


#include "db.h"

/* check a basic expression */

static boolean checkexpr (), checkclause ();



expr *
mkexpr (hrecelm * h, int op) 
{
    expr * rv = malloc (sizeof * rv);
    rv->op = op;
    rv->val = h;
    rv->c = null;
}

clause * 
mkclause (expr * e, clause * c, int op) 
{
    clause * rv = malloc (sizeof (* rv));
    rv->op = op;
    rv->e = e;
    if (c != null) 
        e->c = c;
    return rv;
}


freeclause (clause * c) 
{
    expr * e;
    clause * tmp;
    
    while (c != null) {
        if (c->e != null) {
            tmp = c->e->c;
            free (c->e);
            free (c);
            c = tmp;
        } else
            break;
    }
}


    


static boolean
checkclause (clause * c, hrec * r) 
{
    int e, laste = false;
    
    while (c != null) {
        e = checkexpr (c->e, r);

        switch (c->op) {
            case C_AND:
                if (e == false || laste == false)
                    return false;                
            case C_OR:
                if (e == false && laste == false)
                    return false;
            case C_NOP:
                break;
            default:
                return false;   /* hosed */
        }
        c = c->e->c;        
        laste = e;
    }
    return laste;
}

                
        


static boolean
checkexpr (expr * e, hrec * r) 
{
    switch (e->op) {
        case E_EQ:
            switch (e->val->fmt) {
                case HRF_TEXT:
                    return strcmp (hrget (r, e->val->var), e->val->val) == 0;
                case HRF_CHAR:
                    return strncmp (hrget (r, e->val->var), e->val->cvtval.cval,
                                      e->val->subfmt) == 0;
                case HRF_INT:
                    return hrgetint (r, e->val->var) == e->val->cvtval.ival;
                case HRF_DOUBLE:
                    return hrgetdouble (r, e->val->var) == e->val->cvtval.dval;
                case HRF_BINARY:
                    return memcmp (hrgetbf (r, e->val->var, null), e->val->cvtval.cval,
                                   e->val->subfmt) == 0;
                    break;
            }
            break;
        case E_NEQ:
            switch (e->val->fmt) {
                case HRF_TEXT:
                    return strcmp (hrget (r, e->val->var), e->val->val)  != 0;
                case HRF_CHAR:
                    return strncmp (hrget (r, e->val->var), e->val->cvtval.cval,
                                    e->val->subfmt) != 0;
                case HRF_INT:
                    return hrgetint (r, e->val->var) != e->val->cvtval.ival;
                case HRF_DOUBLE:
                    return hrgetdouble (r, e->val->var) != e->val->cvtval.dval;
                case HRF_BINARY:
                    return memcmp (hrgetbf (r, e->val->var, null), e->val->cvtval.cval,
                                   e->val->subfmt) != 0;
                    break;
            }
            
        case E_GT:
            switch (e->val->fmt) {
                case HRF_TEXT:
                    return strcmp (hrget (r, e->val->var), e->val->val) > 0;
                case HRF_CHAR:
                    return strncmp (hrget (r, e->val->var), e->val->cvtval.cval, e->val->subfmt) > 0;
                case HRF_INT:
                    return hrgetint (r, e->val->var) > e->val->cvtval.ival;
                case HRF_DOUBLE:
                    return hrgetdouble (r, e->val->var) > e->val->cvtval.dval;
                case HRF_BINARY:
                    return memcmp (hrgetbf (r, e->val->var, null),
                                   e->val->cvtval.cval, e->val->subfmt) > 0;
                    break;
            }
            
        case E_GTE:
            switch (e->val->fmt) {
                case HRF_TEXT:
                    return strcmp (hrget (r, e->val->var), e->val->val) >= 0;
                case HRF_CHAR:
                    return strncmp (hrget (r, e->val->var), e->val->cvtval.cval,
                                      e->val->subfmt) >= 0;
                case HRF_INT:
                    return hrgetint (r, e->val->var) >= e->val->cvtval.ival;
                case HRF_DOUBLE:
                    return hrgetdouble (r, e->val->var) >= e->val->cvtval.dval;
                case HRF_BINARY:
                    return memcmp (hrgetbf (r, e->val->var, null), e->val->cvtval.cval,
                                     e->val->subfmt) >= 0;
                    break;
            }
            
        case E_LT:
            switch (e->val->fmt) {
                case HRF_TEXT:
                    return strcmp (hrget (r, e->val->var), e->val->val) < 0;
                case HRF_CHAR:
                    return strncmp (hrget (r, e->val->var), e->val->cvtval.cval,
                                    e->val->subfmt) < 0;
                case HRF_INT:
                    return hrgetint (r, e->val->var) < e->val->cvtval.ival;
                case HRF_DOUBLE:
                    return hrgetdouble (r, e->val->var) < e->val->cvtval.dval;
                case HRF_BINARY:
                    return memcmp (hrgetbf (r, e->val->var, null), e->val->cvtval.cval,
                                     e->val->subfmt) < 0;
                    break;
            }
            
        case E_LTE:
            switch (e->val->fmt) {
                case HRF_TEXT:
                    return strcmp (hrget (r, e->val->var), e->val->val) <= 0;
                case HRF_CHAR:
                    return strncmp (hrget (r, e->val->var), e->val->cvtval.cval,
                                    e->val->subfmt) <= 0;
                case HRF_INT:
                    return hrgetint (r, e->val->var) <= e->val->cvtval.ival;
                case HRF_DOUBLE:
                    return hrgetdouble (r, e->val->var) <= e->val->cvtval.dval;
                case HRF_BINARY:
                    return memcmp (hrgetbf (r, e->val->var, null), e->val->cvtval.cval,
                                   e->val->subfmt) <= 0;
                    break;
            }
            
        case E_REGEX:
            switch (e->val->fmt) {
                case HRF_TEXT:
                    return strcmp (hrget (r, e->val->var), e->val->val);
                case HRF_CHAR:
                    return strncmp (hrget (r, e->val->var), e->val->cvtval.cval, e->val->subfmt);
                case HRF_INT:
                    return hrgetint (r, e->val->var) == e->val->cvtval.ival;
                case HRF_DOUBLE:
                    return hrgetdouble (r, e->val->var) == e->val->cvtval.dval;
                case HRF_BINARY:
                    return memcmp (hrgetbf (r, e->val->var, null), e->val->cvtval.cval, e->val->subfmt);
                    break;
            }            
            break;
    }
}

            
            
boolean
seamatch (clause * c, hrec * r) 
{
    if (c == null)
        return true;
    return checkclause (c, r);
}

