/*
 *                 Copyright (c) Apr 17 16:46:04  MTCC
 *	Author: Michael Thomas
 *	Module: sock.c
 *	Created: Thu Apr 17 16:46:04 1997
 *	Abstract:
 *	   socket interface; isolates socket .h file cruft too.
 */

/* Edit History: 
 */

#include "html.h"               /* must come first, sorts through most includes */
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

/* open a socket at the given port. simplifies the krufty
   bsd socket stuff for the typical case.
 */

int sockopen (char * host, int port) 
{
    struct hostent * hp;
    struct sockaddr_in a;
    int fd;

    if (! (hp = gethostbyname (host))) 
        return -1;
    a.sin_family = AF_INET;
    memcpy (&a.sin_addr, hp->h_addr, 4);
    a.sin_port = htons (port);
    
    if ((fd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) 
        return -1;
    if (connect (fd, (struct sockaddr *) &a, sizeof (a)) < 0) {
        close (fd);
        return -1;
    }
    return fd;
}


FILE * fopenurl (char * url, int * psts, char * content) 
{
    URL purl;
    int sock;
    FILE * fp;
    char * p;
    int sts = 500;
    char bf [256];
    
    if (psts)
        * psts = sts;
    if (content)
        * content = 0;
    if (parseurl (url, &purl, 0) < 0)
        return 0;
    
    if ((sock = sockopen (purl.host, purl.port)) < 0)
        return 0;
    fp = fdopen (sock, "r+");
    /* kill name= urls */
    if ((p = strchr (purl.file, '#')))
        * p = 0;
    sprintf (bf, "GET %s%s HTTP/1.0\n\n", purl.path, purl.file);
    write (sock, bf, strlen (bf));
    /* scan headers */
    if (fgets (bf, 256, fp) == NULL) 
        goto err;
    sscanf (bf, "%*s %d %*s\n", &sts);
    if (sts >= 300)
        goto err;    
    while (fgets (bf, 256, fp) != NULL) {
        if (! strcmp (bf, "\n") || ! strcmp (bf, "\r\n"))
            break;
        if ((p = strchr (bf, '\r')) || (p = strchr (bf, '\n')))
            * p = 0;
        if (content && ! strncasecmp ("Content-type:", bf, 13))
            sscanf (bf, "%*s %s", content);            
    }
    if (psts)
        * psts = sts;
    return fp;
err:
    fclose (fp);
    if (psts)
        * psts = sts;

    return 0;
}

int listenon (char * host, int * startport) 
{
    struct hostent * hp;    
    struct sockaddr_in app;
    int lfd;
    int on = 1, len = sizeof (app);    
    
    app.sin_family = AF_INET;
    app.sin_port = htons (* startport);
    if ((hp = gethostbyname (host)))
        memcpy (&app.sin_addr, hp->h_addr, 4);
    else
        return -1;

    if ((lfd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) 
        return -1;

    /* if they specified a starting port, try to bind to it. */
    if (* startport) {
        if (setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR,
                       (char *)&on, sizeof (on)) < 0) {
            close (lfd);
            return -1;   
        }
        /* find a socket to bind to */
        while (bind (lfd, (struct sockaddr *) &app, sizeof (app)) < 0) {
            app.sin_port = htons (htons (app.sin_port)+1);
            if (app.sin_port == 0) {
                close (lfd);
                return -1;
            }
        }   
    }
    if (listen (lfd, 1) < 0) {
        close (lfd);
        return -1;
    }
    
    if (getsockname (lfd, (struct sockaddr *) &app, &len) < 0) {
        close (lfd);
        return -1;
    }
    * startport = htons (app.sin_port);
    return lfd;
}


int accepton (int fd, struct sockaddr_in * peer) 
{
    int len = sizeof (struct sockaddr_in);
    struct sockaddr_in t;

    if (peer == 0)
        peer = &t;
    return accept (fd, (struct sockaddr *) peer, &len);
}

/*
 * one shot deal to accept and listen to a socket.
 */

int sockaccept (char * host, int * startport, int * endport) 
{
    int fd, afd;
    struct sockaddr_in peer;
    
    if ((fd = listenon (host, startport)) < 0)
        return -1;
    if ((afd = accepton (fd, &peer)) < 0) {
        close (fd);
        return -1;
    }
    if (endport)
        * endport = htons (peer.sin_port);
    close (fd);
    return afd;
}
    
