/*
 *                 Copyright (c) Apr 16 13:48:43  MTCC
 *	Author: Michael Thomas
 *	Module: crawl.c
 *	Created: Wed Apr 16 13:48:43 1997
 *	Abstract:
 *	   crawl through the given web site and update
 *         the search database.
 */

/* Edit History: 
 */

#include <html.h>
#include <crawl.h>



int broken = 0;
int parseonly = 0;
int verify = 0;

void deleterows (void);
void insertrows (char * url);
void startdb (void);
extern int inswords (struct crawlnode * n, htelm * p,
                     int filenum, hrec * w, hrec * db);


int list (struct crawlnode * n) {
    if (broken) {
        if (n->file == 0 && n->sts >= 300)
            printf ("broken: file=%s link=%s sts=%d\n", n->parent, n->url, n->sts);
        else if (verify)
            printf ("file=%s link=%s sts=%d\n", n->parent, n->url, n->sts);
    } else if (parseonly == 0) {
        int i = n->level * 2;
        while (i-- > 0)
            putchar (' ');
        printf ("%s (%s)\n", n->url, n->type);
    }
    return 1;
}





int main (int ac, char ** av) 
{
    char * url;
    int doins = 0, dodel = 0;
    
    
    if (ac < 2) {
        fprintf (stderr, "crawl: url\n");
        exit (1);
    }
    av++;
    ac--;
    while (ac-- > 0) {
        if (av [0][0] != '-')
            break;
        switch (av [0][1]) {
            case 'v':
                verify++;
                break;
            case 'b':
                broken++;
                break;
            case 'p':
                parseonly++;
                hterrors (1);
                break;
            case 'P': 
            {
                URL u, b;
                if (ac > 1) {
                    parseurl (av [1], &b, 0);
                    parseurl (av [2], &u, &b);
                } else
                    parseurl (av [1], &u, 0);
                printf ("proto: %s\n", u.proto);
                printf ("host: %s\n", u.host);
                printf ("path: %s\n", u.path);
                printf ("file: %s\n", u.file);
                exit (0);
                break;
            }
            case 'a':
                doins++;
                break;
            case 'd':
                dodel++;
                break;
        }
        av++;
    }
    url = av [0];
    if (dodel || doins)
        startdb ();
    if (dodel)
        deleterows ();    
    if (doins) {
        insertrows (url);
    } else {
        if (broken == 0)
            hcrawl (url, CRL_EXTRAFIRST, list);
        else
            hcrawl (url, CRL_LEAFVISIT|CRL_EXTRAFIRST, list);
    }
    return 0;
}


#ifdef WANTMSQL

int siteno = 0;
int wordno = 0;
int wdbno = 0;
int dbsock;
struct h_table * dict;

int ins (struct crawlnode * n) 
{
    hrec * h, * w, * db;
    int site;

    /* ignore urls with parameter lists in them */
    if (n->file == 0 || strchr (n->url, '?'))
        return 1;
    printf ("%s: %s\n", n->type, n->url);
    h = newhrec ();
    hraddint (h, "fileno", 0);
    if (hmfind (dbsock, h, "site", "filename", n->url) == 0) {
        hradd (h, "filename", n->url);
        hradd (h, "site", n->parent);
        hraddint (h, "fileno", siteno++);        
        if (hminsert (dbsock, h, "site") < 0)
            fprintf (stderr, "site: %s\n", msqlErrMsg);
    }
    site = hrgetint (h, "fileno");
    hrfree (h);
    if (msqlQueryf (dbsock, "delete from wdb where fileno=%d", site) < 0) {
        fprintf (stderr, "delete: %s\n", msqlErrMsg);
        return 0;
    }
    w = newhrec ();
    db = newhrec ();

    if (inswords (n, n->file, site, w, db) < 0)
        return 0;
    printf ("\nfiles: %d dictionary: %d words: %d\n", site, wordno, wdbno);
    hrfree (w);
    hrfree (db);
    
    return 1;    
}

static char * getword (char * ctx, char * word) 
{
    char * w = word;
    int gotalpha = 0;
    
    for (; * ctx; ctx++) {
        if (isalpha (* ctx)) {
            * w++ = * ctx;
            gotalpha++;
        } else  if (isdigit (* ctx) || * ctx == '-' ||
                    * ctx == '_' || * ctx == '@') {
            * w++ = * ctx;
        } else if (isspace (* ctx)) {
            /* only if we got something. */
            if (w != word && gotalpha)
                break;
            w = word;
            gotalpha = 0;            
        } else if (* ctx == '.' || * ctx == ',' || * ctx == ';' ||
                   * ctx == '!' || * ctx == ':' || * ctx == '\"') {
            /* punctuation normally terminates a word;
             * there are some special cases though...
             */
            /* period embedded in a two strings is prolly part
               of a word like an email address */
            if (* ctx == '.' && isalnum (ctx [1])) {
                * w++ = * ctx;
                continue;
            }
            if (w != word && gotalpha)
                break;
            gotalpha = 0;
            w = word;
            continue;                   
        } else if (* ctx == '\'') {
            /* special case for these */
            if (w == word)
                continue;
            * w++ = * ctx;
        } else {
            /* something bizarre; reset w and gotalpha 
             * and scan for space
             */
            while (* ctx && ! isspace (* ctx))
                ctx++;
            w = word;
            gotalpha = 0;
        }
        
    }
    /* kill trailing quotes */
    if (w > word && w [-1] =='\'')
        w--;
    * w = 0;
    if (* word == 0 || gotalpha == 0)
        return 0;
    strlower (word);
    return ctx;
}


struct dict 
{
    long index;
    long filenum;
};



int inswords (struct crawlnode * n, htelm * p, int filenum, hrec * w, hrec * db)
{
    int rv;
    struct dict * d;


    hradd (w, "word", "");
    hraddint (w, "idx", 0);

    hraddint (db, "fileno", filenum);
    for (; p; p = p->flink) {
        if (p->type == H_TXT) {
            char * ctx = p->body;
            char word [256];
            if (verify == 1)
                printf ("\r#%-6d w=%d db=%d", p->lineno, wordno, wdbno);
            if (verify == 2)
                printf ("%s", p->body);

            fflush (stdout);
            hraddint (db, "lineno", p->lineno);
            while ((ctx = getword (ctx, word))) {
                int index;
                if (verify == 3)
                    printf ("word %s\n", word);
                
                /* check dictionary for word */
                if ((d = h_get (dict, word))) {
                    /* kill repeat words */
                    if (d->filenum == filenum)
                        continue;
                    index = d->index;
                    d->filenum = filenum;
                } else {
                    /* insert this word into the dictionary */
                    word [49] = 0;
                    hradd (w, "word", word);
                    hraddint (w, "idx", index = wordno++);
                    if (hminsert (dbsock, w, "dict") < 0) {
                        fprintf (stderr, "dict: %s\n", msqlErrMsg);
                        return -1;
                    }
                    d = malloc (sizeof (* d));
                    d->index = index;
                    d->filenum = filenum;
                    h_put (dict, word, d);
                }
                hraddint (db, "word", index);
                if ((rv = hminsert (dbsock, db, "wdb")) > 0) {
                    wdbno++;
                    continue;
                }
                if (rv < 0) {
                    fprintf (stderr, "wdb: %s\n", msqlErrMsg);
                    return -1;
                }                
            }   
        }
        if (p->slink)
            if (inswords (n, p->slink, filenum, w, db) < 0)
                return -1;
    }
    return 0;
}



void deleterows (void) 
{
    printf ("deleting...");
    msqlQuery (dbsock, "delete from site");
    msqlQuery (dbsock, "delete from wdb");
    msqlQuery (dbsock, "delete from word");
    printf ("done.\n");
}

void insertrows (char * url) 
{
    int ssiteno, swordno, swdbno;
    int rv;
    void * ctx = 0;
    hrec * hr;
        
    ssiteno = siteno = hmmax (dbsock, "site", "fileno")+1;
    if (verify > 0)
        swdbno = wdbno = msqlQuery (dbsock, "select word from wdb");
    else
        swdbno = wdbno = 0;
    dict = h_init (8191);
    /* read the dictionary into memory */
    hr = newhrec ();
    hraddint (hr, "idx", 0);
    hradd (hr, "word", "");
    
    while ((rv = hmfetchrowsf (dbsock, hr, "dict", &ctx, "")) > 0) {
        struct dict * d;
        d = malloc (sizeof (* d));
        d->filenum = 0;
        d->index = hrgetint (hr, "idx");
        h_put (dict, hrget (hr, "word"), d);
        wordno++;
    }
    swordno = wordno;
    hcrawl (url, 0, ins);
    printf ("\nfiles: %d dictionary: %d words: %d\n",
            siteno-ssiteno, wordno-swordno, wdbno-swdbno);
    h_printstats ("dict", dict, 1);
}

void startdb (void) 
{
    dbsock = msqlConnect ("localhost");
    if (dbsock < 0 || msqlSelectDB (dbsock, "search") < 0) {
        fprintf (stderr, "can't open database");
        exit (1);
    }
}



#else

void insertrows (char * url) 
{
    printf ("insert rows not supported.\n");
}

void deleterows (void) 
{
    printf ("delete rows not supported.\n");
}

void startdb (void) 
{}


#endif
