/*
 *                 Copyright (c) Apr 20 12:56:31  MTCC
 *	Author: Michael Thomas
 *	Module: search.c
 *	Created: Sun Apr 20 12:56:31 1997
 *	Abstract:
 *	   search for keywords given the site database
 */

/* Edit History: 
 */



#include <html.h>

#define GENERIC "/generic.html"
#define BASEURL "http://localhost"
void error (char * fmt, ...);
int search (hrec * in);

hrec * in;

#ifdef WANTMSQL


#define MAXMATCHES 100



int main (int ac, char ** av) 
{


    
    in = newhrec ();
    hradd (in, "server", "localhost");
    hradd (in, "db", "search");
    hradd (in, "template", GENERIC);
    hradd (in, "baseurl", BASEURL);
    
    in = hrgetparams (in, ac, av);
    preHTML ();
    if (htreadurl (hrget (in, "template"), 0, 0) == 0)
        error ("can't open output document.");
    if (htgotomark ("title"))
        htprintf ("MTCC's Search Engine Results");
    if (htgotomark ("pageheader"))
        htprintf ("Search Results");
    htgotomark ("pagebody");
    if (search (in) < 0) {
        error ("can't open database");
    }
    htdumpfile ();
    return 0;
}

char ** parsewords (char * s) 
{
    char ** rv, word [100], * p;
    int n = 0;
    
    rv = calloc (sizeof (* rv), 20);
    while (sscanf (s, "%99s", word) > 0) {
        strlower (word);
        for (p = word; * p; p++)
            if (* p == '*')
                * p = '%';
            else if (* p == '?')
                * p = '_';        
        rv [n++] = strdup (word);
        s += strlen (word);
        if (* s)
            s++;
    }
    return rv;
}


/*
 * Search for a set of words on a given socket.
 *
 * Why, you ask, is this broken into three
 * separate query phases? It's because the database
 * is layed out as a dictionary, a filelist, and
 * a database of wordno,fileno,lineno tuples.
 * The perfect situation for a big old sloppy
 * query to joint the three, right? Yes, if you
 * have more time on earth than Mathusla.
 */

int search (hrec * in) 
{
    hrec * hr, * pr, * xword = 0, * xrows = 0;
    void * ctx = 0;
    char qual [8192], * p, ** words, * base;
    int sock, rv, i;
    int matches = 0;
    
    sock = msqlConnect (hrget (in, "server"));
    if (sock < 0 || msqlSelectDB (sock, hrget (in, "db")))
        return -1;
    words = parsewords (hrget (in, "query"));
    if (* words == 0)
        return 0;
    
    p = qual;
    sprintf (p, "where ");
    p += strlen (p);
    for (i = 0; words [i]; i++) {
        sprintf (p, " %s word%s'%s'", i == 0 ? "" : "or",
                 strpbrk (words [i], "_%") ? " like " : "=",
                 slashify (words [i]));
        p += strlen (p);
    }
    
/*    htprintf ("qual: %s\n", qual);*/
    
    hr = newhrec ();
    hradd (hr, "word", "");
    hraddint (hr, "idx", 0);
    i = 0;
    while ((rv = hmfetchrowsf (sock, hr, "dict", &ctx, "%s", qual)) > 0) {
        hrlink (&xword, hr);
        hr = newhrec ();
        hradd (hr, "word", "");
        hraddint (hr, "idx", 0);
        i++;
    }
    if (i == 0) {
        htprintf ("word(s) not found.\n");
        return 0;
    }
    
    /* now construct the query to the file/word database */

    p = qual;
    sprintf (p, "where (");
    p += strlen (p);
    for (pr = xword; pr; pr = pr->flink) {
        sprintf (p, " word=%d or", hrgetint (pr, "idx"));
        p += strlen (p);
    }
    /* kill the last or */
    p -= 2;
    * p = 0;
    strcat (qual, ")");
/*    htprintf ("qual: %s\n", qual);*/
    
    hr = newhrec ();    
    hraddint (hr, "fileno", 0);
    hraddint (hr, "lineno", 0);
    hraddint (hr, "word", 0);
    ctx = 0;
    while ((rv = hmfetchrowsf (sock, hr, "wdb", &ctx, "%s", qual)) > 0) {
        hrlink (&xrows, hr);
        hr = newhrec ();    
        hraddint (hr, "fileno", 0);
        hraddint (hr, "lineno", 0);
        hraddint (hr, "word", 0);
        matches++;
        if (matches > MAXMATCHES)
            break;
    }
    if (rv < 0)
        htprintf ("wdb: %s\n", msqlErrMsg);
    hrfree (hr);


    p = qual;
    sprintf (p, "where (");
    p += strlen (p);

    for (pr = xrows; pr; pr = pr->flink) {
        sprintf (p, " fileno=%d or", hrgetint (pr, "fileno"));
        p += strlen (p);
    }
    /* kill the last or */
    p -= 2;
    * p = 0;
    strcat (qual, ")");

    ctx = 0;
    bq (1);
    dl (1);

    hr = newhrec ();
    hradd (hr, "filename", "");
    hraddint (hr, "fileno", 0);
    base = hrget (in, "base");
    hl (3);
    htprintf ("%s%d matches found.", matches > MAXMATCHES ? "&gt;" : "",
              matches);
    hl (-3);
    par ();

    while ((rv = hmfetchrowsf (sock, hr, "site", &ctx, "%s", qual)) > 0) {
        hrec * xw;
        char * word = "";
        htelm * ht, * head;

        if (* base && strncmp (base, hrget (hr, "filename"), strlen (base)))
            continue;
        /* lookup the word in the dictionary; we first
           have to relookup ourselves given the incoming
           fileno */
        for (pr = xrows; pr; pr = pr->flink)
            if (hrgetint (pr, "fileno") == hrgetint (hr, "fileno"))
                break;
        for (xw = xword; xw; xw = xw->flink)
            if (hrgetint (xw, "idx") == hrgetint (pr, "word")) {
                word = hrget (xw, "word");
                break;
            }
        par ();
        dt ();
        aHTML (hrget (hr, "filename"), 0);
        htprintf ("%s found in: %s\n", word, hrget (hr, "filename"));
        endaHTML ();
        par ();
        dd ();
#if 1
        if ((head = htpreadurl (0, 0, hrget (hr, "filename"), 0, 0)) &&
            (ht = htpgotoline (head, hrgetint (pr, "lineno")))) {
            for (i = 0; ht && i < 3; ht = ht->flink) {
                if (ht->type == H_TXT) {
                    htprintf ("    %s\n", ht->body);
                    br (1);
                    i++;
                }
            }
        }
        htprelease (head);
#endif        
    }
    if (rv < 0)
        htprintf ("wdb: %s\n", msqlErrMsg);
    dl (0);
    bq (-1);
    return 0;
}

#else

main (int ac, char ** av) 
{
    in = hrgetparams (in, ac, av);
    error ("you must have msql to run search\n");
}

#endif

void 
error (char * fmt, ...) 
{
    va_list ap;
    char bf [512];
    htelm * h;

    VA_START (ap, fmt);
    vsprintf (bf, fmt, ap);
    va_end (ap);
    
    hterase ();
    preHTML ();
    if ((h = htreadurl (hrget (in, "template"), 0, 0)) == 0)
        startHTML ("search error", hrget (in, "baseurl"));
    else {
        htgotomark ("title");
        htprintf ("search error");
    }
    htgotomark ("pageheader");
    par ();
    htprintf ("an error has occurred: %s\n", bf);
    if (h == 0)
        endHTML ();
    else
        htdumpfile ();    
    exit (0);
}
